//-----------------------------------------------------------------------------
// ControllerImpl and Controller method definitions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/Controller.h"
#include "Controller/ControllerFactory.h"
#include <list>
#include <vector>

namespace Cheetah {


//----------------------------------------------------------------------
//
// We store a vector of the registered handlers.
// The tag for each handler is the offset in the vector.
//
//----------------------------------------------------------------------

static std::vector<ControllerImpl::Handler_t> *handlers_s = 0;


//-----------------------------------------------------------------------------
//
// ControllerImpl Constructor.
//
// Initialize the ncontexts and mycontext values to invalid so that
// we can detect when a subclass doesn't initialize things properly.
//
//-----------------------------------------------------------------------------

ControllerImpl::ControllerImpl()
{
}


//-----------------------------------------------------------------------------
//
// ControllerImpl Destructor.
//
// Deletes world group and derived subgroups stored in the groups vector.
//
//-----------------------------------------------------------------------------

ControllerImpl::~ControllerImpl()
{
   std::vector<Group*>::iterator i;

   for (i = groups_m.begin(); i != groups_m.end(); i++) {
      delete *i;
   }
}


//-----------------------------------------------------------------------------
//
// ControllerImpl::registerHandler
//
// Because a function pointer is not sufficiently portable to send
// across the wire, we associate an integer with each local handler.
//
//-----------------------------------------------------------------------------

int ControllerImpl::registerHandler(Handler_t handler)
{
  if (handlers_s == 0)
    handlers_s = new std::vector<Handler_t>;

  handlers_s->push_back(handler);
  return handlers_s->size() - 1;
}


//-----------------------------------------------------------------------------
//
// ControllerImpl::registerHandler
//
// Retrieve the previously registered handler with the given tag
//
//-----------------------------------------------------------------------------

ControllerImpl::Handler_t ControllerImpl::getRegisteredHandler(int tag)
{
  if ((handlers_s==0) || (tag < 0) || (tag >= handlers_s->size()))
    return 0;

  return (*handlers_s)[tag];
}


//-----------------------------------------------------------------------------
//
// ControllerImpl::createGroup
//
// Creates and returns a new subgroup by partioning the old one.
//
//-----------------------------------------------------------------------------

const Group& ControllerImpl::createGroup(const Group& oldGroup,
					 int nContexts,
					 int* ranks)
{
  Group* group = new Group(nContexts, oldGroup.mycontext(), ranks);
  groups_m.push_back(group);
  return *group;
}


//-----------------------------------------------------------------------------
//
// Controller default constructor.
//
// The null ctor just lets the RefCountedPtr initialize itself to an
// invalid value and does not construct an implementation.
//
//-----------------------------------------------------------------------------

Controller::Controller()
{
}


//-----------------------------------------------------------------------------
//
// Controller main constructor.
// Builds the implementation using argc/argv.  It is up to
// the factory in the implementation to define the type of the
// implementation built.
//
//-----------------------------------------------------------------------------

Controller::Controller(int &argc, char ** &argv)
  : impl_m(ControllerFactory::create(argc, argv))
{
}


//-----------------------------------------------------------------------------
//
// Destructor.  The ref-counted implementation pointer will get
// cleaned up properly by the RefCountedPtr destructor.
//
//-----------------------------------------------------------------------------

Controller::~Controller()
{
}


} // namespace Cheetah
