#ifndef CHEETAH_CONTROLLER_CONTROLLER_H
#define CHEETAH_CONTROLLER_CONTROLLER_H

//-----------------------------------------------------------------------------
// Classes:
//   ControllerImpl
//   Controller
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
//
// Controller is a class that stores a reference-counted pointer to a
// ControllerImpl instance.  ControllerImpl is a base class for all Cheetah
// controllers, and provides an interface to the basic messaging operations
// in Cheetah:
//   int ncontexts()
//   int mycontext()
//   registerHandler()
//   ainvoke()
//   get()
//   put()
//   poll()
//   wait()
//   barrier()
//
// The implementations use C routines defined in subroutines of this dir.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/Group.h"
#include "MatchingHandler/Serialize.h"
#include "Utilities/CheetahRefCounted.h"
#include "Utilities/CheetahRefCountedPtr.h"
#include <vector>


namespace Cheetah {


//-----------------------------------------------------------------------------
//
// class ControllerImpl
//
// Abstract base class for all Cheetah controllers.  Defines an abstract
// interface that corresponds with the cheetah-core interface, but
// allows for run-time selection.  It is RefCounted so that the
// envelope class (Controller) can easily delete them at the
// right time.
//
//-----------------------------------------------------------------------------

class ControllerImpl : public RefCounted
{
public:

  //
  // Constructor.
  //
  // Initialize the ncontexts and mycontext values to invalid so that
  // we can detect when a subclass doesn't initialize things properly.
  //
  ControllerImpl();

  //
  // A virtual dtor so that the subclasses will get their dtors called
  // correctly.
  //
  virtual ~ControllerImpl();

  //
  // ncontexts, mycontext
  //
  // Each controller can access a given number of contexts at any one
  // time, and it numbers them from [0..ncontexts-1].  We keep this
  // data here and return it inline for efficiency. The subclasses have
  // to maintain these values.
  // 
  inline int ncontexts() const { return groups_m[0]->ncontexts(); }
  inline int mycontext() const { return groups_m[0]->mycontext(); }

  //
  // ngroups
  //
  // Returns the number of groups contained in this controller.
  // There is always at least one, the world group and additionally
  // there may be subgroups derived from the world group.
  //
  inline int ngroups() const { return groups_m.size(); }

  //
  // group
  //
  // Returns the group associated with the given id.  The world
  // communication group id is 0.
  //
  inline const Group& group(int id = 0) { return *groups_m[id]; }

  //
  // createGroup
  //
  // Creates and returns a new subgroup by partioning the old one.
  //
  virtual
  const Group& createGroup(const Group& oldGroup, int nContexts, int* ranks);

  //
  // Handler_t
  //
  // The handler for ainvoke is always of a particular type.
  //
  typedef void (*Handler_t)(int who, int tag, void* buf, int len);

  //
  // ainvoke
  //
  // Call a function (with a given tag) in another context.  Pass it a
  // buffer of data with len bytes.  Ring the local_bell when the
  // local buffer can be reused.
  //
  virtual void ainvoke(int context, int tag, void *buffer, int len,
		       int* local_bell) = 0;

  //
  // ainvoke
  //
  // Call a function (with a given tag) using the given communication group and
  // context within that group.  Pass it a buffer of data with len bytes.  Ring
  // the local_bell when the local buffer can be reused.
  //
  virtual void ainvoke(const Group& group, int context, int tag, void *buffer,
		       int len, int* local_bell)
    {
      ainvoke(group.rank(context), tag, buffer, len, local_bell);
    }

  //
  // put
  //
  // Move len bytes from the given local buffer to the remote buffer
  // in the another context.  Ring local_bell in this context when the
  // local buffer can be reused, and ring remote_bell in the remote
  // context when the buffer has successfully been written in that
  // context.
  //
  virtual void put(int context, void *remote, void *local, int len,
		   int* local_bell, int *remote_bell) = 0;

  //
  // get
  //
  // Move len bytes from a buffer in a remote context to a local
  // buffer. Ring remote_bell when the remote context can reuse the
  // data and ring local_bell when this context can use the copied
  // data.
  //
  virtual void get(int context, void *remote, void *local, int len,
		   int* local_bell, int *remote_bell) = 0;

  //
  // poll
  //
  // Ensure that things keep moving through the network, since we
  // don't assume that there is a background thread handling things.
  // Calling this could cause an arbitrary number of ainvokes, gets
  // and puts to be handled.
  //
  virtual void poll() = 0;

  //
  // wait
  //
  // Spin on the bell while handling messages (using poll) as long as
  // the bell is less than a given value.
  //
  virtual void wait(volatile int *bell, int value) = 0;

  //
  // barrier
  //
  // Wait (while handling messages) until all contexts have called
  // barrier.  If a group is passed to the barrier, only members of
  // the group participate in the barrier.
  //
  virtual void barrier() = 0;

  virtual void barrier(const Group& group) = 0;

  //
  // registerHandler
  //
  // Because a function pointer is not sufficiently portable to send
  // across the wire, we associate an integer with each local handler.
  //
  static int registerHandler(Handler_t handler);

  //
  // Handler_t getRegisteredHandler(int tag)
  //
  // Retrieve the previously registered handler with the given tag
  //
  static Handler_t getRegisteredHandler(int tag);

protected:

  //
  // Container for all groups, world group is groups_m[0];
  //
  std::vector<Group*> groups_m;

};


//-----------------------------------------------------------------------------
//
// class Controller
//
// The user interface class for all controllers.  It just maintains a
// pointer to a ControllerImpl and manages that memory.
//
//-----------------------------------------------------------------------------

class Controller
{
public:

  //
  // The null ctor just lets the RefCountedPtr initialize itself to an
  // invalid value and does not construct an implementation.
  //
  Controller();

  //
  // Builds the implementation using argc/argv.  It is up to
  // the factory in the implementation to define the type of the
  // implementation built.
  //
  Controller(int &argc, char ** &argv);

  //
  // Destructor.  The ref-counted implementation pointer will get
  // cleaned up properly by the RefCountedPtr destructor.
  //
  ~Controller();

  //
  // Handler_t
  //
  // Import the definition of the type of the handler.
  //
  typedef ControllerImpl::Handler_t Handler_t;

  //
  // isValid
  //
  // Return whether or not there is an implementation there.
  //
  inline bool isValid()
  {
    return impl_m.isValid();
  }

  //
  // Replicate the interface of the implementation here, and just
  // delegate operations to it.  All of these requre that there is an
  // implementation there.
  //

  inline int ncontexts() const
  {
    CheetahRTAssert(impl_m.isValid());
    return impl_m->ncontexts();
  }

  inline int mycontext() const
  {
    CheetahRTAssert(impl_m.isValid());
    return impl_m->mycontext();
  }

  inline int ngroups() const
  {
    CheetahRTAssert(impl_m.isValid());
    return impl_m->ngroups();
  }

  //
  // Group creators and accessors
  //

  inline const Group& createGroup(const Group& oldGroup, int nContexts,
				  int* ranks)
  {
    CheetahRTAssert(impl_m.isValid());
    return impl_m->createGroup(oldGroup, nContexts, ranks);
  }

  inline const Group& group(int id = 0) const
  {
    CheetahRTAssert(impl_m.isValid());
    return impl_m->group(id);
  }

  inline void ainvoke(int context, int tag, void *buffer, int len,
		      int* local_bell)
  {
    CheetahRTAssert(impl_m.isValid());
    impl_m->ainvoke(context,tag,buffer,len,local_bell);
  }

  inline void ainvoke(const Group& group, int context, int tag, void *buffer,
		      int len, int* local_bell)
  {
    ainvoke(group.rank(context), tag, buffer, len, local_bell);
  }

  //----------------------------------------------------------------------
  //
  // Series of ainvokeWith definitions follow.  Separate definitions are
  // provided for from 0 ... 4 arguments.
  //
  //----------------------------------------------------------------------

  template<class SerializeTag>					// 0 arguments
  inline void ainvokeWith(const SerializeTag &,
			  int context, int tag, int* local_bell)
  {
    CheetahRTAssert(impl_m.isValid());
    ainvoke(context, tag, 0, 0, local_bell);
  }

  template<class SerializeTag>					// 0 arguments
  inline void ainvokeWith(const SerializeTag& serial, const Group& group,
			  int context, int tag, int* local_bell)
  {
    ainvokeWith(serial, group.rank(context), tag, local_bell);
  }

  template<class SerializeTag, class T1>			// 1 argument
  inline void ainvokeWith(const SerializeTag &,
			  int context, int tag, int* local_bell,
			  const T1& x1)
  {
    CheetahRTAssert(impl_m.isValid());

    // Calculate buffer size
    int length = Serialize<SerializeTag, T1>::size(x1);
    CheetahRTAssert(length > 0);

    // Create buffer with data to send
    char *buffer = new char[length];
    CheetahRTAssert(buffer != 0);

    // Pack the data into this buffer
    int n = Serialize<SerializeTag, T1>::pack(x1, buffer);
    CheetahRTAssert(n <= length);
  
    // Send the buffer now.  If the last argument is '0' (instead of a
    // pointer to a 'bell' integer), the message will definitely get sent
    // before this returns so that we can free the buffer.
    ainvoke(context, tag, buffer, length, local_bell);

    // Free the buffer.
    delete [] buffer;
  }

  template<class SerializeTag, class T1>			// 1 argument
  inline void ainvokeWith(const SerializeTag& serial, const Group& group,
			  int context, int tag, int* local_bell,
			  const T1& x1)
  {
    ainvokeWith(serial, group.rank(context), tag, local_bell, x1);
  }

  template<class SerializeTag, class T1, class T2>		// 2 arguments
  inline void ainvokeWith(const SerializeTag &,
			  int context, int tag, int* local_bell,
			  const T1& x1, const T2& x2)
  {
    CheetahRTAssert(impl_m.isValid());

    // Calculate buffer size
    int length = Serialize<SerializeTag, T1>::size(x1) + 
                 Serialize<SerializeTag, T2>::size(x2);
    CheetahRTAssert(length > 0);

    // Create buffer with data to send
    char *buffer = new char[length];
    CheetahRTAssert(buffer != 0);

    // Pack the data into this buffer
    int n = Serialize<SerializeTag, T1>::pack(x1, buffer);
    n += Serialize<SerializeTag, T2>::pack(x2, buffer + n);
    CheetahRTAssert(n <= length);
  
    // Send the buffer now.  If the last argument is '0' (instead of a
    // pointer to a 'bell' integer), the message will definitely get sent
    // before this returns so that we can free the buffer.
    ainvoke(context, tag, buffer, length, local_bell);

    // Free the buffer.
    delete [] buffer;
  }

  template<class SerializeTag, class T1, class T2>		// 2 arguments
  inline void ainvokeWith(const SerializeTag& serial, const Group& group,
			  int context, int tag, int* local_bell,
			  const T1& x1, const T2& x2)
  {
    ainvokeWith(serial, group.rank(context), tag, local_bell, x1, x2);
  }

  template<class SerializeTag, class T1, class T2, class T3>	// 3 arguments
  inline void ainvokeWith(const SerializeTag &,
			  int context, int tag, int* local_bell,
			  const T1& x1, const T2& x2, const T3& x3)
  {
    CheetahRTAssert(impl_m.isValid());

    // Calculate buffer size
    int length = Serialize<SerializeTag, T1>::size(x1) +
                 Serialize<SerializeTag, T2>::size(x2) +
                 Serialize<SerializeTag, T3>::size(x3);
    CheetahRTAssert(length > 0);

    // Create buffer with data to send
    char *buffer = new char[length];
    CheetahRTAssert(buffer != 0);

    // Pack the data into this buffer
    int n = Serialize<SerializeTag, T1>::pack(x1, buffer);
    n += Serialize<SerializeTag, T2>::pack(x2, buffer + n);
    n += Serialize<SerializeTag, T3>::pack(x3, buffer + n);
    CheetahRTAssert(n <= length);
  
    // Send the buffer now.  If the last argument is '0' (instead of a
    // pointer to a 'bell' integer), the message will definitely get sent
    // before this returns so that we can free the buffer.
    ainvoke(context, tag, buffer, length, local_bell);

    // Free the buffer.
    delete [] buffer;
  }

  template<class SerializeTag, class T1, class T2, class T3>	// 3 arguments
  inline void ainvokeWith(const SerializeTag& serial, const Group& group,
			  int context, int tag, int* local_bell,
			  const T1& x1, const T2& x2, const T3& x3)
  {
    ainvokeWith(serial, group.rank(context), tag, local_bell, x1, x2, x3);
  }

  template<class SerializeTag, class T1, class T2, class T3, class T4>
  inline void ainvokeWith(const SerializeTag &,
			  int context, int tag, int* local_bell,
			  const T1& x1, const T2& x2,
			  const T3& x3, const T4& x4)
  {
    CheetahRTAssert(impl_m.isValid());

    // Calculate buffer size
    int length = Serialize<SerializeTag, T1>::size(x1) +
                 Serialize<SerializeTag, T2>::size(x2) +
                 Serialize<SerializeTag, T3>::size(x3) +
                 Serialize<SerializeTag, T4>::size(x4);
    CheetahRTAssert(length > 0);

    // Create buffer with data to send
    char *buffer = new char[length];
    CheetahRTAssert(buffer != 0);

    // Pack the data into this buffer
    int n = Serialize<SerializeTag, T1>::pack(x1, buffer);
    n += Serialize<SerializeTag, T2>::pack(x2, buffer + n);
    n += Serialize<SerializeTag, T3>::pack(x3, buffer + n);
    n += Serialize<SerializeTag, T4>::pack(x4, buffer + n);
    CheetahRTAssert(n <= length);
  
    // Send the buffer now.  If the last argument is '0' (instead of a
    // pointer to a 'bell' integer), the message will definitely get sent
    // before this returns so that we can free the buffer.
    ainvoke(context, tag, buffer, length, local_bell);

    // Free the buffer.
    delete [] buffer;
  }

  template<class SerializeTag, class T1, class T2, class T3, class T4>
  inline void ainvokeWith(const SerializeTag& serial, const Group& group,
			  int context, int tag, int* local_bell,
			  const T1& x1, const T2& x2,
			  const T3& x3, const T4& x4)
  {
    ainvokeWith(serial, group.rank(context), tag, local_bell, x1, x2, x3, x4);
  }

  inline void put(int context, void *remote, void *local, int len,
		  int* local_bell, int *remote_bell)
  {
    CheetahRTAssert(impl_m.isValid());
    impl_m->put(context,remote,local,len,local_bell,remote_bell);
  }

  inline void get(int context, void *remote, void *local, int len,
		  int* local_bell, int *remote_bell)
  {
    CheetahRTAssert(impl_m.isValid());
    impl_m->get(context,remote,local,len,local_bell,remote_bell);
  }

  inline void poll()
  {
    CheetahRTAssert(impl_m.isValid());
    impl_m->poll();
  }

  inline void wait(volatile int *bell, int value)
  {
    CheetahRTAssert(impl_m.isValid());
    impl_m->wait(bell,value);
  }

  inline void barrier(const Group& group)
  {
    CheetahRTAssert(impl_m.isValid());
    impl_m->barrier(group);
  }

  inline void barrier()
  {
    CheetahRTAssert(impl_m.isValid());
    impl_m->barrier();
  }

  //
  // registerHandler
  //
  // Because a function pointer is not sufficiently portable to send
  // across the wire, we associate an integer with each local handler.
  //
  inline static int registerHandler(Handler_t handler)
  {
    return ControllerImpl::registerHandler(handler);
  }

  //
  // Handler_t getRegisteredHandler(int tag)
  //
  // Retrieve the previously registered handler with the given tag
  //
  inline static Handler_t getRegisteredHandler(int tag)
  {
    return ControllerImpl::getRegisteredHandler(tag);
  }

private:

  //
  // The pointer to the implementation class.  This is refcounted to
  // give it shallow copy semantics.
  //

  RefCountedPtr<ControllerImpl> impl_m;
};


} // namespace Cheetah

#endif // CHEETAH_CONTROLLER_CONTROLLER_H
