//-----------------------------------------------------------------------------
// ControllerFactory method definitions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/ControllerFactory.h"
#include "Controller/Controller.h"
#include "Controller/SerialController.h"
#include "CheetahConfiguration.h"
#include <list>
#include <string>
#include <fstream>


#if CHEETAH_MPI_CONTROLLER
#include "Controller/MPI/MPIController.h"
#endif

#if CHEETAH_SHMEM_CONTROLLER
#include "Controller/Shmem/ShmemController.h"
#endif

#if CHEETAH_ULM_CONTROLLER
#include "Controller/ULM/ULMController.h"
#endif


namespace Cheetah {

//-----------------------------------------------------------------------------
//
// A static container of functions that will construct particular
// subclasses of ControllerImpl together with the name of the
// controller.
//
//-----------------------------------------------------------------------------

typedef std::list< std::pair<std::string, ControllerFactory::Factory_t> >
  FactoryContainer_t;

typedef FactoryContainer_t::iterator FactoryIterator_t;

static FactoryContainer_t *factories_s = 0;


//-----------------------------------------------------------------------------
//
// ControllerFactory::initialize
//
// Make sure the container has been set up and all the factory
// functions we know of have been registered.  It is necessary to
// list these here explicitely, instead of having some kind of
// global-object-constructor-automatic-registration scheme, because
// when you create a static lib these constructors do not get run on
// some platforms.
//
//-----------------------------------------------------------------------------

void ControllerFactory::initialize()
{
  // We only need to do this once, if the container is null

  if (factories_s == 0)
    {
      factories_s = new FactoryContainer_t;

      SerialController::registerFactory();

#if CHEETAH_MPI_CONTROLLER
      MPIController::registerFactory();
#endif

#if CHEETAH_SHMEM_CONTROLLER
      ShmemController::registerFactory();
#endif

#if CHEETAH_ULM_CONTROLLER
      ULMController::registerFactory();
#endif
    }
}


//-----------------------------------------------------------------------------
//
// ControllerFactory::create
//
// Scan the input args and see if any of them match the names of a
// registered subclass.  If you find one, build one of that type and
// return it.
//
//-----------------------------------------------------------------------------

ControllerImpl* ControllerFactory::create(int &argc, char ** &argv)
{
  //
  // Make sure all factories are registered
  //

  initialize();

  //
  // Loop over all of the command line args.
  //

  for (int i=1; i<argc; ++i)
    {
      //
      // Loop over all of the registered factories.
      //
      FactoryIterator_t p = factories_s->begin();
      FactoryIterator_t e = factories_s->end();
      for (; p!=e; ++p)
	{
	  //
	  // Do we have a match?
	  //
	  if (p->first == argv[i])
	    {
	      //
	      // We have a match!  Delete this arg from the input
	      // list.
	      //
	      for (int j=i+1; j<argc; ++j)
		argv[j-1] = argv[j];
	      --argc;

	      //
	      // Build this factory and return it.
	      //
	      return p->second(argc, argv);
	    }
	}
    }

  //
  // We didn't find a match, look for the controller type (ctype) in
  // the first line of the file CHEETAH_RUNTIME.  This can occur
  // because mpirun (mpich's version) doesn't pass in the correct command
  // line args (argc, argv are not valid until MPI_Init is called).
  //

  std::string ctype, path(argv[0]);	// look for file in app directory
  int ipos = path.find_last_of('/');
  if (ipos == path.npos) path.erase();
  else                   path.erase(ipos + 1);
  path.append("CHEETAH_RUNTIME");

  std::ifstream istrm(path.c_str());
  std::getline(istrm, ctype);
  if (ctype.length() > 0) 
    {
      FactoryIterator_t p = factories_s->begin();
      FactoryIterator_t e = factories_s->end();
      for (; p!=e; ++p)
	{
	  if (p->first == ctype)
	    {
	      return p->second(argc, argv);
	    }
	}
    }

  //
  // We still didn't find a match, return a "serial" controller that just has
  // one process if we can.   We must search through the list again
  // in this case to find the factory method.
  //

  FactoryIterator_t p = factories_s->begin();
  FactoryIterator_t e = factories_s->end();
  for (; p!=e; ++p)
    if (p->first == "-serial")
      return p->second(argc, argv);

  // OK, there is nothing we can do other that return null to indicate
  // we could not create a controller.

  return 0;
}


//-----------------------------------------------------------------------------
//
// ControllerFactory::registerFactory
//
// Sign up a new factory function that will respond to the given name.
// This function can then be called to create a new instance of the
// proper ControllerImpl.
//
//-----------------------------------------------------------------------------

void ControllerFactory::registerFactory(const char *name, Factory_t factory)
{
  initialize();
  std::string sname(name);
  factories_s->push_back(std::make_pair(sname, factory));
}


} // namespace Cheetah
