#ifndef CHEETAH_CONTROLLER_CONTROLLERFACTORY_H
#define CHEETAH_CONTROLLER_CONTROLLERFACTORY_H

//-----------------------------------------------------------------------------
// Classes:
//   ControllerFactory
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
//
// ControllerFactory takes an argc,argv set of arguments, parses the
// values, determines what specific controller implementation they
// request, and creates that instance.  If it cannot do this, it returns
// a null pointer.
//-----------------------------------------------------------------------------


namespace Cheetah {

//-----------------------------------------------------------------------------
// Forward references:
//-----------------------------------------------------------------------------

class ControllerImpl;


//-----------------------------------------------------------------------------
//
// class ControllerFactory
//
// A factory class that maintains a list of ControllerImpl
// factory functions (that generate implementation instances when
// called), and a mechanism to add these functions and use them
// to create an implementation instance.  It also provides
// a static method "initialize" that will make sure all the
// available factory methods have been registered.
//
// All the methods in this class are static, it is not intended
// that you actually create instances of this factory.
//
//-----------------------------------------------------------------------------

class ControllerFactory
{
public:

  //
  // Typedef for factory generation functions.
  //

  typedef ControllerImpl* (*Factory_t)(int &argc, char ** &argv);

  //
  // initialize
  //

  // Checks to see if the factory methods have been registered
  // yet.  If they have not been, this registers them all.

  static void initialize();

  //
  // create
  //

  // Given command line arguments, return a ControllerImpl using
  // them.  This scans the list of registered types of controllers and
  // builds the one that matches one of the command line arguments.
  // The command-line options that are used are then removed.

  static ControllerImpl* create(int &argc, char ** &argv);

  //
  // registerFactory
  //

  // Add a new factory method to our available list, with an
  // associated name.  That name is used as the command-line
  // option to check for when creating new instances.

  static void registerFactory(const char *name, Factory_t);
};


} // namespace Cheetah

#endif // CHEETAH_CONTROLLER_CONTROLLERFACTORY_H

// ACL:rcsinfo
// ----------------------------------------------------------------------
// $RCSfile: ControllerFactory.h,v $   $Author: rasmussn $
// $Revision: 1.1 $   $Date: 2000/04/17 19:01:29 $
// ----------------------------------------------------------------------
// ACL:rcsinfo

