#ifndef CHEETAH_CONTROLLER_GROUP_H
#define CHEETAH_CONTROLLER_GROUP_H

//-----------------------------------------------------------------------------
// Classes:
//   Group
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
//
// Group is a class that stores information about a communication group.
// A default world group is associated with each controller and this group
// can be partitioned into subgroups.  The main role of a group is to present
// a mapping between ranks in the group and ranks in default controller group.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Utilities/CheetahAssert.h"
#if CHEETAH_MPI_CONTROLLER == CHEETAH_YES
# include <mpi.h>
#endif

namespace Cheetah {

//-----------------------------------------------------------------------------
//
// class Group
//
// Group is a class that stores information about a communication group.
// A default world group is associated with each controller and this group can
// be partitioned into subgroups.  The main role of a group is to present a
// mapping between ranks in the group and ranks in the default controller group.
//
//-----------------------------------------------------------------------------

class Group
{
public:

  enum context_t { NOT_A_MEMBER = -1 };

  //
  // Constructors.  As provided, the default constructor creates an invalid
  // state which must be corrected with the initialize() function.
  //

  Group() : nContexts_m(-1), myContext_m(NOT_A_MEMBER), ranks_m(0), id_m(0) { }
  
  Group(int nContexts, int myContext, int* ranks = 0, int id = 0);

  //
  // Declare but don't define to insure that it can't be used
  //
  Group(const Group& g);

  //
  // Destructor.
  //
  virtual ~Group();

  //
  // Declare but don't define to insure that it can't be used
  //
  Group& operator=(const Group& group);

  //
  // Initializes nContexts and myContext.  Assumes that ranks is
  // index ordered.
  //
  void initialize(int nContexts, int myContext, int* ranks = 0, int id = 0);

  //
  // ncontexts, mycontext
  //
  // Each group consists of a number of contexts numbered from
  // [0..ncontexts-1].  This ordering is based on rankings in the
  // controllers world group.
  // 
  inline int ncontexts() const
  {
    CheetahRTAssert(nContexts_m > 0);
    return nContexts_m;
  }

  //
  // If the calling context is not am member of this group, returns
  // Group::NOT_A_MEMBER.
  //
  inline int mycontext() const
  {
    CheetahRTAssert(myContext_m == NOT_A_MEMBER || myContext_m >= 0);
    return myContext_m;
  }

  //
  // rank
  //
  // Returns the global rank at the given index.
  //
  inline int rank(int index) const
  {
    CheetahRTAssert(index >= 0 && index < nContexts_m);
    if (ranks_m == 0)  return index;
    else               return ranks_m[index];
  }

  //
  // id
  //
  // Returns the global identifier of the group
  //
  inline int id() const { return id_m; }

  //
  // isMPIGroup
  //
  // Returns true if this is an MPIGroup, false otherwise.
  //
  virtual bool isMPIGroup() const { return false; }

  //
  // MPICommunicator
  //
  // Returns an MPICommunicator if this is an MPIGroup.
  // WARNING, it is an error to call this method unless
  // isMPIGroup() is true.
  //
#if CHEETAH_MPI_CONTROLLER == CHEETAH_YES
  virtual MPI_Comm MPICommunicator() const
  {
    CheetahRTInsist(false, "Must be an MPI Group obtained from an MPI Controller");
    return MPI_COMM_NULL;
  }
#endif

  
protected:

  //
  // Instance variables.  Note that if ranks_m is NULL, then
  // an index rank ordering is assumed, i.e., ranks[i] = i;
  //
  int myContext_m;
  int nContexts_m;
  int* ranks_m;
  int id_m;	// global identifier for group within a given controller

};
} // namespace Cheetah

#endif // CHEETAH_CONTROLLER_GROUP_H
