#ifndef CHEETAH_CONTROLLER_MPI_COMMDATA_H
#define CHEETAH_CONTROLLER_MPI_COMMDATA_H

//-----------------------------------------------------------------------------
// Classes:
//   MPICommData
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
//
// A class which maintains data for performing gets/puts and barriers
// using MPI communication.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/MPI/cheetah_mpi.h"


namespace Cheetah {

//-----------------------------------------------------------------------------
//
// class MPICommData
//
//-----------------------------------------------------------------------------

class MPICommData
{
public:

  //
  // Declare but don't define to insure that it can't be used
  //
  MPICommData();
  
  //
  // This is the only constructor for the class.  Requires the #
  // of contexts to participate in the barrier.
  //
  MPICommData(int nContexts);

  //
  // Declare but don't define to insure that it can't be used
  //
  MPICommData(const MPICommData&);

  //
  // Destructor.
  //
  virtual ~MPICommData();

  //
  // Declare but don't define to insure that it can't be used
  //
  MPICommData& operator=(const MPICommData&);

  //
  // Accessor methods
  //

  inline get_map_t* getMap() { return &outstanding_gets_m; }
  inline isend_list_t* isendMap() { return &outstanding_isends_m; }

  int next_put_tag();
  int next_get_tag();

  //
  // Methods to keep track of barrier counters
  //

  inline int barrierCounter(int context) const
    {
      return barrier_counters_m[ next_barrier_m[context] ];
    }

  inline void incrementBarrierCounter(int context)
    {
      ++barrier_counters_m[ next_barrier_m[context] ];
    }

  inline void resetBarrierCounter(int context)
    {
      barrier_counters_m[ next_barrier_m[context] ] = 0;
    }

  inline void switchBarrierCounters(int context) const
    {
      next_barrier_m[context] = 1 - next_barrier_m[context];
    }

protected:

  //
  // Holds the outstanding get operations we're expecting to hear back on.
  //

  get_map_t outstanding_gets_m;

  //
  // Holds a list of outstanding isends we're expecting to finish someday.
  //

  isend_list_t outstanding_isends_m;

  //
  // Unique tags are generated for each get and put data message
  // from next_xxx_tag.
  //

  int next_put_tag_m;		// next tag for put messages
  int next_get_tag_m;		// next tag for get messages

//
// Static data structures for barriers.  We have two counters that
// will be incremented in barriers.  There are two so we can switch
// back and forth between them to eliminate race conditions between
// multiple barriers.  next_barrier_s indicates the current active
// barrier for each remote proc.
//

  int barrier_counters_m[2];
  int* next_barrier_m;

};
} // namespace Cheetah

#endif // CHEETAH_CONTROLLER_MPI_COMMDATA_H
