//-----------------------------------------------------------------------------
// MPIController method definitions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/MPI/MPIController.h"
#include "Controller/MPI/cheetah_mpi.h"
#include "Controller/ControllerFactory.h"
#include "Controller/Group.h"
#include "Controller/MPI/MPIGroup.h"

namespace Cheetah {

//-----------------------------------------------------------------------------
//
// MPIController Constructor
//
// Pass the command line arguments to the cheetah_mpi library to set
// up the parallelism.  Create world group (with id = 0).
//
//-----------------------------------------------------------------------------

MPIController::MPIController(int &argc, char **&argv, bool init_MPI)
              : initMPI_m(init_MPI)
{
  cheetah_mpi_setup(&argc, &argv, initMPI_m, this);

  MPIGroup* group = new MPIGroup(MPI_COMM_WORLD, 0);
  groups_m.push_back(group);
  cheetah_mpi_newgroup(0, group->ncontexts());
}


//-----------------------------------------------------------------------------
//
// MPIController::Destructor
//
// Tell the underlying library to tear down the parallelism.
//
//-----------------------------------------------------------------------------

MPIController::~MPIController()
{
  cheetah_mpi_finalize(initMPI_m);
}


//-----------------------------------------------------------------------------
//
// MPIController::ainvoke
//
// To call a remote function, just tell the underlying library to do
// it.
//
//-----------------------------------------------------------------------------

void
MPIController::ainvoke(int context, int tag, void *buf, int len, int *bell)
{
  cheetah_mpi_ainvoke(context,tag,buf,len,bell);
}


//-----------------------------------------------------------------------------
//
// MPIController::put
//
// To put memory, just tell the underlying library to do it.
//
//-----------------------------------------------------------------------------

void
MPIController::put(int c, void *r, void *l, int len, int *lb, int *rb)
{
  cheetah_mpi_put(c,r,l,len,lb,rb);
}


//-----------------------------------------------------------------------------
//
// MPIController::get
//
// To get memory, just tell the underlying library to do it.
//
//-----------------------------------------------------------------------------

void
MPIController::get(int c, void *r, void *l, int len, int *lb, int *rb)
{
  cheetah_mpi_get(c,r,l,len,lb,rb);
}


//-----------------------------------------------------------------------------
//
// MPIController::poll
//
// Poll for potential incoming messages, just tell the underlying library to do
// it.  However, there may be subgroups, so the communicators related to the
// subgroups must also be polled.
//
//-----------------------------------------------------------------------------

void
MPIController::poll()
{
   std::vector<Group*>::iterator i;
   for (i = groups_m.begin(); i != groups_m.end(); i++) {
      MPIGroup* g = dynamic_cast<MPIGroup*>(*i);
      if (g->mycontext() != Cheetah::Group::NOT_A_MEMBER) {
        cheetah_mpi_poll(g);
      }
   }
}


//-----------------------------------------------------------------------------
//
// MPIController::wait
//
// To wait for a bell to reach a given value, just tell the underlying
// library to do it.
//
//-----------------------------------------------------------------------------

void
MPIController::wait(volatile int *bell, int value)
{
  cheetah_mpi_wait(bell,value);
}


//-----------------------------------------------------------------------------
//
// MPIController::barrier
//
// To do a barrier, just tell the underlying library to do it.
//
//-----------------------------------------------------------------------------

void
MPIController::barrier()
{
  cheetah_mpi_barrier();
}

void
MPIController::barrier(const Group& group)
{
  const MPIGroup* g = dynamic_cast<const MPIGroup*>(&group);
  CheetahRTAssert(g);
  if (g->mycontext() == Cheetah::Group::NOT_A_MEMBER) return;

  cheetah_mpi_barrier(g);
}


//-----------------------------------------------------------------------------
//
// The function used to create new implementations.  This is registered
// with a factory.
//
//-----------------------------------------------------------------------------

static ControllerImpl*
mpiControllerFactory(int &argc, char ** &argv)
{
  return new MPIController(argc, argv, true);
}


static ControllerImpl*
mpiControllerFactoryNoInit(int &argc, char ** &argv)
{
  return new MPIController(argc, argv, false);
}


//-----------------------------------------------------------------------------
//
// MPIController::registerFactory
//
// Register this subclass with the controller factory list.
//
//-----------------------------------------------------------------------------

void
MPIController::registerFactory()
{
  ControllerFactory::registerFactory("-mpi", mpiControllerFactory);
  ControllerFactory::registerFactory("-mpi_noinit",
				     mpiControllerFactoryNoInit);
}


//-----------------------------------------------------------------------------
//
// MPIController::createGroup
//
// Create and return a new subgroup by partioning the old one.  An MPIGroup
// is created and MPI specific things must be done.  If ranks == 0, then all
// processors not in the ranks array must be notified to create a new group.
// NOT IMPLEMENTED YET.
//
//-----------------------------------------------------------------------------

const Group& MPIController::createGroup(const Group& oldGroup,
					int nContexts,
					int* ranks)
{
  const MPIGroup* old = dynamic_cast<const MPIGroup*>(&oldGroup);
  CheetahRTAssert(old);

  int id = groups_m.size();

  MPIGroup* group = new MPIGroup(*old, nContexts, ranks, id);
  groups_m.push_back(group);
  cheetah_mpi_newgroup(id, group->ncontexts());

  return *group;
}

} // namespace Cheetah
