//-----------------------------------------------------------------------------
//
// MPI Cheetah function definitions.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/MPI/cheetah_mpi.h"
#include "Controller/MPI/MPIGroup.h"
#include "Controller/MPI/MPIController.h"
#include "Controller/MPI/MPICommData.h"
#include <vector>
#include <stdio.h>

//
// We have a special function for handling barriers since barriers
// need to be set up as part of initialization.  A special function
// is also needed to create new MPI groups for contexts not in
// the new group.  This is needed because ALL MPI contexts
// must create a new communicator.
//

static int barrier_tag_s = -1;
static int newgroup_tag_s = -1;

static void mpi_barrier_handler(int, int, void*, int);
static void mpi_newgroup_handler(int, int, void*, int);

static Cheetah::MPIController* controller_s = 0;

//
// Container for MPI communication data.  Location corresponds
// to group id, i.e., comm_data_s[i] -> group[i]
//
static std::vector<Cheetah::MPICommData*> comm_data_s;


//----------------------------------------------------------------------
//
// cheetah_mpi_setup
//
// Set up the data structures for one sided messaging.
//
//----------------------------------------------------------------------

void cheetah_mpi_setup(int* argc, char*** argv,
		       bool init_MPI, Cheetah::MPIController* con)
{
  //
  // Initialize MPI (conditionally based on init_MPI flag)
  //
  if (init_MPI) MPI_Init(argc,argv);

  controller_s = con;

  //
  // Register the barrier and new group handlers.
  //
  barrier_tag_s = Cheetah::Controller::registerHandler(mpi_barrier_handler);
  newgroup_tag_s = Cheetah::Controller::registerHandler(mpi_newgroup_handler);
}


//----------------------------------------------------------------------
//
// cheetah_mpi_finalize
//
// Shut down the system.  Do a barrier to make sure everyone gets
// here, then shut down MPI.  If doFinalizeMPI is false, don't
// call MPI_Finalize as someone else is responsible for this.
//
//----------------------------------------------------------------------

void cheetah_mpi_finalize(bool finalize_MPI)
{
  if (finalize_MPI) {
    cheetah_mpi_barrier();
    MPI_Finalize();
  }

  for (int i = 0; i < comm_data_s.size(); i++) {
    delete comm_data_s[i];
  }
}


//----------------------------------------------------------------------
//
// cheetah_mpi_newgroup
//
// A new group had been created.  Create a new MPICommData object and
// place it in the comm_data_s vector in the same location as the
// id of the new group.
//
//----------------------------------------------------------------------

void cheetah_mpi_newgroup(int id, int nContexts)
{
  CheetahRTAssert(comm_data_s.size() == id);
  comm_data_s.push_back( new Cheetah::MPICommData(nContexts) );
}


//----------------------------------------------------------------------
//
// cheetah_mpi_ncontexts
// cheetah_mpi_mycontext
//
// Get the total number of contexts participating and which context
// this is.
//
//----------------------------------------------------------------------

int cheetah_mpi_ncontexts(const Cheetah::Group* g)
{
  return ( g ? g->ncontexts() : controller_s->group(0).ncontexts() );
}

int cheetah_mpi_mycontext(const Cheetah::Group* g)
{
  return ( g ? g->mycontext() : controller_s->group(0).mycontext() );
}


//----------------------------------------------------------------------
//
// cheetah_mpi_rfi_poll
//
// Check and see if there are any rfi requests coming in, and if there
// are, handle one.  Return a bool for whether we received something.
//
//----------------------------------------------------------------------

static bool cheetah_mpi_rfi_poll(const Cheetah::MPIGroup* g)
{
  int flag;
  MPI_Status status;

  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);
  if (group->mycontext() == Cheetah::Group::NOT_A_MEMBER) return false;

  MPI_Comm rfi_comm = group->RFICommunicator();

  //
  // Check for an incoming RFI request.  This could have any tag
  // (meaning it is a request for any registered function) but it has
  // to be coming in on the rfi communicator.
  //
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, rfi_comm, &flag, &status);
  if ( flag )
    {
      //
      // There's a message.  Get the size of the message and allocate
      // a buffer for it to land in.
      //
      int count;
      MPI_Get_count(&status, MPI_CHAR, &count);
      char *buffer = new char[ count ];

      //
      // Receive the buffer and block until it is done.  We know it
      // has arrived, so this should be quick.
      //
      MPI_Recv(buffer, count, MPI_CHAR, status.MPI_SOURCE,
	       status.MPI_TAG, rfi_comm, &status);

      //
      // Find the registered function for this message and call it.
      //
      int tag = status.MPI_TAG;
      Cheetah::Controller::Handler_t handler = 
        Cheetah::Controller::getRegisteredHandler(tag);

      if ( handler )
	handler(status.MPI_SOURCE, status.MPI_TAG, buffer, count);
      else
	{
	  //
	  // We didn't find the function!  Print an error and go on.
	  //
	  fprintf(stderr,
		  "CHEETAH-ON-MPI ERROR: Incoming message tag %d "
		  "has not been registered!\n",status.MPI_TAG);
	}

      //
      // Clean up the buffer.
      //
      delete [] buffer;
    }

  //
  // Return whether we did anything.
  //
  return (flag != 0);
}


//----------------------------------------------------------------------
//
// cheetah_mpi_get_poll
//
// Poll for get requests and responses.  If you get a request, send a
// response back.  If you get a response, put it where it belongs and
// ring the bell.
//
//----------------------------------------------------------------------

static bool cheetah_mpi_get_poll(const Cheetah::MPIGroup* g)
{
  int flag;
  MPI_Status status;

  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);
  if (group->mycontext() == Cheetah::Group::NOT_A_MEMBER) return false;

  MPI_Comm get_comm = group->getCommunicator();

  int id = group->id();
  isend_list_t* outstanding_isends = comm_data_s[id]->isendMap();
  get_map_t* outstanding_gets = comm_data_s[id]->getMap();

  //
  // Check to see if get messages came in.
  //
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, get_comm, &flag, &status);
  if ( flag )
    {
      //
      // There is a message, and it is either a request or a response.
      //
      if ( status.MPI_TAG == GET_REQUEST_TAG )
	{
	  //
	  // Pull the request off the wire.  This is blocking because
	  // we know it is there.
	  //
	  MPIGetRequest msg;
	  MPI_Recv(&msg, sizeof(msg), MPI_CHAR, status.MPI_SOURCE,
		   GET_REQUEST_TAG, get_comm, &status);

	  //
	  // Send back the response. We'll need to record the isend so
	  // we can check on it later.
	  //
	  outstanding_isends->push_back( MPIIsendRecord() );
	  MPIIsendRecord& rec = outstanding_isends->back();
	  rec.bell = msg.bell;
	  MPI_Isend(msg.remote, msg.length, MPI_CHAR,
		    status.MPI_SOURCE, msg.tag, get_comm, &rec.request);
	}
      else
	{
	  //
	  // When you get a response message, look up the place to put
	  // it and then put it there.  We know the tag from the
	  // probe. Use that to find the buffer to use for the
	  // receive.
	  //
	  get_iter_t p = outstanding_gets->find(status.MPI_TAG);

	  if ( p != outstanding_gets->end() )
	    MPI_Recv(p->second.local, p->second.length, MPI_CHAR,
		     status.MPI_SOURCE, status.MPI_TAG, get_comm, &status);
	  else
	    {
	      //
	      // It wasn't in the map which means we have an
	      // unrecognized get response!  Print an error message
	      // and go on.
	      //
	      fprintf(stderr,
		      "CHEETAH-ON-MPI ERROR: Unrecognized get response "
		      "tag %d!\n", status.MPI_TAG);
	    }

	  //
	  // If the user asked us to ring a bell when it came in, do so.
	  //
	  if ( p->second.bell )
	    *p->second.bell += 1;

	  //
	  // Get rid of the record of this get.
	  //
	  outstanding_gets->erase(p);
	}
    }
  //
  // Let the caller know if we did anything.
  //
  return (flag != 0);
}


//----------------------------------------------------------------------
//
// cheetah_mpi_put_poll
//
// Check for put messages.  There are two messages involved in a put:
// the header that has the pointer to the place to put it and the tag
// for the payload message, and then there is the payload message.
//
//----------------------------------------------------------------------

static bool cheetah_mpi_put_poll(const Cheetah::MPIGroup* g)
{
  int flag;
  MPI_Status status;

  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);
  if (group->mycontext() == Cheetah::Group::NOT_A_MEMBER) return false;

  MPI_Comm put_comm = group->putCommunicator();

  //
  // Check for a request on the put communicator.
  //
  MPI_Iprobe(MPI_ANY_SOURCE, PUT_REQUEST_TAG, put_comm, &flag, &status);
  if ( flag )
    {
      //
      // We have a request, so get the header.
      //
      MPIPutRequest msg;
      MPI_Recv(&msg, sizeof(msg), MPI_CHAR, status.MPI_SOURCE,
	       PUT_REQUEST_TAG, put_comm, &status);

      //
      // Use the info in the header to pull in the payload.
      //
      MPI_Recv(msg.remote, msg.length, MPI_CHAR, status.MPI_SOURCE,
	       msg.tag, put_comm, &status);

      //
      // If the user requested a doorbell, ring it.
      //
      if ( msg.bell )
	{
	  *msg.bell += 1;
	}
    }
  //
  // Let the user know if we did anything.
  //
  return (flag != 0);
}


//----------------------------------------------------------------------
//
// cheetah_mpi_isend_completion
//
// Check and see if any of the isends have completed.  Return true if
// any have.
//
//----------------------------------------------------------------------

bool cheetah_mpi_isend_completion(const Cheetah::MPIGroup* g)
{
  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);
  if (group->mycontext() == Cheetah::Group::NOT_A_MEMBER) return false;

  isend_list_t* outstanding_isends = comm_data_s[group->id()]->isendMap();

  //
  // Keep track of whether we did anything.
  //
  bool did_something = false;

  //
  // Loop over all of the outstanding isends.
  //
  isend_iter_t p = outstanding_isends->begin();
  isend_iter_t e = outstanding_isends->end();
  while (p != e)
    {
      //
      // Find out if this isend has completed.
      //
      int flag;
      MPI_Status status;
      MPI_Test(&p->request, &flag, &status);
      if ( flag )
	{
	  //
	  // It has completed, so ring the bell for this isend.
	  // 
	  if ( p->bell )
	    *p->bell += 1;

	  //
	  // Delete this entry in the list and go on to the next.
	  // Do this carefully so we don't have any dangling pointers.
	  //
	  isend_iter_t p1 = p;
	  ++p;
	  outstanding_isends->erase(p1);

	  //
	  // Record that we did something.
	  //
	  did_something = true;
	}
      else
	{
	  //
	  // This isend hasn't completed, so go on to the next one.
	  //
	  ++p;
	}
    }
  //
  // Tell the caller if we accomplished something.
  //
  return did_something;
}


//----------------------------------------------------------------------
//
// cheetah_mpi_poll
//
// Check the rfi, get and put queues.  Keep pulling things off the
// wire as long as there are things to do.
//
//----------------------------------------------------------------------

void cheetah_mpi_poll(const Cheetah::MPIGroup* g)
{
  if (g && g->mycontext() == Cheetah::Group::NOT_A_MEMBER) return;

  bool did_something;
  do
    {
      did_something = false;
      did_something |= cheetah_mpi_rfi_poll(g);
      did_something |= cheetah_mpi_get_poll(g);
      did_something |= cheetah_mpi_put_poll(g);
      did_something |= cheetah_mpi_isend_completion(g);
    }
  while ( did_something );
}


//----------------------------------------------------------------------
//
// cheetah_mpi_ainvoke
//
// Invoke a function in a remote context.  The function is given by
// the tag, and we pass a buffer with a given length to the function.
// Since we can use the function tag as the message tag, all we have
// to do is a send.
//
//----------------------------------------------------------------------

void cheetah_mpi_ainvoke(int context, int tag, void *buffer, int len,
		         int *local_bell,
			 const Cheetah::MPIGroup* g)
{
  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);
  if (group->mycontext() == Cheetah::Group::NOT_A_MEMBER) return;

  //
  // Keep things moving.
  //
  cheetah_mpi_poll(group);

  MPI_Comm rfi_comm = group->RFICommunicator();
  isend_list_t* outstanding_isends = comm_data_s[group->id()]->isendMap();

  //
  // Send the request.
  //
  if ( len > 0 )
    {
      //
      // If there is a buffer to send, either send it blocking or
      // nonblocking.
      //
      if ( local_bell )
	{
	  //
	  // We're doing a nonblocking send, so use isend and record
	  // the request so we can check on the isend later.
	  //
	  outstanding_isends->push_back( MPIIsendRecord() );
	  MPIIsendRecord& rec = outstanding_isends->back();
	  rec.bell = local_bell;
	  MPI_Isend(buffer,len,MPI_CHAR,context,tag,rfi_comm,&rec.request);
	}
      else
	{
	  //
	  // We're doing a blocking send, so let MPI do the blocking.
	  //
	  MPI_Send(buffer, len, MPI_CHAR, context, tag, rfi_comm);
	}
    }
  else
    {
      //
      // If there's no buffer to send, use blocking send and hit the
      // bell immediately.
      //
      MPI_Send(buffer, len, MPI_CHAR, context, tag, rfi_comm);
      if ( local_bell )
	*local_bell += 1;
    }
}


//----------------------------------------------------------------------
//
// cheetah_mpi_put
//
// Send a local buffer of a given length to a remote buffer in a given
// context.  Ring a remote doorbell when it is done.
//
// This requires two messages: one for the header and one for the
// payload. The header has the pointer to the remote buffer, the
// length of the message, the bell to ring and the message tag that
// the payload will have.
//
//----------------------------------------------------------------------

void cheetah_mpi_put(int context, void *remote, void *local, int length,
		     int *local_bell, int *remote_bell,
		     const Cheetah::MPIGroup* g)

{
  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);
  if (group->mycontext() == Cheetah::Group::NOT_A_MEMBER) return;

  //
  // Keep things flowing.
  //
  cheetah_mpi_poll(group);

  MPI_Comm put_comm = group->putCommunicator();

  int id = group->id();
  isend_list_t* outstanding_isends = comm_data_s[id]->isendMap();

  //
  // Build the header with the info the receiver needs.
  //
  MPIPutRequest msg;
  msg.remote = remote;
  msg.length = length;
  msg.bell = remote_bell;
  msg.tag = comm_data_s[id]->next_put_tag();

  //
  // Send the header. Block on this one because it is on the stack and
  // small.
  //
  MPI_Send(&msg, sizeof(msg), MPI_CHAR,context, PUT_REQUEST_TAG, put_comm);

  //
  // Send the payload.  If we've specified a local bell, make this
  // asynchronous, otherwise make it synchronous.
  //
  if ( local_bell )
    {
      //
      // Use isend and remember it so we can check on it
      // later.
      //
      outstanding_isends->push_back( MPIIsendRecord() );
      MPIIsendRecord& rec = outstanding_isends->back();
      rec.bell = local_bell;
      MPI_Isend(local,length,MPI_CHAR,context,msg.tag,put_comm,&rec.request);
    }
  else
    {
      //
      // If we're synchronous, just send the message and let MPI do
      // the blocking.
      //
      MPI_Send(local, length, MPI_CHAR, context, msg.tag, put_comm);
    }
}


//----------------------------------------------------------------------
//
// cheetah_mpi_get
//
// Get a buffer of a given length from a remote context and put it in
// a local buffer, and ring a bell when it is here.
//
// This doesn't block, it just sends the request.  Some time later a
// poll will find that it has arrived, and the bell gets rung at that
// time.
//
//----------------------------------------------------------------------

void cheetah_mpi_get(int context, void *remote, void *local, int length,
		     int *local_bell, int *remote_bell,
		     const Cheetah::MPIGroup* g)

{
  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);
  if (group->mycontext() == Cheetah::Group::NOT_A_MEMBER) return;

  //
  // Keep things moving.
  //
  cheetah_mpi_poll(group);

  MPI_Comm get_comm = group->getCommunicator();

  int id = group->id();
  get_map_t* outstanding_gets = comm_data_s[id]->getMap();

  //
  // Build the request with the info the remote processor needs.
  //
  MPIGetRequest msg;
  msg.remote = remote;
  msg.length = length;
  msg.bell = remote_bell;
  msg.tag = comm_data_s[id]->next_get_tag();

  //
  // If no local bell was specified, we need to have a backup.
  //
  int backup_bell=0;

  //
  // Build a record with the info that this context will need when the
  // reply comes back.  Store it in the container of outstanding gets.
  //
  MPIGetRecord rec;
  rec.local = local;
  rec.length = length;
  rec.bell = local_bell ? local_bell : &backup_bell;
  (*outstanding_gets)[msg.tag] = rec;

  //
  // Send the request.  Make this blocking since the thing we're
  // sending is small and on the stack.
  //
  MPI_Send(&msg, sizeof(msg), MPI_CHAR, context, GET_REQUEST_TAG, get_comm);

  //
  // If no local bell was specified, block until the reply arrives.
  //
  if ( local_bell == 0 )
    cheetah_mpi_wait(&backup_bell, 1, g);
}


//----------------------------------------------------------------------
//
// cheetah_mpi_wait
//
// Wait for a given bell to ring (get to a particular value) and
// process messages while you're waiting.
//
//----------------------------------------------------------------------

void cheetah_mpi_wait(volatile int *bell, int value, const Cheetah::MPIGroup* g)
{
  if (g && g->mycontext() == Cheetah::Group::NOT_A_MEMBER) return;

  while ( *bell < value ) {
    cheetah_mpi_poll(g);
  }
}


//----------------------------------------------------------------------
//
// mpi_barrier_handler
//
// When a barrier RFI comes in, call this function.  It increments the
// barrier counter for the proper group, then switches the pointer to
// the other counter for the next barrier.
// 
//----------------------------------------------------------------------

static void mpi_barrier_handler(int who, int, void* buffer, int)
{
  int id = *((int*) buffer);
  Cheetah::MPICommData* commData = comm_data_s[id];

  commData->incrementBarrierCounter(who);
  commData->switchBarrierCounters(who);
}


//----------------------------------------------------------------------
//
// mpi_newgroup_handler
//
// When a new communication group is created, all MPI contexts must
// create the new communicator.  So this functions must be called for
// all contexts NOT in the new group.  This handler creates the
// new MPI communicator.  NOT IMPLEMENTED YET.
// 
//----------------------------------------------------------------------

static void mpi_newgroup_handler(int who, int, void* buffer, int)
{
}


//----------------------------------------------------------------------
//
// cheetah_mpi_barrier
//
// Do a barrier among all of the contexts in the group g.
//
//----------------------------------------------------------------------

void cheetah_mpi_barrier(const Cheetah::MPIGroup* g)
{
  const Cheetah::MPIGroup* group = g ? g : controller_s->mpi_group(0);

  int nContexts = group->ncontexts();
  int myContext = group->mycontext();
  int id = group->id();

  if (myContext == Cheetah::Group::NOT_A_MEMBER) return;

  //
  // Tell all the other contexts that we arrived at this barrier.
  //
  for (int i = 0; i < nContexts; ++i)
    if (i != myContext) {
      cheetah_mpi_ainvoke(i, barrier_tag_s, &id, sizeof(id), 0, g);
    }

  //
  // Wait for all of the remote contexts to tell us that they arrived
  // at this barrier.  Spin until all the expected messages arrive, and
  // poll while spinning to keep things flowing.
  //
  Cheetah::MPICommData* commData = comm_data_s[id];
  while ( commData->barrierCounter(myContext) < nContexts-1 ) {
    cheetah_mpi_poll(g);
  }

  //
  // Reset the counter now that we're through the barrier.
  //
  commData->resetBarrierCounter(myContext);

  //
  // Use the other counter for the next barrier.
  //
  commData->switchBarrierCounters(myContext);
}
