#ifndef CHEETAH_CONTROLLER_MPI_MPI_H
#define CHEETAH_CONTROLLER_MPI_MPI_H

/*
//-----------------------------------------------------------------------------
// Overview:
//
// C interface routines for implementing Cheetah mechanism on MPI.
//
//-----------------------------------------------------------------------------
*/

/*
//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------
*/

#include <mpi.h>
#include <map>
#include <list>

//
// forward references
//

namespace Cheetah {
  class Group;
  class MPIGroup;
  class MPIController;
};

#ifdef __cplusplus
extern "C"
{
#endif

  /*
  // Set up and tear down the messaging system.  This initializes and
  // shuts down MPI.
  // 
  */
  void cheetah_mpi_setup(int*, char***, bool, Cheetah::MPIController*);
  void cheetah_mpi_finalize(bool);

  /*
  // Set up for a new communication group by creating a new MPICommData.
  */
  void cheetah_mpi_newgroup(int id, int nContexts);

  /*
  // Report on the number of contexts and which one this is.
  */
  int cheetah_mpi_ncontexts(const Cheetah::Group* g = 0);
  int cheetah_mpi_mycontext(const Cheetah::Group* g = 0);

  /*
  // Call a remote function with the given tag and buffer of data.  Do
  // not wait for a reply, but ring a bell when the buffer may be
  // reused.
  */
  void cheetah_mpi_ainvoke(int context, int tag, void *buffer, int len,
			   int *local_bell,
			   const Cheetah::MPIGroup* g = 0);

  /*
  // Copy a local buffer to a remote one in the specified context.
  // When the message arrives, ring a bell over there by incrementing
  // it.  When the buffer may be reused, ring a bell locally.
  */
  void cheetah_mpi_put(int context, void *remote, void *local, int len,
		       int *local_bell, int *remote_bell,
		       const Cheetah::MPIGroup* g = 0);

  /*
  // Copy a remote buffer to a local one.  When the data arrives, ring
  // a local bell.  When it has been successfully sent from the far
  // side, ring a bell over there.
  */
  void cheetah_mpi_get(int context, void *remote, void *local, int len,
		       int *local_bell, int *remote_bell,
		       const Cheetah::MPIGroup* g = 0);

  /*
  // Have the system pull things out of the network and handle them.
  // Keeps going as long as there are things to do.
  */
  void cheetah_mpi_poll(const Cheetah::MPIGroup* g = 0);

  /*
  // Wait for a doorbell to ring (get to a given value) while
  // processing messages.
  */
  void cheetah_mpi_wait(volatile int *bell, int value,
			const Cheetah::MPIGroup* g = 0);

  /*
  // Do a barrier among all of the contexts within a communication group.
  */
  void cheetah_mpi_barrier(const Cheetah::MPIGroup* g = 0);

#ifdef __cplusplus
}
#endif


//----------------------------------------------------------------------
//
// Structs needed for sending data across the wire.
//
//----------------------------------------------------------------------

//
// The header information for a put.  This informs the receiver that a
// put is coming in a following message.  This tells the receiver
// what the message tag is, where to put it, how long it is and what
// doorbell to ring when it arrives.
//

struct MPIPutRequest
{
  int *bell;
  void *remote;		// address on remote context to "put" data
  int length;
  int tag;
};

//
// The header information requesting a get.  This tells the receiver
// what data is being requested, how long it is, and what message tag
// to use when sending it back, and what bell to ring when the data
// has been copied.
//

struct MPIGetRequest
{
  void *remote;		// address on remote context to "get" data
  int *bell;
  int length;
  int tag;
};

//
// The record stored when a get is requested.  This has where to put
// it when it arrives, how big to exptect it to be and what doorbell
// to ring when it arrives.  These are stored in a map in which the
// key is the tag it will come in with.  Stored on requestor pending
// return of requested data.
//

struct MPIGetRecord
{
  int *bell;
  void *local;
  int length;
};

//
// The record that is stored when there is an outstanding isend.  We
// record the MPI status data and a pointer to the bell to ring when
// it finishes.
//

struct MPIIsendRecord
{
  MPI_Request request;
  int *bell;
};


//----------------------------------------------------------------------
//
// Typedefs to simplify templated type names.
//
//----------------------------------------------------------------------

typedef std::map<int,MPIGetRecord> get_map_t;
typedef get_map_t::iterator        get_iter_t;

typedef std::list<MPIIsendRecord> isend_list_t;
typedef isend_list_t::iterator    isend_iter_t;

//
// For get and put, there are control messages in addition to data
// messages. The control messages are indicated by a special tag.
//

static const int GET_REQUEST_TAG = 0;
static const int PUT_REQUEST_TAG = 0;


#endif
