//-----------------------------------------------------------------------------
// SerialController method definitions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/SerialController.h"
#include "Controller/ControllerFactory.h"
#include "Controller/Group.h"
#include "Utilities/CheetahAssert.h"
#include <stdio.h>
#include <string.h>


namespace Cheetah {

//-----------------------------------------------------------------------------
//
// SerialController::Constructor
//
//-----------------------------------------------------------------------------

SerialController::SerialController(int &, char **&)
{
  groups_m.push_back( new Group(1,0) );
}


//-----------------------------------------------------------------------------
//
// SerialController::Destructor
//
//-----------------------------------------------------------------------------

SerialController::~SerialController()
{
}


//-----------------------------------------------------------------------------
//
// SerialController::ainvoke
//
// To call a remote function, just tell the underlying library to do
// it.
//
//-----------------------------------------------------------------------------

void SerialController::ainvoke(int context, int tag, void *buf, int len,
			       int *bell)
{
  CheetahRTAssert(context == 0);
  Controller::Handler_t handler = getRegisteredHandler(tag);
  if (handler)
    {
      //
      // We can just call the handler right now, after ringing the bell
      // if necessary.
      //
      if (bell != 0)
	*bell += 1;
      handler(context, tag, buf, len);
    }
  else
    {
      //
      // We didn't find the function!  Print an error and go on.
      //
      fprintf(stderr,
	      "CHEETAH-ON-SERIAL ERROR: Requested message tag %d "
	      "has not been registered!\n", tag);
    }
}


//-----------------------------------------------------------------------------
//
// SerialController::put
// In serial, we can just copy directly.
//
//-----------------------------------------------------------------------------

void SerialController::put(int c, void *r, void *l, int len, int *lb, int *rb)
{
  // just check that the context ID is correct
  CheetahRTAssert(c == 0);

  // copy the data
  memcpy(r,l,len);

  // ring the local and remote bells immediately
  if (lb)
    *lb += 1;
  if (rb)
    *rb += 1;
}


//-----------------------------------------------------------------------------
//
// SerialController::get
// In serial, we can just copy directly.
//
//-----------------------------------------------------------------------------

void SerialController::get(int c, void *r, void *l, int len, int *lb, int *rb)
{
  // just check that the context ID is correct
  CheetahRTAssert(c == 0);

  // Copy the data
  memcpy(l,r,len);

  // ring the local and remote bells immediately
  if (lb)
    *lb += 1;
  if (rb)
    *rb += 1;
}


//-----------------------------------------------------------------------------
//
// SerialController::poll
//
// To handle incoming messages, just tell the underlying library to do
// it.
// Since this is serial, we should not ever have to do anything in a poll.
//
//-----------------------------------------------------------------------------

void SerialController::poll()
{
}


//-----------------------------------------------------------------------------
//
// SerialController::wait
//
// To wait for a bell to reach a given value, just tell the underlying
// library to do it.
// In serial, if the bell is not already set, it is an error.
//
//-----------------------------------------------------------------------------

void SerialController::wait(volatile int *bell, int value)
{
  CheetahRTAssert(*bell == value);
}


//-----------------------------------------------------------------------------
//
// SerialController::barrier
//
// To do a barrier, just tell the underlying library to do it.
// In serial, there is only one context, so just return.
// 
//-----------------------------------------------------------------------------

void SerialController::barrier()
{
}

void SerialController::barrier(const Group& group)
{
}

//-----------------------------------------------------------------------------
//
// The function used to create new implementations.  This is registered
// with a factory.
//
//-----------------------------------------------------------------------------

static ControllerImpl *serialControllerFactory(int &argc, char ** &argv)
{
  return new SerialController(argc, argv);
}


//-----------------------------------------------------------------------------
//
// SerialController::registerFactory
//
// Register this subclass with the controller factory list.
//
//-----------------------------------------------------------------------------

void SerialController::registerFactory()
{
  ControllerFactory::registerFactory("-serial", serialControllerFactory);
}


} // namespace Cheetah
