//-----------------------------------------------------------------------------
//
// Shared memory Cheetah function definitions.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/Shmem/cheetah_shmem.h"
#include "Controller/Shmem/MM_Allocator.h"
#include "Controller/Controller.h"
#include "Utilities/CheetahAssert.h"
#include "Utilities/CheetahSpinlockMutex.h"
#include <map>
#include <list>
#include <vector>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <signal.h>
#include <mm.h>


////////////////////////////////////////////////////////////////////////
// 
// Some gadgets for allocating objects on cache lines.
//
////////////////////////////////////////////////////////////////////////

//
// The default size of a cache line in bytes.
//
#ifndef CHEETAH_SHMEM_CACHE_LINE_SIZE
#define CHEETAH_SHMEM_CACHE_LINE_SIZE (128)
#endif

//
// An object with size rounded up to an integral number of cache lines.
//
template<class T>
struct CacheLine
{
  enum
  {
    SIZE = (((sizeof(T)-1)/CHEETAH_SHMEM_CACHE_LINE_SIZE)+1)
	    *CHEETAH_SHMEM_CACHE_LINE_SIZE
  };
  char cache_line[SIZE];
};

////////////////////////////////////////////////////////////////////////
//
// Type definitions.
//
// AInvokeRequest:     The data we need to transfer for an ainvoke.
// GetRequest:         The data we need to transfer for an get.
// PutRequest:         The data we need to transfer for an put.
// IncomingData:       A union for the above three structs.
// IncomingQueueEntry: A tag for the type of request and the data for it.
// BarrierData:        The data we need for a barrier.
// incoming_queue_t:   The containter type for incoming requests.
// handler_map_t:      The container type for mapping tags to handlers.
//
////////////////////////////////////////////////////////////////////////

//
// AInvokeRequest.  The data we need to transfer for an ainvoke.  We
// have the data for the arguments, the tag for the function to call,
// and the length of the buffer for the arguments.
//

struct AInvokeRequest
{
  void *data;
  int function_tag;
  int length;
  int calling_context;
};


//
// GetRequest. The data we need to request a get.  We need to send the
// buffer the data will come from including its length, the bell to
// ring on the far side.  We also send the requesting context together
// with the buffer to put the reply and the bell to ring when that has
// been done so that we don't need to store any information on the
// requesting side about the get.  The reply comes back as a put.
//

struct GetRequest
{
  void *source_data;
  void *dest_data;
  int *source_bell;
  int *dest_bell;
  int length;
  int requesting_context;
};


//
// PutRequest.  The data we need to request a put.  All we need here
// is the buffer to put the data in and the bell to ring when we've
// done that.
//

struct PutRequest
{
  void *dest_data;
  void *shared_data;
  int *dest_bell;
  int length;
};


//
// IncomingData.  We want to be able to put any of the requests in a
// given list, so we need a union that could hold any of the types of
// requests.
//

union IncomingData
{
  AInvokeRequest ainvoke_request;
  GetRequest get_request;
  PutRequest put_request;
};


//
// The actual entry in the queue includes the type of the request and
// then the union with the data for the request.
//

struct IncomingQueueEntry
{
  enum { AINVOKE_REQUEST=1, GET_REQUEST, PUT_REQUEST };
  int request_type;
  IncomingData data;
};


//
// The data we need to hold for a barrier.  We hold the current number
// of procs waiting at the barrier, the number we're expecting and a
// mutex so we can control access.
//

struct BarrierData
{
  int goal;
  int current;
  Cheetah::SpinlockMutex mutex;
  BarrierData() : goal(0), current(0) {}
};


//
// incoming_queue_t. The type of container for holding the incoming
// requests from other procs.
//

typedef
  std::vector<IncomingQueueEntry,MM_Allocator<IncomingQueueEntry> >
  incoming_queue_t;
typedef incoming_queue_t::iterator incoming_iter_t;

////////////////////////////////////////////////////////////////////////
//
// Static data.
//
// nprocs_s                 The number of distinct contexts.
// myproc_s                 The current context.
// shared_mem_size_s        The size of the shared memory region.
// incoming_queues_s        A place for other procs to drop off msgs.
// waiting_messages_s       Flags for each proc to spin on waiting for msgs.
// local_incoming_s         A queue in local mem for the incoming msgs.
// barrier_data_s           The data in shared memory for a barrier.
// children_s               The pids of the children
// finalized_s              Whether a child has exited gracefully
//
////////////////////////////////////////////////////////////////////////

static int nprocs_s = 1;
static int myproc_s = 0;
static const int shared_mem_size_s = (1<<30);
static CacheLine<incoming_queue_t> *incoming_queues_s = 0;
static CacheLine<int> *waiting_messages_s = 0;
static BarrierData *barrier_data_s = 0;
static pid_t *children_s = 0;
static bool *finalized_s = 0;
static Cheetah::SpinlockMutex bell_mutex;

//
// Since incoming_queues_s and waiting_messages_s are really just
// blocks of cache lines, we use helper functions to look up elements
// and set the types correctly.
//

static inline incoming_queue_t&
incoming_queue(int context)
{
  return *((incoming_queue_t*)(incoming_queues_s+context));
}

static inline Cheetah::SpinlockMutex&
incoming_mutex(int context)
{
  return *((Cheetah::SpinlockMutex*)((char*)(incoming_queues_s+context)+
			    sizeof(incoming_queue_t)));
}

static inline int&
waiting_message(int context)
{
  return *((int*)(waiting_messages_s+context));
}


////////////////////////////////////////////////////////////////////////
//
// Static Functions
//
// print_usage_and_exit
// parse_args
// fork_children
// wait_for_children_to_exit
// setup_shared_memory
// teardown_shared_memory
//
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
//
// If there was an error in the input, tell them how to run this and
// get out of here.
//
//----------------------------------------------------------------------

static void
print_usage_and_exit()
{
  fprintf(stderr,"cheetah_shmem usage: executable -np <CPUCOUNT> [app args]\n");
  exit(1);
}


//----------------------------------------------------------------------
//
// parse_args
//
// Parse the input arguments.  The only thing currently recognized is
// -np <NPROCS>.  If that is found, the others are shifted down so
// that the user application doesn't see it.
//
//----------------------------------------------------------------------

static void
parse_args(int *argc, char ***argv)
{
  //
  // Make sure we have some args to parse.
  //
  if ( (argc==0) || (argv==0) || (*argc<=1) )
    print_usage_and_exit();

  //
  // Loop over the arguments.
  //
  int i=1;
  while (i<*argc)
    {
      //
      // Is this the one we're looking for?
      //
      if ( strcmp( (*argv)[i] , "-np" )==0 )
	{
	  //
	  // If this was the last one, there's no integer after it,
	  // which is an error.
	  //
	  if ( i == (*argc)-1 )
	    print_usage_and_exit();

	  //
	  // Grab the next arg as an integer, and print an error if it
	  // doesn't make sense.
	  //
	  nprocs_s = atoi((*argv)[i+1]);
	  if ( nprocs_s <= 0 )
	    print_usage_and_exit();

	  //
	  // Shift the other args down.
	  //
	  for (int j=i+2; j<*argc; ++j)
	    (*argv)[j-2] = (*argv)[j];

	  //
	  // Reduce the total number of args and get out of here.
	  //
	  *argc -= 2;
	  return;
	}
      else
	{
	  //
	  // This wasn't the one we're looking for.  Go on to the next.
	  //
	  ++i;
	}
    }

  //
  // If we get to here there was no -np in the args.  Print an error
  // and exit.
  //
  print_usage_and_exit();
}


//----------------------------------------------------------------------
//
// kill_and_exit
//
// Violently kill children, self.
//
//----------------------------------------------------------------------

static void
kill_and_exit()
{
  for (int i=0; i<nprocs_s; ++i)
    if ( children_s[i] != -1 )
      kill(children_s[i],SIGKILL);
  exit(1);
}


//----------------------------------------------------------------------
//
// fork_children
//
// Fork off nprocs_s child processes.  Return true in the parent,
// false in the children.  Set the myproc_s variable for each child.
//
//----------------------------------------------------------------------

static bool
fork_children()
{
  //
  // Initialize an array with one entry per child.  This array will
  // have the pid of each child together with a bool for whether that
  // child exited in a controlled manner.  We use this to detect a
  // child dying unexpectedly.
  //
  children_s = (pid_t*)MM_malloc(sizeof(pid_t)*nprocs_s);
  finalized_s = (bool*)MM_malloc(sizeof(bool)*nprocs_s);

  int i;
  for (i=0; i<nprocs_s; ++i)
    {
      children_s[i] = -1;
      finalized_s[i] = false;
    }

  //
  // Loop over all of the children.
  //
  for (i=0; i<nprocs_s; ++i)
    {
      //
      // Fork the i'th child.
      //
      pid_t child = fork();
      if ( child == 0 )
	{
	  //
	  // Write down which child this is and return that this is
	  // not the parent.
	  //
	  myproc_s = i;
	  return false;
	}
      else if ( child == -1 )
	{
	  //
	  // Some sort of error.  Scream and die.
	  //
	  fprintf(stderr,"Could not fork child %d!\n",i);
	  kill_and_exit();
	}
      //
      // We're in the parent, so save the pid of the child.
      //
      children_s[i] = child;
    }

  //
  // Set myproc_s in the parent, in case we need to check.
  //
  myproc_s = -1;
  return true;
}

//----------------------------------------------------------------------
//
// wait_for_children_to_exit
//
// After the parent spawns all of the children, wait for them to die.
//
//----------------------------------------------------------------------

static void
wait_for_children_to_exit()
{
  int status;
  for (int i=0; i<nprocs_s; ++i)
    {
      pid_t pid = wait(&status);

      //
      // Look for this pid in the children.
      //
      for (int j=0; j<nprocs_s; ++j)
	if ( (children_s[j]==pid) && !finalized_s[j] )
	  {
	    //
	    // This child died unexpectedly!  Kill the rest and exit.
	    //
	    fprintf(stderr,
		    "%d: Child died unexpectedly! "
		    "Everybody out of the pool!\n",
		    cheetah_shmem_mycontext());	
	    kill_and_exit();
	  }
    }
}


//----------------------------------------------------------------------
//
// setup_shared_memory
//
// Initialize the shared memory.  This must be called before the fork.
// Creates the shared memory region, sets up the malloc system, and
// builds the data structures in shared memory that we need.
//
//----------------------------------------------------------------------

static void
setup_shared_memory()
{
  //
  // Create the shared memory region and initialize the malloc system
  // that uses it.
  //
  MM_create(shared_mem_size_s,NULL);

  //
  // Allocate shared memory space for the incoming message queues.
  // Call placement new for each to initialize them since we're doing
  // this damn cache line thing.
  //
  incoming_queues_s = (CacheLine<incoming_queue_t>*)
    MM_malloc(sizeof(CacheLine<incoming_queue_t>)*nprocs_s);

  int i;
  for (i=0; i<nprocs_s; ++i)
    {
      new (&incoming_queue(i)) incoming_queue_t;
      incoming_queue(i).reserve(20);
    }

  //
  // Allocate shared memory space for the flag for a proc to spin on
  // waiting for a message to arrive.  Initialize each one to true.
  //
  waiting_messages_s = (CacheLine<int>*)
    MM_malloc(sizeof(CacheLine<int>)*nprocs_s);
  for (i=0; i<nprocs_s; ++i)
    waiting_message(i) = false;

  //
  // Allocate data for the barrier data structure and initialize it.
  //
  barrier_data_s = (BarrierData*)
    MM_malloc(sizeof(BarrierData));
  new ((void*)barrier_data_s) BarrierData;
}


//----------------------------------------------------------------------
//
// teardown_shared_memory
//
// When exiting, clean up the shared memory region.
//
//----------------------------------------------------------------------

static void
teardown_shared_memory()
{
  MM_destroy();

  incoming_queues_s = 0;

  //  incoming_pqueues_s = 0;

  waiting_messages_s = 0;
  barrier_data_s = 0;
  children_s = 0;
  finalized_s = 0;
}


//----------------------------------------------------------------------

static void
handle_incoming(IncomingQueueEntry& incoming)
{
  switch (incoming.request_type)
    {
    case IncomingQueueEntry::AINVOKE_REQUEST:
      {
	AInvokeRequest& request = incoming.data.ainvoke_request;

	//
	// Given the function tag in the incoming message, find
	// the function pointer that we will be calling.
	//
	int tag = request.function_tag;
	Cheetah::Controller::Handler_t handler = 
          Cheetah::Controller::getRegisteredHandler(tag);

	//
	// Call the indicated function.
	//
	if ( handler )
	  {
	    handler(request.calling_context, request.function_tag,
		    request.data, request.length);
	  }
	else
	  {
	    //
	    // We didn't find the function!  Print an error and go on.
	    //
	    fprintf(stderr,
		    "CHEETAH-ON-SHMEM ERROR: Incoming message tag %d "
		    "has not been registered!\n",request.function_tag);
	  }

	//
	// Free the buffer in shared memory.
	//
	if ( request.data )
	  MM_free(request.data);

	break;
      }
    case IncomingQueueEntry::GET_REQUEST:
      {
	GetRequest &request = incoming.data.get_request;
	//
	// Handle a get by doing a put back to the requester.
	//
	cheetah_shmem_put(request.requesting_context,
			request.dest_data,
			request.source_data,
			request.length,
			request.dest_bell,
			request.source_bell);

	break;
      }
    case IncomingQueueEntry::PUT_REQUEST:
      {
	PutRequest &request = incoming.data.put_request;
	//
	// Handle a put by copying the data to the correct local place.
	//
	if ( request.length > 0 )
	  {
	    memcpy(request.dest_data, request.shared_data, request.length );
	  }

	//
	// Ring the bell to indicate that the data is here.
	//
	if ( request.dest_bell )
	  {
	    bell_mutex.lock();
	    *request.dest_bell += 1;
	    bell_mutex.unlock();
	  }

	//
	// If there was a shared buffer, delete it.
	//
	if ( request.shared_data )
	  MM_free(request.shared_data);

	break;
      }
    default:
      fprintf(stderr,"%d: got unrecognized message type %d!\n",
	      myproc_s,incoming.request_type);
      break;
    }
}


////////////////////////////////////////////////////////////////////////
//
// Externally visible functions.
//
// cheetah_shmem_setup
// cheetah_shmem_finalize
// cheetah_shmem_ncontexts
// cheetah_shmem_mycontext
// cheetah_shmem_register
// cheetah_shmem_ainvoke
// cheetah_shmem_put
// cheetah_shmem_get
// cheetah_shmem_poll
// cheetah_shmem_wait
// cheetah_shmem_barrier
//
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
//
// cheetah_shmem_setup
//
// Initialize the cheetah_shmem system.  The input is argc and argv from
// the command line.  The tasks are:
// 1. Parse the arguments for nprocs.
// 2. Initialize the shared memory region.
// 3. Fork the children.  The children then return.
// 4. The parent then waits for the children to exit.
// 5. The parent tears down the shared memory and exits itself.
//
//----------------------------------------------------------------------

void
cheetah_shmem_setup(int *argc, char ***argv)
{
  parse_args(argc,argv);
  setup_shared_memory();
  if ( fork_children() == 0 )
    return;
  wait_for_children_to_exit();
  teardown_shared_memory();
  exit(0);
}


//----------------------------------------------------------------------
//
// cheetah_shmem_ncontexts
// cheetah_shmem_mycontext
//
// Return the number of contexts and the current context.
//
//----------------------------------------------------------------------

int cheetah_shmem_ncontexts()
{
  return nprocs_s;
}

int cheetah_shmem_mycontext()
{
  return myproc_s;
}


//----------------------------------------------------------------------
//
// cheetah_shmem_finalize
//
// Called by a child when it is ready to exit, this waits for all the
// other children to call finalize and then exits.
//
//----------------------------------------------------------------------

void
cheetah_shmem_finalize()
{
  cheetah_shmem_barrier();
  finalized_s[myproc_s] = true;
}


//----------------------------------------------------------------------
//
// cheetah_shmem_poll
//
// Check the incoming message queues and handle anything that is
// there.
//
//----------------------------------------------------------------------

void
cheetah_shmem_poll()
{
  //
  // Check and see if there are any incoming messages.
  // 
  //  static bool first = true;
  static incoming_queue_t local_incoming_s(20);

  if ( waiting_message(myproc_s) )
    { 
       local_incoming_s.clear();

      // To minimize the size of the critical section when handling
      // incoming messages, we first transfer them to a local queue.

      incoming_mutex(myproc_s).lock();
      local_incoming_s.swap( incoming_queue(myproc_s) ); 
      waiting_message(myproc_s) = false;
      incoming_mutex(myproc_s).unlock();

      //
      // Loop over the incoming messages, handling them in turn.
      //

      for (int pos = local_incoming_s.size()-1;pos >=0; --pos)
	handle_incoming(local_incoming_s[pos]);
      local_incoming_s.clear();
     }
}


//----------------------------------------------------------------------
//
// cheetah_shmem_barrier
//
// A normal barrier among all of the contexts.
//
//----------------------------------------------------------------------

void
cheetah_shmem_barrier()
{
  //
  // Keep a copy of the pointer to the barrier data locally, so we can
  // keep better track of the memory references.
  //
  register BarrierData *bd = barrier_data_s;

  //
  // Lock the mutex so we can do these operations without others
  // stepping on us.
  //
  bd->mutex.lock();

  //
  // Get copies of the current goal and the current number of procs
  // here.  Explicitly say these are register variable to help keep
  // track of the memory references.
  //
  register int goal = bd->goal;
  register int current = bd->current;

  //
  // If the current counter is the same as the goal, then we're the
  // first ones here.  We'll need to increment the goal for the rest
  // of the procs when they arrive.
  if ( current == goal )
    {
      goal += nprocs_s;
      bd->goal = goal;
    }

  //
  // Record that one more proc has arrived at the barrier, and unlock
  // the mutex to let others in.
  //
  bd->current = current + 1;
  bd->mutex.unlock();

  //
  // Wait for everyone else to get here, and keep things moving while
  // we wait.
  //
  while ( bd->current < goal )
    cheetah_shmem_poll();
}


//----------------------------------------------------------------------
//
// cheetah_shmem_ainvoke
//
// Request that another context execute a function with a given tag
// using the provided buffer with a given length.  Ring a local bell
// when the buffer can be reused.
//
//----------------------------------------------------------------------

void
cheetah_shmem_ainvoke(int context, int tag, void *buffer, int len,
		    int *local_bell)
{
  //
  // If the length is nonzero, the buffer pointer must be nonzero.
  // The context has to be in [0..nprocs).
  //
  CheetahRTAssert( !((len!=0)&&(buffer==0)) );
  CheetahRTAssert( (context>=0) && (context<nprocs_s) );

  //
  // If we're sending any data, allocate a shared buffer and copy the
  // buffer into it.
  //
  void *shared_buffer=0;
  if ( len>0 )
    {
      shared_buffer = MM_malloc(len);
      memcpy(shared_buffer,buffer,len);
    }

  //
  // Strobe the local bell so that this context can reuse the buffer.
  //
  if ( local_bell )
    {
      bell_mutex.lock();
      *local_bell += 1;
      bell_mutex.unlock();
    }

  //
  // Construct the entry for the queue with the shared memory data.
  // 
  IncomingQueueEntry entry;
  entry.request_type = IncomingQueueEntry::AINVOKE_REQUEST;
  entry.data.ainvoke_request.data = shared_buffer;
  entry.data.ainvoke_request.function_tag = tag;
  entry.data.ainvoke_request.length = len;
  entry.data.ainvoke_request.calling_context = myproc_s;
  //
  // Put an entry in the appropriate queue. Strobe the variable that
  // indicates there is a message waiting.  Protect this operation
  // with a mutex.
  //

  incoming_mutex(context).lock();   
  incoming_queue(context).push_back(entry);
  waiting_message(context) = true;
  incoming_mutex(context).unlock();

}


//----------------------------------------------------------------------
//
// cheetah_shmem_get
//
//----------------------------------------------------------------------

void
cheetah_shmem_get(int context, void *remote, void *local, int length,
		int *local_bell, int *remote_bell)
{
  //
  // If the length is nonzero, the buffer pointers must be nonzero.
  // The context has to be in [0..nprocs).
  //
  CheetahRTAssert( !((length!=0)&&((local==0)||(remote==0)) ) );
  CheetahRTAssert( (context>=0) && (context<nprocs_s) );

  //
  // If no local bell was specified (by inputing local_bell==0), we'll
  // make a bell here and wait until it gets rung before exiting.
  //
  int backup_bell = 0;

  //
  // Build the instruction to give to the remote processor.  This
  // includes enough information for the remote processor to do a put
  // to the correct location in this context.
  //
  IncomingQueueEntry entry;
  entry.request_type = IncomingQueueEntry::GET_REQUEST; 
  entry.data.get_request.source_data = remote;
  entry.data.get_request.dest_data   = local;
  entry.data.get_request.source_bell = remote_bell;
  entry.data.get_request.dest_bell   =(local_bell ? local_bell : &backup_bell);
  entry.data.get_request.length      = length;
  entry.data.get_request.requesting_context = myproc_s;

  //
  // Put an entry in the appropriate queue. Strobe the variable that
  // indicates there is a message waiting.  Protect this operation
  // with a mutex.
  //

  incoming_mutex(context).lock();
  incoming_queue(context).push_back(entry);
  waiting_message(context) = true;
  incoming_mutex(context).unlock();

  // If there was no local bell send in, wait until the backup bell
  // gets rung.
  //
  if ( local_bell == 0 )
    cheetah_shmem_wait(&backup_bell,1);
}


//----------------------------------------------------------------------
//
// cheetah_shmem_put
//
//----------------------------------------------------------------------

void
cheetah_shmem_put(int context, void *remote, void *local, int length,
		int *local_bell, int *remote_bell)
{
  //
  // If the length is nonzero, the buffer pointers must be nonzero.
  // The context has to be in [0..nprocs).
  //
  CheetahRTAssert( !((length!=0)&&((local==0)||(remote==0)) ) );
  CheetahRTAssert( (context>=0) && (context<nprocs_s) );

  //
  // If we're sending any data, allocate a shared buffer and copy the
  // source data into it.
  //
  void *shared_buffer=0;
  if ( length>0 )
    {
      shared_buffer = MM_malloc(length);
      memcpy(shared_buffer,local,length);
    }

  //
  // Strobe the local bell so that this context can reuse the buffer.
  //
  if ( local_bell )
    {
      bell_mutex.lock();
      *local_bell += 1;
      bell_mutex.unlock();
    }

  //
  // Build the instruction to give to the remote processor.  This
  // includes enough information for the remote processor to do a put
  // to the correct location in this context.
  //
  IncomingQueueEntry entry;
  entry.request_type = IncomingQueueEntry::PUT_REQUEST;
  entry.data.put_request.dest_data   = remote;
  entry.data.put_request.shared_data = shared_buffer;
  entry.data.put_request.dest_bell   = local_bell;
  entry.data.put_request.length      = length;

  //
  // Put an entry in the appropriate queue. Strobe the variable that
  // indicates there is a message waiting.  Protect this operation
  // with a mutex.
  //
  incoming_mutex(context).lock();
  incoming_queue(context).push_back(entry);
  waiting_message(context) = true;
  incoming_mutex(context).unlock();
}


//----------------------------------------------------------------------
//
// cheetah_shmem_wait
//
// Wait for a bell to get up to a particular value, and keep calling
// poll to make sure things are flowing through the system.
//
//----------------------------------------------------------------------

void
cheetah_shmem_wait(volatile int *bell, int value)
{
  if ( bell )
    while (*bell < value)
      cheetah_shmem_poll();
}

// ACL:rcsinfo
// ----------------------------------------------------------------------
// $RCSfile: cheetah_shmem.cpp,v $   $Author: rasmussn $
// $Revision: 1.3 $   $Date: 2000/05/02 22:59:38 $
// ----------------------------------------------------------------------
// ACL:rcsinfo
