//-----------------------------------------------------------------------------
//
// ULM Cheetah function definitions.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/ULM/cheetah_ulm.h"
#include "Controller/Controller.h"
#include <map>
#include <list>
#include <stdio.h>
#include "ULM.h"


#define ULM_DEBUG_MSG(x)
//#define ULM_DEBUG_MSG(x) x


//----------------------------------------------------------------------
//
// Structs we'll be sending across the wire.
//
//----------------------------------------------------------------------

//
// The header information for a put.  This informs the receiver that a
// put is coming in a following message.  This tells the receiver
// what the message tag is, where to put it, how long it is and what
// doorbell to ring when it arrives.
//

struct ULMPutRequest
{
  int *bell;
  void *remote;
  int length;
};

//
// The header information requesting a get.  This tells the receiver
// what data is being requested, how long it is, and what message tag
// to use when sending it back, and what bell to ring when the data
// has been copied.
//

struct ULMGetRequest
{
  void *remote;
  int *bell;
  int length;
};

//
// The record stored when a get is requested.  This has where to put
// it when it arrives, how big to exptect it to be and what doorbell
// to ring when it arrives.  These are stored in a map in which the
// key is the tag it will come in with.
//

struct ULMGetRecord
{
  int *bell;
  void *local;
  int length;
};

//
// The record that is stored when there is an outstanding isend.  We
// record the ULM status data and a pointer to the bell to ring when
// it finishes.
//

struct ULMIsendRecord
{
  TPMsg request;
  int *bell;
};


//----------------------------------------------------------------------
//
// Static data structures for keeping track of the messages.
//
//----------------------------------------------------------------------

//
// Cache the number of contexts and the current context.
//

static int ncontexts_s;
static int mycontext_s;

//
// ulm_outstanding_gets_s
// Holds the outstanding get operations we're expecting to hear back on.
//

typedef std::map<int,ULMGetRecord> ulm_get_map_t;
typedef ulm_get_map_t::iterator    ulm_get_iter_t;
static ulm_get_map_t               ulm_outstanding_gets_s;

//
// ulm_outstanding_isends_s 
// Holds a list of outstanding isends we're expecting to finish someday.
//

typedef std::list<ULMIsendRecord>  ulm_isend_list_t;
typedef ulm_isend_list_t::iterator ulm_isend_iter_t;
static ulm_isend_list_t            ulm_outstanding_isends_s;

//
// Three opcodes for keeping the messages straight.
// rfi_comm_s: all messages about remote function invocation.
// get_comm_s: all get messages, including requests and replies.
// put_comm_s: all put messages.
//

static const int rfi_comm_s         = 0;
static const int get_request_comm_s = 1;
static const int get_respond_comm_s = 2;
static const int put_header_comm_s  = 3;
static const int put_payload_comm_s = 4;
static const int number_comm_s      = 8;

//
// For get and put, there are control messages in addition to data
// messages. The control messages are indicated by a special tag.  The
// data messages are all other messages, and we generate a unique one
// for each transaction.
//

static int ulm_next_put_tag_s  = 1;
static int ulm_next_get_tag_s  = 1;
static const int ulm_max_tag_s = 65536;

//
// We have a special function for handling barriers since barriers
// need to be set up as part of initialization.  We also have
// static data structures for barriers.  We have two counters that
// will be incremented in barriers.  There are two so we can switch
// back and forth between them to eliminate race conditions between
// multiple barriers.  next_barrier_s indicates the current active
// barrier for each remote proc.
//

static int ulm_barrier_tag_s = -1;
static int ulm_barrier_counters_s[2];
static int *ulm_next_barrier_s = 0;


//----------------------------------------------------------------------
//
// ulm_barrier_handler
//
// When a barrier RFI comes in, call this function.  It increments the
// barrier counter, switches the pointer to the other counter for the
// next barrier.
// 
//----------------------------------------------------------------------

static void
ulm_barrier_handler(int who, int, void *, int)
{
  ulm_barrier_counters_s[ulm_next_barrier_s[who]] += 1;
  ulm_next_barrier_s[who] = 1 - ulm_next_barrier_s[who];
}

//----------------------------------------------------------------------
//
// ulm_allocate_buffer(long length)
//
// allocate storage for a new buffer of the given length
//
//----------------------------------------------------------------------

static char ulm_dummy_item = 0;

inline char *
ulm_allocate_buffer(long length)
{
  return (length > 0 ? new char[length] : &ulm_dummy_item);
}

//----------------------------------------------------------------------
//
// ulm_free_buffer(char *buffer)
//
// free storage previously allocated for a new buffer
//
//----------------------------------------------------------------------

inline void
ulm_free_buffer(char *buffer)
{
  if (buffer != &ulm_dummy_item)
    delete [] buffer;
}

//----------------------------------------------------------------------
//
// ulm_setup_tag(int usertag, int opcode)
//
// Combine the user-provided tag and the opcode for the operation (get,
// put, ainvoke, etc) into a single ULM tag used for send and receive
// operations.  The two values are encoded as:
//
//   ulmtag = number_comm_s*usertag + opcode;
//
// so that you can recover the original usertag and opcode by:
//
//   usertag = ulmtag / number_comm_s;
//   opcode  = ulmtag % number_comm_s;
//
// Thus, the opcodes are limited to values from 0 ... number_comm_s - 1.
// These could all also be expressed as bit-shift operations if
// performance is an issue for this.
//
//----------------------------------------------------------------------

inline int
ulm_setup_tag(int usertag, int opcode)
{
  return (usertag * number_comm_s + opcode);
}

//----------------------------------------------------------------------
//
// ulm_tag_to_user_tag(int tag)
//
// Convert the tag used to send messages via ULM to the one the user used.
// 
//----------------------------------------------------------------------

inline int
ulm_tag_to_user_tag(int ulmtag)
{
  return (ulmtag / number_comm_s);
}

//----------------------------------------------------------------------
//
// ulm_tag_to_opcode(int tag)
//
// Convert the tag used to send messages via ULM to the opcode
// embedded in the ULM tag.
// 
//----------------------------------------------------------------------

inline int
ulm_tag_to_opcode(int ulmtag)
{
  return (ulmtag % number_comm_s);
}

//----------------------------------------------------------------------
//
// ulm_next_put_tag 
//
// Generate the next unique tag for a put.  If we wrap around the tag
// space, start at the beginning but be sure to skip the special
// PUT_REQUEST_TAG!
//
//----------------------------------------------------------------------

static int
ulm_next_put_tag()
{
  ulm_next_put_tag_s += 1;
  if (ulm_next_put_tag_s == ulm_max_tag_s)
    ulm_next_put_tag_s = 1;
  return ulm_next_put_tag_s;
}

//----------------------------------------------------------------------
//
// ulm_next_get_tag 
//
// Generate the next unique tag for a get.
//
//----------------------------------------------------------------------

static int
ulm_next_get_tag()
{
  ulm_next_get_tag_s += 1;
  if (ulm_next_get_tag_s == ulm_max_tag_s)
    ulm_next_get_tag_s = 1;
  return ulm_next_get_tag_s;
}

//----------------------------------------------------------------------
//
// ulm_invoke_handler
//
// For the given arguments, find what handler has been registered
// to be called, and call it.  If no handler matches the tag,
// it is an error.
// The 'tag' here is the user-supplied tag.  It is NOT the tag
// generated for sending and receiving message via ulm calls; that tag
// encodes the user-supplied tag and an opcode.  This routine is
// intended to be called after you decode the ulm tag, or before you
// encode it.
//
//----------------------------------------------------------------------

static void
ulm_invoke_handler(int who, int tag, void *buf, int len)
{
  Cheetah::Controller::Handler_t handler = 
    Cheetah::Controller::getRegisteredHandler(tag);

  if (handler)
    {
      ULM_DEBUG_MSG(cerr << "RFI handler found, invoking ..." << endl;)
      handler(who, tag, buf, len);
    }
  else
    {
      //
      // We didn't find the function!  Print an error and go on.
      //
      fprintf(stderr,
	      "CHEETAH-ON-ULM ERROR: Incoming message tag %d "
	      "has not been registered!\n", tag);
    }
}

//----------------------------------------------------------------------
//
// ulm_blocking_recv
//
// Call the ULM routines to receive a message, and block until
// it arrives.  This should really only be called after you have
// determined that there is indeed a message of the proper type
// ready to be received.  If an error occurs, print out a message.
// The 'tag' here is the internal ULM-ON-CHEETAH tag, not the
// user-supplied tag.  This tag encodes the user-supplied tag and
// the code indicating what type of message it is (rfi, put, get, etc).
// The 'checklen' is a length that we expect len to be.  If they
// are not the same, print an error message.
// Return success.
//
//----------------------------------------------------------------------

static bool
ulm_blocking_recv(int who, int tag, void *buf, int len, int checklen)
{
  // Make sure the lengths match up

  if (len != checklen)
    {
      fprintf(stderr,
	      "CHEETAH-ON-ULM ERROR: Expected receive buffer length %d",
	      checklen);
      fprintf(stderr,
	      " is different than available message length %d",
	      len);
      fprintf(stderr,
	      " from node %d in ulm_blocking_recv on node %d",
	      who, mycontext_s);
      return false;
    }

  // Call the async receive, which fills returnhandle with info
  // on the operation.

  TPMsg returnhandle;
  tp_irecv(tag, who, buf, len, &returnhandle);

  // Wait for the receive to complete, also check for an error.
  // We might want to make this a loop where we check for completed
  // isends as well if the received message has not yet arrived.

  if (tp_wait(returnhandle, 0) < 0)
    {
      fprintf(stderr,
	      "CHEETAH-ON-ULM ERROR: Error from tp_wait while waiting");
      fprintf(stderr,
	      " on node %d for a message from node %d with tag %d and",
	      mycontext_s, who, tag);
      fprintf(stderr,
	      " length %d\n",
	      len);
      return false;
    }

  return true;
}

//----------------------------------------------------------------------
//
// ulm_rfi_receive
//
// Handle an incoming RFI request.  This should be called after you
// have determined that a RFI message is available.
//
//----------------------------------------------------------------------

static void
ulm_rfi_receive(MessageStruct_t *status)
{
  // Convert the sending tag to the one the user used

  int usertag = ulm_tag_to_user_tag(status->tag);

  // Allocate a buffer for it to land in.

  ULM_DEBUG_MSG(cerr << "Allocating buffer for rfi_receive of length = ";)
  ULM_DEBUG_MSG(cerr << status->MessageLength << endl;)
  char *buffer = ulm_allocate_buffer(status->MessageLength);

  // Receive the buffer and block until it is done.  We know it
  // has arrived, so this should be quick.

  ulm_blocking_recv(status->SourceProc, status->tag, (void *)buffer,
		    status->MessageLength, status->MessageLength);

  //
  // Find the registered function for this message and call it.
  //

  ulm_invoke_handler(status->SourceProc,usertag,buffer,status->MessageLength);

  //
  // Clean up the buffer.
  //

  ULM_DEBUG_MSG(cerr << "RFI handler complete, deleting buffer." << endl;)
  ulm_free_buffer(buffer);
}

//----------------------------------------------------------------------
//
// ulm_get_request_receive
//
// Process get requests.  If you get a request, send a
// response back.  This should be called after you
// have determined that a get message is available.
//
//----------------------------------------------------------------------

static void
ulm_get_request_receive(MessageStruct_t *status)
{
  // Convert the sending tag to the one the user used

  int usertag = ulm_tag_to_user_tag(status->tag);

  // Receive the buffer and block until it is done.  We know it
  // has arrived, so this should be quick.  The data goes into
  // a ULMGetRequest struct.

  ULMGetRequest msg;
  ulm_blocking_recv(status->SourceProc, status->tag, (void *)(&msg),
		    status->MessageLength, sizeof(ULMGetRequest));

  //
  // Send back the response. We'll need to record the isend so
  // we can check on it later.  The 'msg' struct contains the
  // information about where to get the data, how many bytes are in
  // the data, and what tag to use.
  //

  ulm_outstanding_isends_s.push_back(ULMIsendRecord());
  ULMIsendRecord& rec = ulm_outstanding_isends_s.back();
  rec.bell = msg.bell;
  int sendtag = ulm_setup_tag(usertag, get_respond_comm_s);
  tp_isend(sendtag, status->SourceProc, (void *)(msg.remote), msg.length,
	   &rec.request);
}

//----------------------------------------------------------------------
//
// ulm_get_respond_receive
//
// Process get responses.  Put it where it belongs and
// ring the bell.  This should be called after you
// have determined that a get message is available.
//
//----------------------------------------------------------------------

static void
ulm_get_respond_receive(MessageStruct_t *status)
{
  // Convert the sending tag to the one the user used

  int usertag = ulm_tag_to_user_tag(status->tag);

  // When you get a response message, look up the place to put
  // it and then put it there.  We know the tag from the
  // probe. Use that to find the buffer to use for the
  // receive.

  ulm_get_iter_t p = ulm_outstanding_gets_s.find(usertag);
  if (p != ulm_outstanding_gets_s.end())
    {
      //
      // Receive the message now, since we know it is available.
      //
      ulm_blocking_recv(status->SourceProc, status->tag,
			(void *)(p->second.local),
			status->MessageLength, p->second.length);

      //
      // If the user asked us to ring a bell when it came in, do so.
      //
      if (p->second.bell)
	*p->second.bell += 1;

      //
      // Get rid of the record of this get.
      //
      ulm_outstanding_gets_s.erase(p);
    }
  else
    {
      //
      // It wasn't in the map which means we have an
      // unrecognized get response!  Print an error message
      // and go on.
      //
      fprintf(stderr,
	      "CHEETAH-ON-ULM ERROR: Unrecognized get response "
	      "tag %d!\n", usertag);
    }
}

//----------------------------------------------------------------------
//
// ulm_put_receive
//
// Check for put messages.  There are two messages involved in a put:
// the header that has the pointer to the place to put it and the tag
// for the payload message, and then there is the payload message.
//
// NOTE: For this to work, we must have a guarantee of correct message
// ordering from each context!!!  This is because the data is sent
// in two pieces, first a header and then a payload.  This routine
// is activated when the header comes in, and each header must arrive
// before the corresponding payload.
//----------------------------------------------------------------------

static void
ulm_put_receive(MessageStruct_t *status)
{
  // Convert the sending tag to the one the user used.  This will be
  // the tag used for the payload as well.

  int usertag = ulm_tag_to_user_tag(status->tag);

  //
  // We have a put request, so get the header.
  //
  ULMPutRequest msg;
  ulm_blocking_recv(status->SourceProc, status->tag, (void *)(&msg),
		    status->MessageLength, sizeof(ULMPutRequest));

  //
  // Get the user tag for the payload from the ULMPutRequest info, and
  // convert it into a ULM tag for the put payload
  //
  int ulmputtag = ulm_setup_tag(usertag, put_payload_comm_s);

  //
  // Use the info in the header to pull in the payload.  This will
  // be a message from the same sender as the header, but with a different
  // tag (ulmputtag, derived from tag info in the header).  We do this
  // with a probe since we know that if a node sent the header, it will
  // send they payload right afterwards.
  //
  ulm_blocking_recv(status->SourceProc, ulmputtag, (void *)(msg.remote),
		    msg.length, msg.length);

  //
  // If the user requested a doorbell, ring it.
  //
  if (msg.bell)
    {
      *msg.bell += 1;
    }
}

//----------------------------------------------------------------------
//
// ulm_isend_completion
//
// Check and see if any of the isends have completed.  Return true if
// any have.
//
//----------------------------------------------------------------------

static bool
ulm_isend_completion()
{
  ULM_DEBUG_MSG(cerr << "Checking for isend completions on node ";)
  ULM_DEBUG_MSG(cerr << mycontext_s << endl;)

  //
  // Keep track of whether we did anything.
  //
  bool did_something = false;

  //
  // Loop over all of the outstanding isends.
  //
  ulm_isend_iter_t p = ulm_outstanding_isends_s.begin();
  ulm_isend_iter_t e = ulm_outstanding_isends_s.end();

  ULM_DEBUG_MSG(cerr << "There are " << ulm_outstanding_isends_s.size();)
  ULM_DEBUG_MSG(cerr << " outstanding isends on node " << mycontext_s << endl;)

  while (p != e)
    {
      ULM_DEBUG_MSG(cerr << "Checking isend seq = " << p->request.seq;)
      ULM_DEBUG_MSG(cerr << " on node " << mycontext_s;)
      ULM_DEBUG_MSG(cerr << ", destined for node " << p->request.proc << endl;)

      //
      // Find out if this isend has completed.
      //
      int tval = tp_test(p->request, 0);
      if (tval < 0)
	{
	  fprintf(stderr, "CHEETAH-ON-ULM ERROR: Error from tp_isend!!\n");
	}
      else if (tval > 0)
	{
	  //
	  // It has completed, so ring the bell for this isend.
	  // 
	  if (p->bell)
	    *p->bell += 1;

	  //
	  // Delete this entry in the list and go on to the next.
	  // Do this carefully so we don't have any dangling pointers.
	  //
	  ulm_isend_iter_t p1 = p;
	  ++p;
	  ulm_outstanding_isends_s.erase(p1);

	  //
	  // Record that we did something.
	  //
	  did_something = true;
	}
      else
	{
	  //
	  // This isend hasn't completed, so go on to the next one.
	  //
	  ++p;
	}
    }

  //
  // Tell the caller if we accomplished something.
  //
  return did_something;
}

//----------------------------------------------------------------------
//
// ulm_message_poll
//
// Just look for an incoming ulm-specific message, for rfi, get, or
// put.  If one arrives, decode the tag to determine what operation
// should be done, and call the appropriate function.
//
//----------------------------------------------------------------------

static bool
ulm_message_poll()
{
  // Prepare struct indicating we want to look for a ULM message.

  MessageStruct_t ulmprobedata;
  ulmprobedata.SourceProc = (-1);
  ulmprobedata.tag = (-1);
  ulmprobedata.MessageLength = 0;

  // Look for any kind of message at all

  ULM_DEBUG_MSG(cerr << "Checking for ULM message ..." << endl;)

  if (tp_iprobe(&ulmprobedata))
    {
      ULM_DEBUG_MSG(cerr << "Probe returned true: tag = " << ulmprobedata.tag;)
      ULM_DEBUG_MSG(cerr << ", source = " << ulmprobedata.SourceProc;)
      ULM_DEBUG_MSG(cerr << ", size = " << ulmprobedata.MessageLength << endl;)

      // Also compute what operation this was for

      int opcode = ulm_tag_to_opcode(ulmprobedata.tag);

      // Based on the opcode, invoke the proper operation

      if (opcode == rfi_comm_s)
	{
	  ulm_rfi_receive(&ulmprobedata);
	}
      else if (opcode == get_request_comm_s)
	{
	  ulm_get_request_receive(&ulmprobedata);
	}
      else if (opcode == get_respond_comm_s)
	{
	  ulm_get_respond_receive(&ulmprobedata);
	}
      else if (opcode == put_header_comm_s)
	{
	  ulm_put_receive(&ulmprobedata);
	}
      else
	{
	  fprintf(stderr, "ERROR: Unknown opcode %d in %s\n", opcode,
		  "cheetah_ulm_message_poll");
	  return false;
	}

      // If we're here, we were able to receive something

      return true;
    }
  else
    {
      ULM_DEBUG_MSG(cerr << "tp_iprobe did not find anything on node ";)
      ULM_DEBUG_MSG(cerr << mycontext_s << endl;)
    }

  // If we're here, we did not find anything with a probe

  return false;
}

//----------------------------------------------------------------------
//
// cheetah_ulm_poll
//
// Check the rfi, get and put queues.  Keep pulling things off the
// wire as long as there are things to do.
//
//----------------------------------------------------------------------

void
cheetah_ulm_poll()
{
  bool did_something;
  do
    {
      did_something = false;
      did_something |= ulm_message_poll();
      did_something |= ulm_isend_completion();
    }
  while ( did_something );
}

//----------------------------------------------------------------------
//
// cheetah_ulm_ainvoke
//
// Invoke a function in a remote context.  The function is given by
// the tag, and we pass a buffer with a given length to the function.
// Since we can use the function tag as the message tag, all we have
// to do is a send.
//
//----------------------------------------------------------------------

void
cheetah_ulm_ainvoke(int context, int tag, void *buffer, int len,
		  int *local_bell)
{
  ULM_DEBUG_MSG(cerr << "Calling cheetah_ulm_ainvoke on node " << mycontext_s;)
  ULM_DEBUG_MSG(cerr << " with tag=" << tag << ", len=" << len << " to node ";)
  ULM_DEBUG_MSG(cerr << context << endl;)

  //
  // Keep things moving.
  //
  cheetah_ulm_poll();

  //
  // If we are ainvoking on our same context, just try to call the
  // handler right now.
  //
  if (context == mycontext_s)
    {
      // Find the registered function for this message and call it.
      ulm_invoke_handler(context, tag, buffer, len);

      // Ring the bell if necesary.
      if (local_bell)
	*local_bell += 1;

      // And get the heck out of here.
      return;
    }

  //
  // Determine the actual tag we'll use in the ULM call.  This will
  // encode the user's tag and the fact that we're doing an ainvoke
  // 
  int sendtag = ulm_setup_tag(tag, rfi_comm_s);

  //
  // Determine what actual bell to use.  Either we use the user's,
  // or we use our own and make sure the item is sent before we
  // return.
  //
  int ulmbell = 0;
  int *bell = (local_bell != 0 ? local_bell : &ulmbell);
  int waitvalue = *bell + 1;

  //
  // Send the request.
  //
  ulm_outstanding_isends_s.push_back( ULMIsendRecord() );
  ULMIsendRecord& rec = ulm_outstanding_isends_s.back();
  rec.bell = bell;
  ULM_DEBUG_MSG(cerr << "About to call tp_isend, with sendtag = ";)
  ULM_DEBUG_MSG(cerr << sendtag << ", bell = " << *bell << endl;)
  tp_isend(sendtag, context, buffer, len, &rec.request);

  //
  // If the user did not provide a bell, we must wait here for
  // our own locally-provide bell to be rung.  This makes this send
  // blocking to the user's point of view.
  //
  if (local_bell == 0)
    {
      ULM_DEBUG_MSG(cerr << "Waiting for ulmbell in blocking ainvoke" << endl;)
      cheetah_ulm_wait(bell, waitvalue);
    }
}


//----------------------------------------------------------------------
//
// cheetah_ulm_put
//
// Send a local buffer of a given length to a remote buffer in a given
// context.  Ring a remote doorbell when it is done.
//
// This requires two messages: one for the header and one for the
// payload. The header has the pointer to the remote buffer, the
// length of the message, the bell to ring and the message tag that
// the payload will have.
//
//----------------------------------------------------------------------

void
cheetah_ulm_put(int context, void *remote, void *local, int length,
	      int *local_bell, int *remote_bell)
{
  //
  // Keep things flowing.
  //
  cheetah_ulm_poll();

  //
  // Determine the actual tag we'll use in the ULM call.  This will
  // encode the user's tag and the fact that we're doing an ainvoke
  //
  int puttag = ulm_next_put_tag();
  int headertag = ulm_setup_tag(puttag, put_header_comm_s);
  int payloadtag = ulm_setup_tag(puttag, put_payload_comm_s);

  //
  // Build the header with the info the receiver needs.
  //
  ULMPutRequest msg;
  msg.remote = remote;
  msg.length = length;
  msg.bell = remote_bell;

  //
  // Send the header. Block on this one because it is on the stack and
  // small.
  //
  TPMsg ReturnHandle;
  tp_isend(headertag, context, (void *)(&msg), sizeof(msg), &ReturnHandle);
  if (tp_wait(ReturnHandle, 0) < 0)
    {
      fprintf(stderr, "CHEETAH-ON-ULM ERROR: Error from tp_wait!!\n");
    }

  //
  // Send the payload.  If we've specified a local bell, make this
  // asynchronous, otherwise make it synchronous.
  //
  if (local_bell)
    {
      //
      // Use isend and remember it so we can check on it
      // later.
      //
      ulm_outstanding_isends_s.push_back( ULMIsendRecord() );
      ULMIsendRecord& rec = ulm_outstanding_isends_s.back();
      rec.bell = local_bell;
      tp_isend(payloadtag, context, local, length, &rec.request);
    }
  else
    {
      //
      // If we're synchronous, just send the message and let ULM do
      // the blocking.
      //
      TPMsg ReturnHandle;
      tp_isend(payloadtag, context, local, length, &ReturnHandle);
      if (tp_wait(ReturnHandle, 0) < 0)
	{
	  fprintf(stderr, "CHEETAH-ON-ULM ERROR: Error from tp_wait!!\n");
	}
    }
}

//----------------------------------------------------------------------
//
// cheetah_ulm_get
//
// Get a buffer of a given length from a remote context and put it in
// a local buffer, and ring a bell when it is here.
//
// This doesn't block, it just sends the request.  Some time later a
// poll will find that it has arrived, and the bell gets rung at that
// time.
//
//----------------------------------------------------------------------

void
cheetah_ulm_get(int context, void *remote, void *local, int length,
	      int *local_bell, int *remote_bell)
{
  //
  // Keep things moving.
  //
  cheetah_ulm_poll();

  //
  // Determine the actual tag we'll use in the ULM call.  This will
  // encode the user's tag and the fact that we're doing a get request.
  // The 'gettag' value will then be used in the reply.
  //
  int gettag = ulm_next_get_tag();
  int requesttag = ulm_setup_tag(gettag, get_request_comm_s);

  //
  // Build the request with the info the remote processor needs.
  //
  ULMGetRequest msg;
  msg.remote = remote;
  msg.length = length;
  msg.bell = remote_bell;

  //
  // If no local bell was specified, we need to have a backup.
  //
  int backup_bell = 0;

  //
  // Build a record with the info that this context will need when the
  // reply comes back.  Store it in the container of outstanding gets.
  //
  ULMGetRecord rec;
  rec.local = local;
  rec.length = length;
  rec.bell = local_bell ? local_bell : &backup_bell;
  ulm_outstanding_gets_s[gettag] = rec;

  //
  // Send the request.  Make this blocking since the thing we're
  // sending is small and on the stack.
  //
  TPMsg ReturnHandle;
  tp_isend(requesttag, context, (void *)(&msg), sizeof(msg), &ReturnHandle);
  if (tp_wait(ReturnHandle, 0) < 0)
    {
      fprintf(stderr, "CHEETAH-ON-ULM ERROR: Error from tp_wait!!\n");
    }

  //
  // If no local bell was specified, block until the reply arrives.
  //
  if (local_bell == 0)
    cheetah_ulm_wait(&backup_bell,1);
}

//----------------------------------------------------------------------
//
// cheetah_ulm_wait
//
// Wait for a given bell to ring (get to a particular value) and
// process messages while you're waiting.
//
//----------------------------------------------------------------------

void
cheetah_ulm_wait(volatile int *bell, int value)
{
  while (*bell < value)
    cheetah_ulm_poll();
}

//----------------------------------------------------------------------
//
// cheetah_ulm_barrier
//
// Do a barrier among all of the contexts.
//
//----------------------------------------------------------------------

void
cheetah_ulm_barrier()
{
  ULM_DEBUG_MSG(cerr << "Starting ULM barrier on node " << mycontext_s<<endl;)

  //
  // Tell all the other contexts that we arrived at this barrier.
  //
  for (int i=0; i < ncontexts_s; ++i)
    if (i != mycontext_s)
      cheetah_ulm_ainvoke(i, ulm_barrier_tag_s, 0, 0, 0);

  //
  // Wait for all of the remote contexts to tell us that they arrived
  // at this barrier.  First get a reference to the counter that they
  // will increment.
  //
  int &counter = ulm_barrier_counters_s[ulm_next_barrier_s[mycontext_s]];

  //
  // Spin until all the expected messages arrive, and poll while you
  // spin to keep things flowing.
  //
  ULM_DEBUG_MSG(cerr << "ULM barrier polling on node " << mycontext_s << endl;)

  while (counter < ncontexts_s-1)
    cheetah_ulm_poll();

  //
  // Reset the counter now that we're through the barrier.
  //
  counter = 0;

  //
  // Use the other counter for the next barrier.
  //
  ulm_next_barrier_s[mycontext_s] = 1-ulm_next_barrier_s[mycontext_s];
}

//----------------------------------------------------------------------
//
// cheetah_ulm_setup
//
// Set up the data structures for one sided messaging.
//
//----------------------------------------------------------------------

void
cheetah_ulm_setup(int *argc, char ***argv)
{
  ULM_DEBUG_MSG(cerr << "Initializing ulm ..." << endl;)

  //
  // Initialize ULM
  //
  tp_init();

  //
  // Cache the number of contexts and the current context.
  //
  ncontexts_s = tp_GetNProcs();
  mycontext_s = tp_GetMyProcID();
  
  ULM_DEBUG_MSG(cerr << "Initialization complete, on node " << mycontext_s;)
  ULM_DEBUG_MSG(cerr << " of " << ncontexts_s << endl;)

  //
  // Initialize the array that holds the next slot to use in the
  // barrier for each remote CPU.
  // Register the barrier handler.
  //
  ulm_next_barrier_s = new int[ncontexts_s];
  for (int i=0; i<ncontexts_s; ++i)
    ulm_next_barrier_s[i] = 0;

  ulm_barrier_tag_s = Cheetah::Controller::registerHandler(ulm_barrier_handler);
}

//----------------------------------------------------------------------
//
// cheetah_ulm_finalize
//
// Shut down the system.  Do a barrier to make sure everyone gets
// here, then shut down ULM.
//
//----------------------------------------------------------------------

void
cheetah_ulm_finalize()
{
  cheetah_ulm_barrier();
  tp_finalize();
}

//----------------------------------------------------------------------
//
// cheetah_ulm_ncontexts
// cheetah_ulm_mycontext
//
// Get the total number of contexts participating and which context
// this is.
//
//----------------------------------------------------------------------

int
cheetah_ulm_ncontexts()
{
  return ncontexts_s;
}

int
cheetah_ulm_mycontext()
{
  return mycontext_s;
}
