#ifndef CHEETAH_CORE_GLOBALS_H
#define CHEETAH_CORE_GLOBALS_H

//----------------------------------------------------------------------
//
// Classes:
//    ConstGlobalRef
//    ConstGlobalPtr
//    GlobalRef
//    GlobalPtr
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
// NOTE: The following routines are based on an older design that used
// a 'core' interface and these C++ functions as wrappers around that
// core.  This design has since been changed to use polymorphic 'controllers'
// for the core functionality, and these routines have NOT been updated to
// that new design.  They are here for historical reference.  Because
// of this, they are disabled with an #if 0 / #endif block.
//
//-----------------------------------------------------------------------------

#if 0

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include <cheetah_core.h>
#include <cheetah_retrieve.h>
#include <cheetah_store.h>
#include <cheetah_operate.h>

//-----------------------------------------------------------------------------
// Forward References
//-----------------------------------------------------------------------------

template<class T> class ConstGlobalRef;
template<class T> class ConstGlobalPtr;
template<class T> class GlobalRef;
template<class T> class GlobalPtr;

//----------------------------------------------------------------------
//
// ConstGlobalRef
//
// A reference to a possibly remote const object of type T.  This will
// eventually have as much of the interface of a smart reference as
// C++ will allow.
//
//----------------------------------------------------------------------

template<class T>
class ConstGlobalRef
{
  friend ConstGlobalPtr<T>;
protected:

  //
  // We store a pointer to the object and the context it is in.  If
  // the context is not this context then we had better not
  // dereference this pointer here!
  //
  T* object_m;
  int context_m;

  //
  // Construct with the object and the context.  Keep this protected
  // so that only subclasses and friends can use it.
  //
  ConstGlobalRef(const T *obj, int context)
    : object_m((T*)obj), context_m(context) {}

public:

  //
  // Default ctor constructs a pointer to nothing.  That means a null
  // ptr and a context -1.
  //
  inline ConstGlobalRef() : object_m(0), context_m(-1) {}

  //
  // Construct pointing to the given object.  Just get the pointer to
  // the object and use mycontext().
  //
  inline ConstGlobalRef( const T& p )
    : object_m( (T*)&p ) , context_m(cheetah_core_mycontext()) {}

  //
  // Copy ctor.
  //
  inline ConstGlobalRef( const ConstGlobalRef<T>& x )
    : object_m(x.object_m) , context_m(x.context_m) {}

  //
  // Attach to a given local variable.
  //
  inline const ConstGlobalRef<T>& attach(const T& x)
    {
      object_m = &x;
      context_m = cheetah_core_mycontext();
      return *this;
    }

  //
  // Detach from whatever you're pointing to.
  //
  inline const ConstGlobalRef<T>& detach(const T& x)
    {
      object_m = 0;
      context_m = -1;
      return *this;
    }

  //
  // Return whether the object is in this context.
  //
  inline bool is_local() const
    {
      return context_m == cheetah_core_mycontext();
    }

  //
  // Get a pointer to the object.  If the object is not in this
  // context, we have to retrieve it.  This routine is set up so that
  // the default path just returns the pointer.
  //
  inline const T* pointer() const
    {
      T* p = object_m;
      if ( ! is_local() )
	p = CheetahRetrieve<T>::retrieve(context_m,object_m);
      return p;
    }

  //
  // The 'find the pointer' operator returns a GlobalPtr, not a
  // pointer to the current object.
  //
  /*
  ConstGlobalPtr<T> operator&()
    {
      return ConstGlobalPtr<T>(object_m, context_m);
    }
  */
  //
  // Print the component data to a stream.  This gets packed up below
  // inside of operator<<.
  //
  template<class Stream>
  void print(Stream& s) const;

};


//----------------------------------------------------------------------
//
// GlobalRef
//
// A non-const version of ConstGlobalRef.  The difference is that we
// are allowed to modify the object we're pointing to through this
// interface through operations like assignment.  That is purely an
// extension to the interface, so GlobalRef inherits from
// ConstGlobalRef.
//
//----------------------------------------------------------------------

template<class T>
class GlobalRef : public ConstGlobalRef<T>
{
  friend GlobalPtr<T>;

protected:

  //
  // Construct with the object and the context.  Keep this protected
  // so that only subclasses and friends can use it.
  //
  GlobalRef(const T *obj, int context)
    : ConstGlobalRef<T>(obj,context) {}

public:

  //
  // Construct with no args, an object of type T, or another
  // GlobalRef.  Just have the base class construct itself.
  //
  inline GlobalRef() {}
  inline explicit GlobalRef( T &x ) : ConstGlobalRef<T>(x) {}
  inline GlobalRef( const GlobalRef<T>& x ) : ConstGlobalRef<T>(x) {}

  //
  // Attach to a given local variable.
  //
  inline const GlobalRef<T>& attach(T& x)
    {
      object_m = &x;
      context_m = cheetah_core_mycontext();
      return *this;
    }

  //
  // Detach from whatever you're pointing to.
  //
  inline const GlobalRef<T>& detach(T& x)
    {
      object_m = 0;
      context_m = -1;
      return *this;
    }

  //
  // Get a pointer to the object, retrieving it if necessary.  Just
  // use the routine in the base class and cast the pointer.
  //
  inline T* pointer()
    {
      T* p = object_m;
      if ( ! is_local() )
	p = CheetahRetrieve<T>::retrieve(context_m,object_m);
      return p;
    }

  //
  // The 'find the pointer' operator returns a GlobalPtr, not a
  // pointer to the current object.
  //
  /*
  GlobalPtr<T> operator&()
    {
      return GlobalPtr<T>(object_m, context_m);
    }
  */
  //
  // Assign a value to the object we are pointing to.  If it is remote
  // we'll have to do a put.
  //
  const ConstGlobalRef<T>& operator=(const T&) const;
  const ConstGlobalRef<T>& operator+=(const T&) const;
  const ConstGlobalRef<T>& operator-=(const T&) const;
  const ConstGlobalRef<T>& operator*=(const T&) const;
  const ConstGlobalRef<T>& operator/=(const T&) const;

  const GlobalRef<T>& operator++() const;
  const GlobalRef<T>& operator--() const;
};


//----------------------------------------------------------------------
//
// ConstGlobalPtr
//
// Similar to GlobalRef except that the various operations are
// shallow: is they operate on the pointer instead of on the object
// we're pointing to. The goal here is to implement as much of the
// smart pointer interface as makes sense.
//
//----------------------------------------------------------------------

template<class T>
class ConstGlobalPtr
{
  friend ConstGlobalRef<T>;
protected:

  //
  // A pointer to the (possibly) remote object, and the context it is
  // in.
  //
  const T* object_m;
  int context_m;

  //
  // Construct directly from the object and the context.  Keep this
  // protected so that only friends and subclasses can see it.
  //
  ConstGlobalPtr(const T* obj, int context)
    : object_m(obj), context_m(context) {}

public:

  //
  // The default ctor initializes to an invalid state.
  //
  ConstGlobalPtr() : object_m(0), context_m(-1) {}

  //
  // Initialize pointing to the given object.
  //
  explicit ConstGlobalPtr(const T* x)
    : object_m(x), context_m(cheetah_core_mycontext()) {}

  //
  // Assign a local pointer to the ConstGlobalPtr.  If we're assigning
  // a null ptr, then we're making this invalid, so set the context to
  // an invalid value.
  //
  const ConstGlobalPtr<T>& operator=(const T* x)
    {
      object_m = x;
      if ( x == 0 )
	context_m = -1;
      else
	context_m = cheetah_core_mycontext();
      return *this;
    }

  //
  // Return whether or not the object is local.
  //
  bool is_local() const
    {
      return context_m == cheetah_core_mycontext();
    }

  //
  // Get a pointer to the object.  If it is local, just return the
  // pointer. If it isn't local, go get it.
  //
  const T* pointer() const
    {
      T* p = object_m;
      if ( ! is_local() )
	p = CheetahRetrieve<T>::retrieve(context_m,object_m);
      return p;
    }

  //
  // Dereference to get a global reference.  We're friends with
  // ConstGlobalRef, so just initialize the data members.
  //
  ConstGlobalRef<T> operator*() const
    {
      return ConstGlobalRef<T>(object_m,context_m);
    }

  //
  // Add and subtract an integer type.  Builds a new ConstGlobalPtr
  // which is offset from the original.  We handle these as friends so
  // that we can do int+ptr as well as ptr+int.  We template on the
  // integer type so that any of the integral types will work.
  //
  template<class Int>
  ConstGlobalPtr<T> operator+(Int x) const
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      return ConstGlobalPtr<T>(object_m+x,context_m);      
    }
  template<class Int>
  ConstGlobalPtr<T> operator-(Int x) const
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      return ConstGlobalPtr<T>(object_m-x,context_m);      
    }

  //
  // operator+=, operator-=
  //
  // Perform the operation on the pointer only.  We template on the
  // integer type so that any integer type could be used.
  //
  template<class Int>
  ConstGlobalPtr<T>& operator+=(Int x)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      object_m += x;
      return *this;
    }

  template<class Int>
  ConstGlobalPtr<T>& operator-=(Int x)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      object_m -= x;
      return *this;
    }

  //
  // Preincrement and predecrement.
  // Perform the operation on the pointer only.
  //
  ConstGlobalPtr<T>& operator++()
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      ++object_m;
      return *this;
    }

  ConstGlobalPtr<T>& operator--()
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      --object_m;
      return *this;
    }

  //
  // Post-increment and post-decrement.  
  // 
  // In the usual fashion, first we save the current state, then
  // increment/decrement the pointer, then return the saved state.
  //
  ConstGlobalPtr<T> operator++(int)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      GlobalPtr<T> save = *this;
      ++object_m;
      return save;
    }

  ConstGlobalPtr<T> operator--(int)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      GlobalPtr<T> save = *this;
      --object_m;
      return save;
    }

  //
  // operator[](Int)
  //
  // Offset from the current pointer and dereference.
  //
  template<class Int>
  ConstGlobalRef<T> operator[](Int x)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      return ConstGlobalRef<T>(object_m+x,context_m);
    }

  //
  // Comparison operators.  Make this a total ordering: first sort on
  // context, then on pointer.
  //
  bool operator==(const ConstGlobalPtr<T>& x) const
    {
      return (context_m == x.context_m) && (object_m == x.object_m);
    }
  bool operator<(const ConstGlobalPtr<T>& x) const
    {
      return
	(context_m < x.context_m) ||
	( (context_m == x.context_m) && (object_m < x.object_m) );
    }
  bool operator!=(const ConstGlobalPtr<T>& x) const
    {
      return !((*this)==x);
    }
  bool operator>(const ConstGlobalPtr<T>& x) const
    {
      return x < *this;
    }
  bool operator<=(const ConstGlobalPtr<T>& x) const
    {
      return !(x<(*this));
    }
  bool operator>=(const ConstGlobalPtr<T>& x) const
    {
      return !((*this)<x);
    }
};


//----------------------------------------------------------------------
//
// GlobalPtr<T>
//
// Much like ConstGlobalPtr above, except that when it returns
// pointers to the object or a GlobalRef for the object, it is
// non-const.
//
//----------------------------------------------------------------------

template<class T>
class GlobalPtr : public ConstGlobalPtr<T>
{
  friend GlobalRef<T>;

protected:

  //
  // Construct directly from the object and the context.  Keep this
  // protected so that only friends and subclasses can see it.
  //
  GlobalPtr(const T* obj, int context)
    : ConstGlobalPtr<T>(obj,context) {}

public:

  //
  // The default ctor initializes to an invalid state.
  //
  GlobalPtr() {}

  //
  // Initialize pointing to the given object.
  //
  explicit GlobalPtr(T* x)
    : ConstGlobalPtr<T>(x) {}

  //
  // Assign a local pointer to the ConstGlobalPtr.  If we're assigning
  // a null ptr, then we're making this invalid, so set the context to
  // an invalid value.
  //
  const ConstGlobalPtr<T>& operator=(T* x)
    {
      object_m = x;
      if ( x == 0 )
	context_m = -1;
      else
	context_m = cheetah_core_mycontext();
      return *this;
    }

  //
  // Get a pointer to the object.  If it is local, just return the
  // pointer. If it isn't local, go get it.
  //
  T* pointer() const
    {
      T* p = object_m;
      if ( ! is_local() )
	p = CheetahRetrieve<T>::retrieve(context_m,object_m);
      return p;
    }

  //
  // Dereference to get a global reference.  We're friends with
  // ConstGlobalRef, so just initialize the data members.
  //
  GlobalRef<T> operator*() const
    {
      return GlobalRef<T>(object_m,context_m);
    }

  //
  // Add and subtract an integer type.  Builds a new ConstGlobalPtr
  // which is offset from the original.  We handle these as friends so
  // that we can do int+ptr as well as ptr+int.  We template on the
  // integer type so that any of the integral types will work.
  //
  template<class Int>
  GlobalPtr<T> operator+(Int x) const
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      return GlobalPtr<T>(object_m+x,context_m);
    }
  template<class Int>
  GlobalPtr<T> operator-(Int x) const
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      return GlobalPtr<T>(object_m-x,context_m);
    }

  //
  // operator+=, operator-=.
  //
  // Perform the operation on the pointer only.  We template on the
  // integer type so that any integer type could be used.
  //
  template<class Int>
  GlobalPtr<T>& operator+=(Int x)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      object_m += x;
      return *this;
    }

  template<class Int>
  GlobalPtr<T>& operator-=(Int x)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      object_m -= x;
      return *this;
    }

  //
  // Preincrement and predecrement.
  //
  // Perform the operation on the pointer only.
  //
  GlobalPtr<T>& operator++()
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      ++object_m;
      return *this;
    }

  GlobalPtr<T>& operator--()
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      --object_m;
      return *this;
    }

  //
  // Post-increment and post-decrement.
  //
  // In the normal fashion, first we save the current state, then
  // increment/decrement the pointer, then return the saved state.
  //
  GlobalPtr<T> operator++(int)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      GlobalPtr<T> save = *this;
      ++object_m;
      return save;
    }

  GlobalPtr<T> operator--(int)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      GlobalPtr<T> save = *this;
      --object_m;
      return save;
    }

  //
  // operator[](Int)
  //
  // Offset from the current pointer and dereference.
  //
  template<class Int>
  GlobalRef<T> operator[](Int x)
    {
      assert( (context_m >= 0) && (context_m < cheetah_core_ncontexts()) );
      assert( object_m != 0 );
      return GlobalRef<T>(object_m+x,context_m);
    }
};


//----------------------------------------------------------------------
//
// Add or subtract an integer from ConstGlobalPtr and GlobalPtr.
// These all do basically the same thing: add or subtract the integer
// type from the pointer in the GlobalPtr.  The ones that are pointer
// plus integer are member functions, these are for integer plus
// pointer.  There is no integer minus pointer.
//
//----------------------------------------------------------------------

template<class T, class Int>
inline ConstGlobalPtr<T>
operator+(Int x,const ConstGlobalPtr<T> &p)
{
  return p+x;
}

template<class T, class Int>
inline GlobalPtr<T>
operator+(Int x,const GlobalPtr<T> &p)
{
  return p+x;
}


//----------------------------------------------------------------------
//
// ConstGlobalRef::print(Stream)
// operator<<(Stream&,ConstGlobalRef&)
//
// print outputs the ConstGlobalRef to a stream.  operator<< is
// syntactic sugar on top of that.
//
//----------------------------------------------------------------------

template<class T>
template<class Stream>
void 
ConstGlobalRef<T>::print(Stream& s) const
{
  s << "[" << object_m << "," << context_m << "]";
}

template<class Stream, class T>
inline Stream&
operator<<(Stream& s, const ConstGlobalRef<T>& g)
{
  g.print(s);
  return s;
}


//----------------------------------------------------------------------
//
// Assign to a remote object.  If it is actually here, just do it, but
// if not send it.
//
// If it is local, we just do the assignment, and if it is remote we
// send an instruction to have the remote side to the assignment.
//
// Below we have plain assignment, add assign, subtract assign,
// multiply assign and divide assign.
//
//----------------------------------------------------------------------

//
// Assign to a GlobalRef<T>
//
template<class T>
const ConstGlobalRef<T>&
GlobalRef<T>::operator=(const T& x) const
{
  assert(object_m != 0);
  if ( is_local() )
    *object_m = x;
  else
    CheetahStore<T>::store(context_m,object_m,x,CheetahStore<T>::ASSIGN);
  return *this;
}

//
// Add Assign to a GlobalRef<T>
//
template<class T>
const ConstGlobalRef<T>&
GlobalRef<T>::operator+=(const T& x) const
{
  assert(object_m != 0);
  if ( is_local() )
    *object_m += x;
  else
    CheetahStore<T>::store(context_m,object_m,x,CheetahStore<T>::ADD_ASSIGN);
  return *this;
}

//
// Subtract Assign to a GlobalRef<T>
//
template<class T>
const ConstGlobalRef<T>&
GlobalRef<T>::operator-=(const T& x) const
{
  assert(object_m != 0);
  if ( is_local() )
    *object_m -= x;
  else
    CheetahStore<T>::store(context_m,object_m,x,CheetahStore<T>::SUB_ASSIGN);
  return *this;
}

//
// Multiply Assign to a GlobalRef<T>
//
template<class T>
const ConstGlobalRef<T>&
GlobalRef<T>::operator*=(const T& x) const
{
  assert(object_m != 0);
  if ( is_local() )
    *object_m *= x;
  else
    CheetahStore<T>::store(context_m,object_m,x,CheetahStore<T>::MUL_ASSIGN);
  return *this;
}

//
// Divide Assign to a GlobalRef<T>
//
template<class T>
const ConstGlobalRef<T>&
GlobalRef<T>::operator/=(const T& x) const
{
  assert(object_m != 0);
  if ( is_local() )
    *object_m /= x;
  else
    CheetahStore<T>::store(context_m,object_m,x,CheetahStore<T>::DIV_ASSIGN);
  return *this;
}


//----------------------------------------------------------------------
//
// Increment or decrement a global ref.  This does the operation here
// if the object is local or passes the instruction to the far side if
// it is not.
//
//----------------------------------------------------------------------

//
// Increment.
//
template<class T>
const GlobalRef<T>&
GlobalRef<T>::operator++() const
{
  assert(object_m != 0);
  if ( is_local() )
    ++*object_m;
  else
    CheetahOperate<T>::operate(context_m,object_m,CheetahOperate<T>::INCREMENT);
  return *this;
}

//
// Decrement
//
template<class T>
const GlobalRef<T>&
GlobalRef<T>::operator--() const
{
  assert(object_m != 0);
  if ( is_local() )
    --*object_m;
  else
    CheetahOperate<T>::operate(context_m,object_m,CheetahOperate<T>::DECREMENT);
  return *this;
}

//----------------------------------------------------------------------

#endif   // #if 0

#endif // CHEETAH_CORE_GLOBALS_H
