#ifndef CHEETAH_CORE_REDUCTIONS_H
#define CHEETAH_CORE_REDUCTIONS_H

//-----------------------------------------------------------------------------
// C++ function interface to Cheetah capabilities.
//
// NOTE: The following routines are based on an older design that used
// a 'core' interface and these C++ functions as wrappers around that
// core.  This design has since been changed to use polymorphic 'controllers'
// for the core functionality, and these routines have NOT been updated to
// that new design.  They are here for historical reference.  Because
// of this, they are disabled with an #if 0 / #endif block.
//
//-----------------------------------------------------------------------------

#if 0

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include <assert.h>
#include <cheetah_core.h>
#include <cheetah_ainvoke.h>
#include <cheetah_register.h>
#include <vector>


//----------------------------------------------------------------------
//
// void cheetah_sum(const T& input, T& output, int *bell)
//
// An asynchronous reduce capability built with ainvoke.  The inputs
// from all of the contexts are accumulated into the output, and the
// bell is rung when the sum is ready.
//
// This is synchronized based on calling order.  If you have ten calls
// to cheetah_sum they keep each other separate by depending on all of
// the contexts calling cheetah_sum in the same order.
//
// The algorithm used here is not scalable: everybody just sends to
// context zero, which broadcasts the answer.
//
//----------------------------------------------------------------------


//----------------------------------------------------------------------
//
// class ReduceQueues
//
// This class contains all of the functionality of the reductions.  It
// has no non-static members, and is mainly used for scoping so that
// the static data is not at global scope and so that it is not
// globally visible.
//
//----------------------------------------------------------------------

template<class T>
class ReduceQueues
{
public:

  //----------------------------------------------------------------------
  //
  // cheetah_sum(input,output,bell).
  //
  // If called on someone other than context zero, send your input to
  // zero. If called on zero, get things ready for the data to come
  // in.
  //
  //----------------------------------------------------------------------

  static void cheetah_sum(const T& input, T& output, int *bell)
    {
      //
      // Context zero does different things from the other processors.
      //
      if ( cheetah_core_mycontext() == 0 )
	{
	  if ( cheetah_core_ncontexts() > 1 )
	    {
	      sum_on_zero(input,output,bell);
	    }
	  else
	    {
	      //
	      // If there is only one context in the whole thing, then
	      // just copy the input to the output and ring the bell.
	      //
	      output = input;
	      ++*bell;
	    }
	}
      else
	{
	  //
	  // If we're not processor zero, generate a unique integer to
	  // identify this reduction, save a record of where to put
	  // the output when the sum comes back, and send our input to
	  // context zero.
	  //
	  int id = unique();
	  replies_s.push_back( Reply(id,&output,bell) );
	  cheetah_ainvoke(0,reduce_tag_s,0,input,id);
	}
    }

private:

  //----------------------------------------------------------------------
  //
  // struct Accumulator
  //
  // The data structure used on context zero to accumulate the sum
  // before sending it back out.
  //
  //----------------------------------------------------------------------

  struct Accumulator
  {
    //
    // The unique id for this reduction.
    //
    int id_m;

    //
    // The number of values that have come in so far and their sum.
    //
    int count_m;
    T input_m;

    //
    // Where to store the result on context zero and the bell to ring
    // when it comes in.
    //
    T* output_m;
    int *bell_m;

    //
    // Two constructors.  The first is used when a value comes off the
    // wire before context zero has called a reduction.  It stores a
    // null ptr for where to put the result and for the local bell to
    // ring when we're done.  The second is used when context zero
    // calls sum before anything comes off the wire, and includes the
    // local bell and the destination location.
    //
    Accumulator(int u, const T& i)
      : id_m(u), bell_m(0), count_m(1), output_m(0), input_m(i) {}
    Accumulator(int u, int *b, const T& i, T* o)
      : id_m(u), bell_m(b), count_m(1), output_m(o), input_m(i) {}
  };

  //----------------------------------------------------------------------
  //
  // struct Reply
  //
  // The data structure that is stored on contexts other than zero.
  // It has the unique id so it can recognize the accumulated sum when
  // it comes back, the location to store the result, and the bell to
  // ring when it gets here.
  //
  //----------------------------------------------------------------------

  struct Reply
  {
    int id_m;
    T* output_m;
    int *bell_m;
    Reply(int u, T* o, int* b) : id_m(u), output_m(o), bell_m(b) {}
  };

  //----------------------------------------------------------------------
  //
  // Types for the containers on context zero and the other contexts.
  //
  //----------------------------------------------------------------------

  typedef std::vector<Accumulator> ResultVector_t;
  typedef std::vector<Reply> ReplyVector_t;

  //----------------------------------------------------------------------
  //
  // unique()
  //
  // Generate a unique ID for a reduction.
  //
  //----------------------------------------------------------------------

  static inline int unique()
    {
      return next_unique_s++;
    }

  //----------------------------------------------------------------------
  //
  // broadcast_result()
  // 
  // Once context zero has accumulated the result, broadcast it back
  // to the other contexts and remove this entry from the queue of
  // ongoing reductions.
  //
  //----------------------------------------------------------------------

  static void broadcast_result(ResultVector_t::iterator p) 
    {
      //
      // Loop over all of the other contexts, telling each one the
      // final value and the unique id.  Use one doorbell for all of
      // the ainvokes.
      //
      int nprocs = cheetah_core_ncontexts();
      int lb = 0;
      for (int i=1; i<nprocs; ++i)
	cheetah_ainvoke(i,reply_tag_s,&lb,p->input_m,p->id_m);
      cheetah_core_wait(&lb,nprocs-1);

      //
      // Store the local result and ring the local bell.
      //
      *p->output_m = p->input_m;
      ++*p->bell_m;

      //
      // Take this accumulation out of the queue, since it is done.
      //
      results_s.erase(p);
    }

  //----------------------------------------------------------------------
  //
  // sum()
  //
  // The function called in context zero by the other contexts to add
  // an element to the sum.  It tries to find the local record of the
  // sum, and if it does, adds to it, and if it doesn't it starts out
  // the sum.
  //
  //----------------------------------------------------------------------

  static void sum(T& input, int id)
    {
      //
      // Loop over all of the reductions currently under way.
      //
      ResultVector_t::iterator p = results_s.begin();
      ResultVector_t::iterator e = results_s.end();
      for (; p!=e; ++p)
	{
	  //
	  // If the id's match, this record is from the same
	  // reduction.
	  //
	  if ( p->id_m == id )
	    {
	      //
	      // Add in this value and increment the number that have
	      // arrived so far.
	      //
	      p->input_m += input;
	      ++p->count_m;

	      //
	      // If this is the last one, we store the final value
	      // locally, increment the local bell to indicate we're
	      // done, and then broadcast the result to the other
	      // contexts.
	      //
	      if ( p->count_m == cheetah_core_ncontexts() )
		broadcast_result(p);

	      //
	      // We found what we're looking for, so get out of here.
	      //
	      return;
	    }
	}
      //
      // We didn't find any records with a matching id, so start out a
      // new record for this reduction.
      //
      results_s.push_back( Accumulator(id,input) );
    }

  //----------------------------------------------------------------------
  //
  // reply()
  //
  // When the reduction is done, the contexts other than zero get this
  // called to inform them of the final answer for the reduction with
  // the given id.
  //
  //----------------------------------------------------------------------

  static void reply(T& final, int id)
    {
      //
      // Loop over all the outstanding reductions.
      //
      ReplyVector_t::iterator p = replies_s.begin();
      ReplyVector_t::iterator e = replies_s.end();
      for (; p!=e; ++p)
	{
	  //
	  // If the id's match, save the final result, increment the
	  // local bell, and remove the entry from the outstanding
	  // reductions.
	  //
	  if ( p->id_m == id )
	    {
	      *p->output_m = final;
	      ++(*p->bell_m);
	      replies_s.erase(p);
	      return;
	    }
	}
      //
      // If we didn't find it then something is seriously wrong.
      //
      assert(false);
    }

  //----------------------------------------------------------------------
  //
  // sum_on_zero()
  //
  // In context zero, initiate a reduction of a given input into a
  // given output and ring the given bell when we're done.
  //
  //----------------------------------------------------------------------

  static void sum_on_zero(const T& input, T& output, int *bell)
    {
      //
      // Generate a unique id for this context, and then loop over all
      // of the outstanding reductions to see if one of the other
      // contexts has beaten us to this point and the reduction is
      // already under way.
      //
      int id = unique();
      ResultVector_t::iterator p = results_s.begin();
      ResultVector_t::iterator e = results_s.end();
      for (; p!=e; ++p)
	{
	  //
	  // If the ids match, one of the other contexts got here
	  // ahead of us.
	  //
	  if ( p->id_m == id )
	    {
	      //
	      // Add this value to the sum and increment the number of
	      // contexts that have gotten to here. Record where to
	      // put the final answer, and what bell to ring when we
	      // finish.
	      //
	      p->input_m += input;
	      p->count_m += 1;
	      p->bell_m = bell;
	      p->output_m = &output;

	      //
	      // If this was the last one, store the local result,
	      // ring the local bell, and broadcast the result back to
	      // the other contexts.
	      //
	      if ( cheetah_core_ncontexts() == p->count_m )
		broadcast_result(p);

	      return;
	    }
	}

      //
      // If we didn't find a record here, then context zero is the
      // first one to call sum, so we just write down the input data.
      //
      results_s.push_back( Accumulator(id,bell,input,&output) );

    }

  //----------------------------------------------------------------------
  // Static data structures.
  //
  // int next_unique_s: An integer we increment to generate unique tags.
  // int reduce_tag_s: the registered tag for ReduceQueues<T>::sum
  // int reply_tag_s: the registered tag for ReduceQueues<T>::reply
  // results_s: The vector in context zero of outstanding reductions.
  // replies_s: The vector on the other contexts of outstanding reductions.
  // 
  //----------------------------------------------------------------------

  static int next_unique_s;
  static int reduce_tag_s;
  static int reply_tag_s;
  static ResultVector_t results_s;
  static ReplyVector_t replies_s;
};

//
// Definitions of the static data structures.
//

template<class T> 
int ReduceQueues<T>::next_unique_s = 10;

template<class T> 
int ReduceQueues<T>::reduce_tag_s = cheetah_register(ReduceQueues<T>::sum);

template<class T> 
int ReduceQueues<T>::reply_tag_s = cheetah_register(ReduceQueues<T>::reply);

template<class T>
ReduceQueues<T>::ResultVector_t ReduceQueues<T>::results_s;

template<class T>
ReduceQueues<T>::ReplyVector_t ReduceQueues<T>::replies_s;


//----------------------------------------------------------------------
//
// cheetah_sum
//
// The user visible function.  It takes in the value to be reduced in
// this context, the place to put the result, and the bell to ring
// when it arrives.
// 
//----------------------------------------------------------------------

template<class T>
inline void cheetah_sum(const T& input, T& output, int *bell)
{
  ReduceQueues<T>::cheetah_sum(input,output,bell);
}

//----------------------------------------------------------------------

#endif   // #if 0

#endif // CHEETAH_CORE_REDUCTIONS_H
