#ifndef CHEETAH_MATCHINGHANDLER_MATCHINGACTION_H
#define CHEETAH_MATCHINGHANDLER_MATCHINGACTION_H

//-----------------------------------------------------------------------------
// Classes:
//   MatchingAction
//   RemoveReference
//   MethodHandler
//   ExtraHandler1
//   FunctionAction0 ... FunctionAction4
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
//
// These routines are used to wrap around user-provided functions an
// functors; they store the user prvoided item and implement a virtual
// function 'run' used to unpack a given buffer and invoke the
// user handler.  MatchingAction is the base class; other classes
// are used for 1,2,3,4 arguments with and without extra other local values.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "MatchingHandler/Serialize.h"
#include "Utilities/CheetahAssert.h"


namespace Cheetah {

//----------------------------------------------------------------------
//
// class MatchingAction
//
// Action is a base class for object that will implement a "action" to
// take when a MatchingHandler message-handler receives a message from
// another context saying "execute this function".  Action wraps up
// the function to call, since different functions have different types
// and arguments.  MatchingAction is an abstract base class; derived
// classes handle performing the action for different numbers of arguments,
// etc.
//
//----------------------------------------------------------------------

class MatchingAction
{
public:
  MatchingAction()
  {
  }

  virtual ~MatchingAction()
  {
  }

  //
  // The one virtual method in MatchingAction is "run", which is called
  // with the buffer information received from another context.
  //

  virtual void run(char *buffer, int length) = 0;
};


//----------------------------------------------------------------------
//
// class RemoveReference
//
// Given a class T, if it is a reference type, strip off the
// reference to give a regular type.  This is used in many of the
// subclasses of MatchingAction.
//
//----------------------------------------------------------------------

template<class T>
struct RemoveReference
{
  typedef T type_t;
};

template<class T>
struct RemoveReference<T&>
{
  typedef T type_t;
};


//----------------------------------------------------------------------
//
// class Methodhandler
//
// This class wraps around pointer to an object of type T, and a pointer
// to a method in that class.  This is then used as a functor object
// for the FuncitonAction0 ... 4 classes, and provides templated op()()
// methods for different argument counts.  The proper op()() method will
// be called by
// the corresponding FunctionAction0 ... FunctionAction4 class, with
// the right number of arguments, and then the requested method in the
// given instance of class T will be called with the arguments.
//
//----------------------------------------------------------------------

template<class Method, class T>
class MethodHandler
{
public:
  //
  // constructor must be called with a pointer to the class, and
  // the method pointer.
  //

  inline MethodHandler(const Method &m, T *p)
    : method_m(m), ptr_m(p)
  {
  }

  inline MethodHandler(const MethodHandler<Method, T> &mh)
    : method_m(mh.method_m), ptr_m(mh.ptr_m)
  {
  }

  //
  // destructor does nothing
  //

  inline ~MethodHandler()
  {
  }

  //
  // Versions of operator() for from 0 ... 4 arguments.  Only the relevant one
  // will end up being called from FunctionAction*.
  //

  inline void operator()()
  {
    (ptr_m->*method_m)();
  }

  template<class T1>
  inline void operator()(T1 &x1)
  {
    (ptr_m->*method_m)(x1);
  }

  template<class T1, class T2>
  inline void operator()(T1 &x1, T2 &x2)
  {
    (ptr_m->*method_m)(x1, x2);
  }

  template<class T1, class T2, class T3>
  inline void operator()(T1 &x1, T2 &x2, T3 &x3)
  {
    (ptr_m->*method_m)(x1, x2, x3);
  }

  template<class T1, class T2, class T3, class T4>
  inline void operator()(T1 &x1, T2 &x2, T3 &x3, T4 &x4)
  {
    (ptr_m->*method_m)(x1, x2, x3, x4);
  }

private:
  // The method to invoke in our run method.
  Method method_m;

  // The object pointer
  T *ptr_m;
};


//----------------------------------------------------------------------
//
// class ExtraHandler1
//
// This class wraps around a function pointer that takes N-1 arguments
// and an extra object that should be passed in as the first argument
// to the function.  It stores a copy of the extra argument given
// to it at construction time, and provides templated op()() methods for
// different argument counts.  The proper op()() method will be called by
// the corresponding FunctionAction0 ... FunctionAction4 class, with
// the right number of arguments, and then the extra argument will be
// included.
//
//----------------------------------------------------------------------

template<class Handler, class T>
class ExtraHandler1
{
public:
  //
  // constructor must be called with the handler function and the extra
  // argument to give to it when op()() is called.
  //

  inline ExtraHandler1(const Handler &h, const T &extra)
    : handler_m(h), extra_m(extra)
  {
  }

  inline ExtraHandler1(const ExtraHandler1<Handler, T> &eh)
    : handler_m(eh.handler_m), extra_m(eh.extra_m)
  {
  }

  //
  // destructor does nothing
  //

  inline ~ExtraHandler1()
  {
  }

  //
  // Versions of operator() for from 0 ... 4 arguments.  Only the relevant one
  // will end up being called from FunctionAction*.
  //

  inline void operator()()
  {
    handler_m(extra_m);
  }

  template<class T1>
  inline void operator()(T1 &x1)
  {
    handler_m(extra_m, x1);
  }

  template<class T1, class T2>
  inline void operator()(T1 &x1, T2 &x2)
  {
    handler_m(extra_m, x1, x2);
  }

  template<class T1, class T2, class T3>
  inline void operator()(T1 &x1, T2 &x2, T3 &x3)
  {
    handler_m(extra_m, x1, x2, x3);
  }

  template<class T1, class T2, class T3, class T4>
  inline void operator()(T1 &x1, T2 &x2, T3 &x3, T4 &x4)
  {
    handler_m(extra_m, x1, x2, x3, x4);
  }

private:
  // The handler to invoke in our run method.
  Handler handler_m;

  // The extra argument to provide to the handler as its first argument.
  T extra_m;
};


//----------------------------------------------------------------------
//
// class FunctionAction0 ... FunctionAction4
//
// Versions of MatchingAction that wrap around functions of from 0 ...
// four arguments.  These different versions will unpack the message
// into the proper types of arguments, then call the function, when run()
// is called.
//
// SerializeTag is a tag type used to identify what specialization of
// Cheetah::Serialize<Tag, T> should be used to unpack items.
//
//----------------------------------------------------------------------

template<class SerializeTag, class Handler>
class FunctionAction0 : public MatchingAction
{
public:
  //
  // constructor must be called with the handler to invoke
  //

  FunctionAction0(const Handler &h)
    : handler_m(h)
  {
  }

  //
  // destructor does nothing
  //

  virtual ~FunctionAction0()
  {
  }

  //
  // The one virtual method in MatchingAction is "run", which is called
  // with the buffer information received from another context.
  //

  virtual void run(char *buffer, int length)
  {
    //
    // There should not be anything left in the buffer if this is called.
    //
    CheetahRTAssert(length == 0);

    //
    // Execute the handler, with no arguments.
    //
    handler_m();
  }

private:
  // The handler to invoke in our run method.
  Handler handler_m;
};


//----------------------------------------------------------------------

template<class SerializeTag, class Handler, class T1>
class FunctionAction1 : public MatchingAction
{
public:
  //
  // constructor must be called with the handler to invoke
  //

  FunctionAction1(const Handler& h)
    : handler_m(h)
  {
  }

  //
  // destructor does nothing
  //

  virtual ~FunctionAction1()
  {
  }

  //
  // The one virtual method in MatchingAction is "run", which is called
  // with the buffer information received from another context.
  //

  virtual void run(char *buffer, int length)
  {
    CheetahRTAssert(buffer != 0);
    CheetahRTAssert(length >= 0);

    //
    // Unpack the one argument from the buffer.  If the user
    // function takes its value by reference, strip that off first.
    //
    typedef typename RemoveReference<T1>::type_t P1_t;
    P1_t *p1 = 0;
    int n = Serialize<SerializeTag,P1_t>::unpack(p1, buffer);
    CheetahRTAssert(n <= length);

    //
    // There should not be anything left in the buffer at this point.
    //
    CheetahRTAssert(n == length);

    //
    // Call the function with this argument.
    //
    handler_m(*p1);

    //
    // Delete the argument in a manner consistent with how it was
    // unpacked.
    //
    Serialize<SerializeTag,P1_t>::cleanup(p1);
  }

private:
  // The handler to invoke in our run method.
  Handler handler_m;
};


//----------------------------------------------------------------------

template<class SerializeTag, class Handler, class T1, class T2>
class FunctionAction2 : public MatchingAction
{
public:
  //
  // constructor must be called with the handler to invoke
  //

  FunctionAction2(const Handler& h)
    : handler_m(h)
  {
  }

  //
  // destructor does nothing
  //

  virtual ~FunctionAction2()
  {
  }

  //
  // The one virtual method in MatchingAction is "run", which is called
  // with the buffer information received from another context.
  //

  virtual void run(char *buffer, int length)
  {
    CheetahRTAssert(buffer != 0);
    CheetahRTAssert(length >= 0);

    //
    // Unpack the one argument from the buffer.  If the user
    // function takes its value by reference, strip that off first.
    //
    typedef typename RemoveReference<T1>::type_t P1_t;
    P1_t *p1 = 0;
    int n = Serialize<SerializeTag,P1_t>::unpack(p1, buffer);
    CheetahRTAssert(n <= length);

    typedef typename RemoveReference<T2>::type_t P2_t;
    P2_t *p2 = 0;
    n += Serialize<SerializeTag,P2_t>::unpack(p2, buffer + n);
    CheetahRTAssert(n <= length);

    //
    // There should not be anything left in the buffer at this point.
    //
    CheetahRTAssert(n == length);

    //
    // Call the function with this argument.
    //
    handler_m(*p1, *p2);

    //
    // Delete the argument in a manner consistent with how it was
    // unpacked.
    //
    Serialize<SerializeTag,P1_t>::cleanup(p1);
    Serialize<SerializeTag,P2_t>::cleanup(p2);
  }

private:
  // The handler to invoke in our run method.
  Handler handler_m;
};


//----------------------------------------------------------------------

template<class SerializeTag, class Handler, class T1, class T2, class T3>
class FunctionAction3 : public MatchingAction
{
public:
  //
  // constructor must be called with the handler to invoke
  //

  FunctionAction3(const Handler& h)
    : handler_m(h)
  {
  }

  //
  // destructor does nothing
  //

  virtual ~FunctionAction3()
  {
  }

  //
  // The one virtual method in MatchingAction is "run", which is called
  // with the buffer information received from another context.
  //

  virtual void run(char *buffer, int length)
  {
    CheetahRTAssert(buffer != 0);
    CheetahRTAssert(length >= 0);

    //
    // Unpack the one argument from the buffer.  If the user
    // function takes its value by reference, strip that off first.
    //
    typedef typename RemoveReference<T1>::type_t P1_t;
    P1_t *p1 = 0;
    int n = Serialize<SerializeTag,P1_t>::unpack(p1, buffer);
    CheetahRTAssert(n <= length);

    typedef typename RemoveReference<T2>::type_t P2_t;
    P2_t *p2 = 0;
    n += Serialize<SerializeTag,P2_t>::unpack(p2, buffer + n);
    CheetahRTAssert(n <= length);

    typedef typename RemoveReference<T3>::type_t P3_t;
    P3_t *p3 = 0;
    n += Serialize<SerializeTag,P3_t>::unpack(p3, buffer + n);
    CheetahRTAssert(n <= length);

    //
    // There should not be anything left in the buffer at this point.
    //
    CheetahRTAssert(n == length);

    //
    // Call the function with this argument.
    //
    handler_m(*p1, *p2, *p3);

    //
    // Delete the argument in a manner consistent with how it was
    // unpacked.
    //
    Serialize<SerializeTag,P1_t>::cleanup(p1);
    Serialize<SerializeTag,P2_t>::cleanup(p2);
    Serialize<SerializeTag,P3_t>::cleanup(p3);
  }

private:
  // The handler to invoke in our run method.
  Handler handler_m;
};


//----------------------------------------------------------------------

template<class SerializeTag, class Handler, class T1, class T2, class T3,
         class T4>
class FunctionAction4 : public MatchingAction
{
public:
  //
  // constructor must be called with the handler to invoke
  //

  FunctionAction4(const Handler& h)
    : handler_m(h)
  {
  }

  //
  // destructor does nothing
  //

  virtual ~FunctionAction4()
  {
  }

  //
  // The one virtual method in MatchingAction is "run", which is called
  // with the buffer information received from another context.
  //

  virtual void run(char *buffer, int length)
  {
    CheetahRTAssert(buffer != 0);
    CheetahRTAssert(length >= 0);

    //
    // Unpack the one argument from the buffer.  If the user
    // function takes its value by reference, strip that off first.
    //
    typedef typename RemoveReference<T1>::type_t P1_t;
    P1_t *p1 = 0;
    int n = Serialize<SerializeTag,P1_t>::unpack(p1, buffer);
    CheetahRTAssert(n <= length);

    typedef typename RemoveReference<T2>::type_t P2_t;
    P2_t *p2 = 0;
    n += Serialize<SerializeTag,P2_t>::unpack(p2, buffer + n);
    CheetahRTAssert(n <= length);

    typedef typename RemoveReference<T3>::type_t P3_t;
    P3_t *p3 = 0;
    n += Serialize<SerializeTag,P3_t>::unpack(p3, buffer + n);
    CheetahRTAssert(n <= length);

    typedef typename RemoveReference<T4>::type_t P4_t;
    P4_t *p4 = 0;
    n += Serialize<SerializeTag,P4_t>::unpack(p4, buffer + n);
    CheetahRTAssert(n <= length);

    //
    // There should not be anything left in the buffer at this point.
    //
    CheetahRTAssert(n == length);

    //
    // Call the function with this argument.
    //
    handler_m(*p1, *p2, *p3, *p4);

    //
    // Delete the argument in a manner consistent with how it was
    // unpacked.
    //
    Serialize<SerializeTag,P1_t>::cleanup(p1);
    Serialize<SerializeTag,P2_t>::cleanup(p2);
    Serialize<SerializeTag,P3_t>::cleanup(p3);
    Serialize<SerializeTag,P4_t>::cleanup(p4);
  }

private:
  // The handler to invoke in our run method.
  Handler handler_m;
};


} // namespace Cheetah

#endif // CHEETAH_MATCHINGHANDLER_MATCHINGACTION_H
