//----------------------------------------------------------------------
// MatchingHandler method definitions.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// Include Files:
//----------------------------------------------------------------------

#include "MatchingHandler/MatchingHandler.h"
#include "Utilities/CheetahAssert.h"


namespace Cheetah {

//----------------------------------------------------------------------
//
// The function that gets registered with Cheetah is a static function,
// and since many MatchingHandlers could be created, we need to keep
// track of them.  Each MatchingHandler has a unique message tag, and
// we use that to find the right instance.
//
//----------------------------------------------------------------------

typedef std::vector<MatchingHandler*> MatchingHandlerVector_t;
static MatchingHandlerVector_t handlers_s;


//----------------------------------------------------------------------
//
// MatchingHandler::registerSelf
//
// Put this MatchingHandler in the container of all MatchingHandlers.
// Use the tag from Controller::registerHandler for the position
// in the vector so we can do constant time lookup.
//
//----------------------------------------------------------------------

void
MatchingHandler::registerSelf()
{
  CheetahRTAssert(!registered());
  assertConsistent();

  // 
  // If we somehow were already registered, deregister first.
  //
  if ( registered() )
    deregisterSelf();

  //
  // Register with the communication layer, and get the handler tag
  // that all ainvokes for this handler will have.
  //
  CheetahRTAssert(controller_m.isValid());
  int tag = controller_m.registerHandler(handlerFunction);

  //
  // Record the message tag and this message handler.  Make space in
  // the array of MatchingHandlers if you have to.
  //
  handlerTag_m = tag;
  if (tag >= handlers_s.size())
    handlers_s.resize(tag+1,0);
  handlers_s[handlerTag_m] = this;

  CheetahRTAssert(registered());
  assertConsistent();
}


//----------------------------------------------------------------------
//
// MatchingHandler::deregisterSelf
//
// If we are already registered, take yourself out of the registry.
//
//----------------------------------------------------------------------

void
MatchingHandler::deregisterSelf()
{
  CheetahRTAssert(registered());
  assertConsistent();

  //
  // Erase this entry in the container and forget the tag.
  //
  handlers_s[handlerTag_m] = 0;
  handlerTag_m = -1;

  //
  // Clear out the queues.
  //
  clearQueues();

  CheetahRTAssert(!registered());
  assertConsistent();
}


//----------------------------------------------------------------------
//
// Delete any outstanding actions and messages.
//
//----------------------------------------------------------------------

void
MatchingHandler::clearQueues()
{
  //
  // Loop over all the actions.
  //
  ActionMapVector_t::iterator pa = actions_m.begin();
  ActionMapVector_t::iterator ea = actions_m.end();
  for (; pa != ea; ++pa)
    {
      //
      // Loop over all the outstanding actions for this remote context
      // and delete any actions there.  We have to delete the action
      // by hand since it is a pointer.
      //
      ActionMap_t::iterator p = pa->begin();
      ActionMap_t::iterator e = pa->end();
      for (; p!=e; ++p)
	delete p->second;

      pa->clear();
    }

  //
  // Loop over all the messages.
  //
  MessageMapVector_t::iterator pm = messages_m.begin();
  MessageMapVector_t::iterator em = messages_m.end();
  for (; pm != em; ++pm)
    pm->clear();
}


//----------------------------------------------------------------------
//
// Null ctor initializes to an unusable state.
//
//----------------------------------------------------------------------

MatchingHandler::MatchingHandler()
  : handlerTag_m(-1), group_m(Group())
{
  assertConsistent();
}


//----------------------------------------------------------------------
//
// Construct with a given controller.  This is now a usable
//  MatchingHandler.  This performs a barrier because every context has
// to construct a corresponding MatchingHandler.
//
//----------------------------------------------------------------------

MatchingHandler::MatchingHandler(Controller p)
  : handlerTag_m(-1), group_m(p.group(0))
{
  controller(p);
  assertConsistent();
  p.barrier(group_m);
}


//----------------------------------------------------------------------
// Construct with a given controller and group.  This is now a usable
// MatchingHandler.  This performs a barrier because every context has
// to construct a corresponding MatchingHandler.
//
//----------------------------------------------------------------------

MatchingHandler::MatchingHandler(Controller p, const Group& g)
  : handlerTag_m(-1), group_m(g)
{
  controller(p);
  assertConsistent();
  p.barrier(group_m);
}


//----------------------------------------------------------------------
//
// MatchingHandler::Dtor
//
//----------------------------------------------------------------------

MatchingHandler::~MatchingHandler()
{
  //
  // Make sure we do not have any outstanding actions or messages
  //

  CheetahRTAssert(actions() == 0);
  CheetahRTAssert(messages() == 0);
  assertConsistent();

  //
  // Deregister yourself from the list of MatchingHandlers.
  //
  if ( registered() )
    deregisterSelf();
}


//----------------------------------------------------------------------
//
// MatchingHandler::handlerFunction
//
// The actual static handler function that gets called whenever a message
// is sent from one context (via a MatchingHandler) to another context.
// This function will be called when a message arrives at the local
// context, sent here from another context.  The message buffer will
// contain some header information on what handler should be used, what
// user-supplied function should be called, and then what the arguments are.
// Since there is just one function for all the handlers and all the
// messages, this handler has to do the following when it is called
//   1. Find out what actual handler this message is for.
//   2. For THAT handler, find the corresponding 'action' to perform.
//   3. If there is an available action, invoke it with the message.
//   4. If there is no available action, queue up the message in the
//      other handler.
//----------------------------------------------------------------------

void
MatchingHandler::handlerFunction(int otherContext, int tag,
				 void *buf, int length)
{
  //
  // Pull out the pointer to the MatchingHandler object that has the
  // queues we're looking for.
  //
  CheetahRTAssert( (tag >= 0) && (tag < handlers_s.size() ) );
  MatchingHandler *mh = handlers_s[tag];

  CheetahRTAssert( mh );
  mh->assertConsistent();
  CheetahRTAssert( (otherContext >= 0) &&
	       (otherContext < mh->controller_m.ncontexts() ) );

  //
  // Pull out the user's tag for this message, by unpacking the header.
  //
  int matchingTag = 0;
  int headerLen = unpackHeader(static_cast<char *>(buf), matchingTag);
  CheetahRTAssert(matchingTag >= 0);
  CheetahRTAssert(headerLen > 0);
  CheetahRTAssert(headerLen <= length);

  //
  // Try to find the action that matches this tag and source processor.
  //
  ActionMap_t &actions = mh->actions_m[otherContext];
  ActionMap_t::iterator p = actions.find(matchingTag);

  //
  // If we found it, do the action and erase it.
  // If not, save the message for later.
  //
  if ( p != actions.end() )
    {
      p->second->run(static_cast<char *>(buf) + headerLen, length - headerLen);
      delete (p->second);
      actions.erase(p);
    }
  else
    {
      mh->messages_m[otherContext][matchingTag] =
	Message_t(buf, length, headerLen);
    }

  mh->assertConsistent();
}


//----------------------------------------------------------------------
//
// MatchingHandler::controller
//
// Set the controller, which lets us set the number of cpus and
// register the handler that gets called by ainvoke.
//
//----------------------------------------------------------------------

void
MatchingHandler::controller(Controller p) 
{
  assertConsistent();

  //
  // If we have previously registered this MatchingHandler, deregister
  // it.
  //
  if ( registered() )
    deregisterSelf();

  //
  // Record the handler.
  //
  controller_m = p;

  //
  // If this is a valid controller, set up other things.
  //
  if ( controller_m.isValid() )
    {
      //
      // One actions and messages list per context.
      //
      actions_m.resize( controller_m.ncontexts() );
      messages_m.resize( controller_m.ncontexts() );

      //
      // Set up the handler tag for ainvoke use.
      //
      registerSelf();
    }
  assertConsistent();
}


//----------------------------------------------------------------------
//
// MatchingHandler::assertConsistent
//
// Check the consistency of this MatchingHandler in various ways.
//
//----------------------------------------------------------------------

void
MatchingHandler::assertConsistent()
{
  //
  // The actions and messages arrays must always be the same size.
  //
  CheetahRTAssert( actions_m.size() == messages_m.size() );

  //
  // If there is no controller, we should not be registered.
  // If there is a controller, we might or might not be registered.
  //
  CheetahRTAssert( (controller_m.isValid()) || (!registered()) );

  //
  // If we are registered, various things must be true, and if we are
  // not, then different things must be true.
  //
  if ( registered() )
    {
      //
      // Both the actions and messages should have a queue for each
      // context.
      //
      int nprocs = controller_m.ncontexts();
      CheetahRTAssert( actions_m.size() == nprocs );
      CheetahRTAssert( messages_m.size() == nprocs );

      //
      // The static list of all MatchingHandlers must have exactly one
      // reference to this MatchingHandler.
      //
      CheetahRTAssert( (handlerTag_m >= 0) && (handlerTag_m<handlers_s.size()) );
      for (int i=0; i<handlers_s.size(); ++i)
	CheetahRTAssert( (this!=handlers_s[i]) ^ (i==handlerTag_m) );
    }
  else
    {
      //
      // There should be no pending actions or messages.
      //

      CheetahRTAssert(actions() == 0);
      CheetahRTAssert(messages() == 0);

      //
      // None of the handlers should refer to this MatchingHandler.
      //
      for (int i=0; i<handlers_s.size(); ++i)
	CheetahRTAssert(this != handlers_s[i] );
    }
}


//----------------------------------------------------------------------
//
// MatchingHandler::actions
//
// Return the current number of actions that are pending for
// messages arriving from the given other context.
//
//----------------------------------------------------------------------

int
MatchingHandler::actions(int fromContext)
{
  int asum = 0;
  if (fromContext < 0)
    {
      int nc = actions_m.size();
      for (int i=0; i < nc; ++i)
	asum += actions_m[i].size();
    }
  else
    {
      CheetahRTAssert(fromContext < actions_m.size());
      asum = actions_m[fromContext].size();
    }

  return asum;
}


//----------------------------------------------------------------------
//
// MatchingHandler::messages
//
// Return the current number of messages that have arrived from
// the given other context, but have not yet been given to an action.
//
//----------------------------------------------------------------------

int
MatchingHandler::messages(int fromContext)
{
  int msum = 0;
  if (fromContext < 0)
    {
      int nc = messages_m.size();
      for (int i=0; i < nc; ++i)
	msum += messages_m[i].size();
    }
  else
    {
      CheetahRTAssert(fromContext < messages_m.size());
      msum = messages_m[fromContext].size();
    }

  return msum;
}


//----------------------------------------------------------------------
//
// MatchingHandler::ainvoke
//
// The basic 'ainvoke' capability.  This will invoke the requested
// handler on the receiving context with the given buffer of bytes
// of the specified length.  The buffer is not changed in any way,
// so the extra amount of information that MatchingHandler includes
// in each message must have been put into the message beforehand.
// This header includes the 'tag' value normally specified in send,
// plus any other MatchingHandler-specific information.  That is why
// no tag value is specified in this routines argument list.
// Header info can be added by calling the methods headerLength() and
// packHeader() (see the comments for these routines for their arguments).
//
//----------------------------------------------------------------------

void
MatchingHandler::ainvoke(int toContext, char *buffer, int length, int *bell)
{
  CheetahRTAssert(controller_m.isValid());
  assertConsistent();
  controller_m.ainvoke(toContext,handlerTag_m,buffer,length,bell);
  assertConsistent();
}


} // namespace Cheetah
