#ifndef CHEETAH_MATCHINGHANDLER_MATCHINGHANDLER_H
#define CHEETAH_MATCHINGHANDLER_MATCHINGHANDLER_H

//-----------------------------------------------------------------------------
// Classes:
//   MatchingHandler
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
//
// MatchingHandler sets up queues to match incoming ainvokes with data
// they need on the receiving side.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Controller/Controller.h"
#include "MatchingHandler/Serialize.h"
#include "MatchingHandler/MatchingAction.h"
#include "MatchingHandler/MatchingMessage.h"
#include "Utilities/CheetahAssert.h"
#include <map>
#include <vector>
#include <memory>


namespace Cheetah {

//----------------------------------------------------------------------
//
// class MatchingHandler
//
// MatchingHandler sets up queues to match incoming ainvokes with data
// they need on the receiving side.  The reason for this is that the
// object that the ainvoke needs to talk to may not be ready to talk
// to it -- it may not even exist yet.
//
// One way to handle this is by making sure that you always use global
// pointers with ainvoke, which means that you don't try to do the
// ainvoke before the object has been created and the pointer sent.
//
// For a number of systems though, that is not the most convenient way
// to do things.  Often you will have some sort of token on the
// sending side which indicates which which object on the receiving
// side this ainvoke is for, but that token is not an actual pointer.
// That is particulary useful when there are thousands of contexts and
// there are corresponding objects on all of them.  The sender will
// want to store only a token for all of them and let the destination
// translate that to an actual pointer.
//
// MatchingHandler is a data-parallel class in that it assumes that
// all the contexts build their MatchingHandlers in the same order.
// The MatchingHandler on this context corresponds with the
// MatchingHandler in the other classes in that messages sent through
// this MatchingHandler connect with the corresponding MatchingHandler
// in the other contexts.
//
// The template parameter 'SerializeTag' for MatchingHandler::send and
// ::request routines is for a pack/unpack
// functor.  It is a tag type, used to identify a specialization of
// the class
//
//    template<class Tag, class T>
//    class Serialize { };
// 
// These specializations are used to pack/unpack at item of type T into
// or out of a buffer.  This is used in the send and request calls.
// Each specialization should provide the following four static functions:
//
//   // Return the storage needed to pack the item of type T
//   static int size(const T &item);
//
//   // Pack an item of type T into the given buffer.  Return space used.
//   static int pack(const T &item, char *buffer, int length);
//
//   // Unpack an item of type T from the given buffer.  Set the given
//   // pointer to point at this item.  Return bytes unpacked.
//   static int unpack(T* &p, char *buffer, int length);
//
//   // Delete the item pointed to by the given pointer, that was
//   // unpacked with a previous call to unpack().
//   static void cleanup(T *p);
//
// See matching/Serialize.h for the general template, and a specialization
// for cheetah-provided pack-unpack with the tag 'CHEETAH'.
//
//----------------------------------------------------------------------

class MatchingHandler
{
public:

  //
  // Default ctor yields a MatchingHandler that can't be used until it
  // is given a controller.
  //
  MatchingHandler();

  //
  // Initialize with a given controller.  It is now ready to use.
  //
  MatchingHandler(Controller);

  //
  // Initialize with a given controller and group.  This group is to be used
  // in subsequent messaging in that all contexts are contexts within the
  // given group.  The handler is now ready to use.
  //
  MatchingHandler(Controller, const Group& g);

  //
  // Dtor cleans up the containers and deregisters this MatchingHandler.
  //
  ~MatchingHandler();

  //
  // Return the controller so you have your way with it.
  //
  Controller controller() const
  {
    return controller_m;
  }

  //
  // Give it a new controller.
  //
  void controller(Controller p);

  //
  // Return the current number of actions that are pending for
  // messages arriving from the given other context.  If no
  // argument is given, count ALL actions.
  //
  int actions(int fromContext = (-1));

  //
  // Return the current number of messages that have arrived from
  // the given other context, but have not yet been given to an action.
  // If no argument is given, count ALL actions.
  int messages(int fromContext = (-1));

  //
  // Send a message to another context, with from 0 ... 4 arguments.
  // toContext indicates what context should request this data.
  // matchingTag indicates what action on the receiving context should
  // be invoked with these arguments.  The receiving side should
  // then call "request" with a handler function for this same tag value
  // within the same MatchingHandler object in the receiving context.
  //
  // All send routines have two forms:
  //   o sendWith(tag, args), with a tag templated used to
  //     select the proper specialization of Serialize<Tag,T>.
  //   o send(args), without a tag.  This calls sendWith(CHEETAH(), args).
  //

  // Send a message with zero arguments.

  template<class SerializeTag>
  void sendWith(const SerializeTag &, int toContext, int matchingTag);

  inline void send(int toContext, int matchingTag)
  {
    sendWith(CHEETAH(), toContext, matchingTag);
  }

  // Send a message with one argument.

  template<class SerializeTag, class T1>
  void sendWith(const SerializeTag &, int toContext, int matchingTag,
		const T1 &);

  template<class T1>
  inline void send(int toContext, int matchingTag, const T1 &x1)
  {
    sendWith(CHEETAH(), toContext, matchingTag, x1);
  }

  // Send a message with two arguments.

  template<class SerializeTag, class T1, class T2>
  void sendWith(const SerializeTag &, int toContext, int matchingTag,
		const T1 &, const T2 &);

  template<class T1, class T2>
  inline void send(int toContext, int matchingTag, const T1 &x1, const T2 &x2)
  {
    sendWith(CHEETAH(), toContext, matchingTag, x1, x2);
  }

  // Send a message with three arguments.

  template<class SerializeTag, class T1, class T2, class T3>
  void sendWith(const SerializeTag &, int toContext, int matchingTag,
		const T1 &, const T2 &, const T3 &);

  template<class T1, class T2, class T3>
  inline void send(int toContext, int matchingTag, const T1 &x1, const T2 &x2,
		   const T3 &x3)
  {
    sendWith(CHEETAH(), toContext, matchingTag, x1, x2, x3);
  }

  // Send a message with four arguments.

  template<class SerializeTag, class T1, class T2, class T3, class T4>
  void sendWith(const SerializeTag &, int toContext, int matchingTag,
		const T1 &, const T2 &, const T3 &, const T4 &);

  template<class T1, class T2, class T3, class T4>
  inline void send(int toContext, int matchingTag, const T1 &x1, const T2 &x2,
		   const T3 &x3, const T4 &x4)
  {
    sendWith(CHEETAH(), toContext, matchingTag, x1, x2, x3, x4);
  }


  //
  // Expect to receive a message from another context with a given
  // tag.  When it comes in, call the handler, and pass it the arguments
  // sent here.  The handler must be a function that takes arguments
  // of the proper types, that match the types sent from the sending
  // context.
  //
  // There are different forms of request:
  //   1. Where the handler is a function that takes N arguments, and
  //      all N are sent from the sending context.
  //   2. Where the handler is a function that takes N arguments, where
  //      N-1 are sent from the sending context and one is supplied
  //      on the local context when request is called.  In this case, this
  //      'extra' argument is passed to the handler as its first argument.
  //   3. and 4. Routines 'requestWith', where a tag type is given for
  //      the type of serializer to use.
  //

  // Handler with zero arguments.

  template<class SerializeTag, class Ret>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)());

  template<class SerializeTag, class Ret, class T>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (T::*handler)(), T *obj);

  template<class Ret>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)())
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler);
  }

  template<class Ret, class T>
  inline void request(int fromContext, int matchingTag,
		      Ret (T::*handler)(), T *obj)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, obj);
  }

  // Handlers with one argument, either from other context or local info.

  template<class SerializeTag, class Ret, class T1>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(T1));

  template<class SerializeTag, class Ret, class T, class T1>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (T::*handler)(T1), T *obj);

  template<class SerializeTag, class Ret, class TLH, class TL1>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(TLH), const TL1 &x0);

  template<class Ret, class T1>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(T1))
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler);
  }

  template<class Ret, class T, class T1>
  inline void request(int fromContext, int matchingTag,
		      Ret (T::*handler)(T1), T *obj)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, obj);
  }

  template<class Ret, class TLH, class TL1>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(TLH), const TL1 &x0)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, x0);
  }

  // Handlers with two arguments, either from other context or local info.

  template<class SerializeTag, class Ret, class T1, class T2>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(T1, T2));

  template<class SerializeTag, class Ret, class T, class T1, class T2>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (T::*handler)(T1,T2), T *obj);

  template<class SerializeTag, class Ret, class TLH, class TL1, class T1>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(TLH, T1), const TL1 &x0);

  template<class Ret, class T1, class T2>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(T1, T2))
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler);
  }

  template<class Ret, class T, class T1, class T2>
  inline void request(int fromContext, int matchingTag,
		      Ret (T::*handler)(T1,T2), T *obj)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, obj);
  }

  template<class Ret, class TLH, class TL1, class T1>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(TLH, T1), const TL1 &x0)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, x0);
  }

  // Handlers with three arguments, either from other context or local info.

  template<class SerializeTag, class Ret, class T1, class T2, class T3>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(T1, T2, T3));

  template<class SerializeTag, class Ret, class T, class T1, class T2,
           class T3>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (T::*handler)(T1,T2,T3), T *obj);

  template<class SerializeTag, class Ret, class TLH, class TL1, class T1,
           class T2>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(TLH, T1, T2), const TL1 &x0);

  template<class Ret, class T1, class T2, class T3>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(T1, T2, T3))
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler);
  }

  template<class Ret, class T, class T1, class T2, class T3>
  inline void request(int fromContext, int matchingTag,
		      Ret (T::*handler)(T1,T2,T3), T *obj)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, obj);
  }

  template<class Ret, class TLH, class TL1, class T1, class T2>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(TLH, T1, T2), const TL1 &x0)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, x0);
  }

  // Handlers with four arguments, either from other context or local info.

  template<class SerializeTag, class Ret, class T1, class T2, class T3,
           class T4>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(T1, T2, T3, T4));

  template<class SerializeTag, class Ret, class T, class T1, class T2,
           class T3, class T4>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (T::*handler)(T1,T2,T3,T4), T *obj);

  template<class SerializeTag, class Ret, class TLH, class TL1, class T1,
           class T2, class T3>
  void requestWith(const SerializeTag &, int fromContext, int matchingTag,
		   Ret (*handler)(TLH, T1, T2, T3), const TL1 &x0);

  template<class Ret, class T1, class T2, class T3, class T4>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(T1, T2, T3, T4))
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler);
  }

  template<class Ret, class T, class T1, class T2, class T3, class T4>
  inline void request(int fromContext, int matchingTag,
		      Ret (T::*handler)(T1,T2,T3,T4), T *obj)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, obj);
  }

  template<class Ret, class TLH, class TL1, class T1, class T2, class T3>
  inline void request(int fromContext, int matchingTag,
		      Ret (*handler)(TLH, T1, T2, T3), const TL1 &x0)
  {
    requestWith(CHEETAH(), fromContext, matchingTag, handler, x0);
  }

  //
  // Make sure the state is consistent.  This is guarded so it will
  // introduce no overhead if assertions are turned off.
  //
#ifdef NDEBUG
  inline void assertConsistent() {}
#else
  void assertConsistent();
#endif

private:
  //
  // Typedefs for the base class for action handlers, and for messages
  //
  typedef MatchingAction  Action_t;
  typedef MatchingMessage Message_t;

  //
  // Containers for actions and messages.  These are maps keyed on an
  // integer tag so we can match them easily.
  //
  typedef std::map<int,Action_t *> ActionMap_t;
  typedef std::map<int,Message_t>  MessageMap_t;

  //
  // We keep one of these containers for each of the remote contexts.
  //
  typedef std::vector<ActionMap_t>  ActionMapVector_t;
  typedef std::vector<MessageMap_t> MessageMapVector_t;

  //
  // Keep track of the controller to use for all the messaging.
  //
  Controller controller_m;

  //
  // The group to use in sending and requesting messages.  All contexts
  // are based on this group.
  //
  const Group& group_m;  

  //
  // Queues for actions and messages from each of the remote contexts.
  // Also a function that clears them out.
  //
  ActionMapVector_t actions_m;
  MessageMapVector_t messages_m;
  void clearQueues();

  //
  // The ainvoke calls this function, which looks in the queues to see
  // if an action has been posted for it.  We also keep track of the
  // tag to use when calling ainvoke on the sending side.
  //
  static void handlerFunction(int who, int tag, void *buf, int length);
  int handlerTag_m;

  //
  // We need to keep track of all of the MatchingHandlers that have
  // been created, so we keep a registry.
  //
  inline bool registered() { return handlerTag_m>=0; }
  void registerSelf();
  void deregisterSelf();

  //
  // The "raw" interface to the actions taken to prepare a buffer for
  // sending, to send a buffer, and to handle an incoming message.
  //

  //
  // Calculate the total number of bytes (including any padding for
  // memory alignment) that is required to add our header information to
  // an outgoing message buffer.  This must be static since it is called
  // from the static handler function for this class.
  //
  static inline int headerLength()
  {
    return Serialize<CHEETAH, int>::size(0);
  }

  //
  // Pack our header information into the provided buffer, and return
  // the number of bytes packed in.  The header consists of:
  //     int = tag of receiving handler, as specified by user
  // This does not have to be static, but we do so to be consistent with
  // other routines that manage the header info.
  //
  static inline int packHeader(char *buf, int tag)
  {
    return Serialize<CHEETAH, int>::pack(tag, buf);
  }

  //
  // Unpack our header information into the provided storage, and return
  // the number of bytes unpacked.
  // This must be static since it is called from the static handler function
  // for this class.
  //
  static inline int unpackHeader(char *buf, int &tag)
  {
    int *p = 0;
    int r = Serialize<CHEETAH, int>::unpack(p, buf);
    tag = *p;
    Serialize<CHEETAH, int>::cleanup(p);
    return r;
  }

  //
  // The main routine that actually performs a 'send' to a given context.
  // This is the basic 'ainvoke' capability.  This will invoke the requested
  // handler on the receiving context with the given buffer of bytes
  // of the specified length.  The buffer is not changed in any way,
  // so the extra amount of information that MatchingHandler includes
  // in each message must have been put into the message beforehand.
  // This header includes the 'tag' value normally specified in send,
  // plus any other MatchingHandler-specific information.  That is why
  // no tag value is specified in this routines argument list.
  // Header info can be added by calling the methods headerLength() and
  // packHeader() (see the comments for these routines for their arguments).
  //
  void ainvoke(int toContext, char *buffer, int length, int *bell = 0); 

  //
  // The main routine that actually performs a 'request' for a given
  // handler.  'Handler' must be a type that derives from MatchingAction
  // since we might have to make a copy of it and store a pointer to its
  // base class in our action-map.  But we don't always have to do that,
  // if the message is available that we are trying to request with this
  // call, we can just invoke the handler directly and not copy it.
  //
  template<class Handler>
  void handle(int otherContext, int matchingTag, Handler &handler);
};


//----------------------------------------------------------------------
//
// MatchingHandler<SerializeTag>::send
//
// Send a message to another context, with from 0 ... 4 arguments.
//
//----------------------------------------------------------------------

template<class SerializeTag>
void
MatchingHandler::sendWith(const SerializeTag &,
			  int toContext,
			  int matchingTag)
{
  // Calculate buffer size, including the header
  int length = headerLength();
  CheetahRTAssert(length > 0);

  // Create buffer with data to send
  char *buffer = new char[length];
  CheetahRTAssert(buffer != 0);

  // Pack the data into this buffer, including the header
  int n = packHeader(buffer, matchingTag);
  CheetahRTAssert(n <= length);

  // Send the buffer now.  If the last argument is '0' (instead of a
  // pointer to a 'bell' integer), the message will definitely get sent
  // before this returns so that we can free the buffer.
  ainvoke(group_m.rank(toContext), buffer, length, 0);

  // Free the buffer.
  delete [] buffer;
}

template<class SerializeTag, class T1>
void
MatchingHandler::sendWith(const SerializeTag &,
			  int toContext,
			  int matchingTag,
			  const T1 &x1)
{
  // Calculate buffer size, including the header
  int length = headerLength() +
    Serialize<SerializeTag, T1>::size(x1);
  CheetahRTAssert(length > 0);

  // Create buffer with data to send
  char *buffer = new char[length];
  CheetahRTAssert(buffer != 0);

  // Pack the data into this buffer, including the header
  int n = packHeader(buffer, matchingTag);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T1>::pack(x1, buffer + n);
  CheetahRTAssert(n <= length);
  
  // Send the buffer now.  If the last argument is '0' (instead of a
  // pointer to a 'bell' integer), the message will definitely get sent
  // before this returns so that we can free the buffer.
  ainvoke(group_m.rank(toContext), buffer, length, 0);

  // Free the buffer.
  delete [] buffer;
}

template<class SerializeTag, class T1, class T2>
void
MatchingHandler::sendWith(const SerializeTag &,
			  int toContext,
			  int matchingTag,
			  const T1 &x1, const T2 &x2)
{
  // Calculate buffer size, including the header
  int length = headerLength() +
    Serialize<SerializeTag, T1>::size(x1) +
    Serialize<SerializeTag, T2>::size(x2);
  CheetahRTAssert(length > 0);

  // Create buffer with data to send
  char *buffer = new char[length];
  CheetahRTAssert(buffer != 0);

  // Pack the data into this buffer, including the header
  int n = packHeader(buffer, matchingTag);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T1>::pack(x1, buffer + n);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T2>::pack(x2, buffer + n);
  CheetahRTAssert(n <= length);

  // Send the buffer now.  If the last argument is '0' (instead of a
  // pointer to a 'bell' integer), the message will definitely get sent
  // before this returns so that we can free the buffer.
  ainvoke(group_m.rank(toContext), buffer, length, 0);

  // Free the buffer.
  delete [] buffer;
}

template<class SerializeTag, class T1, class T2, class T3>
void
MatchingHandler::sendWith(const SerializeTag &,
			  int toContext,
			  int matchingTag,
			  const T1 &x1, const T2 &x2,
			  const T3 &x3)
{
  // Calculate buffer size, including the header
  int length = headerLength() +
    Serialize<SerializeTag, T1>::size(x1) +
    Serialize<SerializeTag, T2>::size(x2) +
    Serialize<SerializeTag, T3>::size(x3);
  CheetahRTAssert(length > 0);

  // Create buffer with data to send
  char *buffer = new char[length];
  CheetahRTAssert(buffer != 0);

  // Pack the data into this buffer, including the header
  int n = packHeader(buffer, matchingTag);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T1>::pack(x1, buffer + n);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T2>::pack(x2, buffer + n);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T3>::pack(x3, buffer + n);
  CheetahRTAssert(n <= length);

  // Send the buffer now.  If the last argument is '0' (instead of a
  // pointer to a 'bell' integer), the message will definitely get sent
  // before this returns so that we can free the buffer.
  ainvoke(group_m.rank(toContext), buffer, length, 0);

  // Free the buffer.
  delete [] buffer;
}

template<class SerializeTag, class T1, class T2, class T3, class T4>
void
MatchingHandler::sendWith(const SerializeTag &,
			  int toContext,
			  int matchingTag,
			  const T1 &x1, const T2 &x2,
			  const T3 &x3, const T4 &x4)
{
  // Calculate buffer size, including the header
  int length = headerLength() +
    Serialize<SerializeTag, T1>::size(x1) +
    Serialize<SerializeTag, T2>::size(x2) +
    Serialize<SerializeTag, T3>::size(x3) +
    Serialize<SerializeTag, T4>::size(x4);
  CheetahRTAssert(length > 0);

  // Create buffer with data to send
  char *buffer = new char[length];
  CheetahRTAssert(buffer != 0);

  // Pack the data into this buffer, including the header
  int n = packHeader(buffer, matchingTag);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T1>::pack(x1, buffer + n);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T2>::pack(x2, buffer + n);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T3>::pack(x3, buffer + n);
  CheetahRTAssert(n <= length);
  n += Serialize<SerializeTag, T4>::pack(x4, buffer + n);
  CheetahRTAssert(n <= length);

  // Send the buffer now.  If the last argument is '0' (instead of a
  // pointer to a 'bell' integer), the message will definitely get sent
  // before this returns so that we can free the buffer.
  ainvoke(group_m.rank(toContext), buffer, length, 0);

  // Free the buffer.
  delete [] buffer;
}


//----------------------------------------------------------------------
//
// MatchingHandler<SerializeTag>::request()
//
// Register a handler for a message sent from another context, with
// 0 ... 4 arguments.
//
//----------------------------------------------------------------------

// Handlers with no arguments.

template<class SerializeTag, class Ret>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)())
{
  typedef Ret (*H_t)();
  FunctionAction0<SerializeTag,H_t> action(handler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class T>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (T::*handler)(),
			     T *obj)
{
  typedef Ret (T::*Method_t)();
  typedef MethodHandler<Method_t, T> H_t;
  H_t methodhandler(handler, obj);
  FunctionAction0<SerializeTag,H_t> action(methodhandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}


// Handlers with one argument, either from other context or local info.

template<class SerializeTag, class Ret, class T1>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(T1))
{
  typedef Ret (*H_t)(T1);
  FunctionAction1<SerializeTag,H_t,T1> action(handler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class T, class T1>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (T::*handler)(T1), T *obj)
{
  typedef Ret (T::*Method_t)(T1);
  typedef MethodHandler<Method_t, T> H_t;
  H_t methodhandler(handler, obj);
  FunctionAction1<SerializeTag,H_t,T1> action(methodhandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class TLH, class TL1>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(TLH),
			     const TL1 &x0)
{
  typedef Ret (*BaseH_t)(TLH);
  typedef ExtraHandler1<BaseH_t, TL1> H_t;
  H_t extrahandler(handler, x0);
  FunctionAction0<SerializeTag,H_t> action(extrahandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

// Handlers with two arguments, either from other context or local info.

template<class SerializeTag, class Ret, class T1, class T2>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(T1, T2))
{
  typedef Ret (*H_t)(T1, T2);
  FunctionAction2<SerializeTag,H_t,T1,T2> action(handler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class T, class T1, class T2>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (T::*handler)(T1,T2),
			     T *obj)
{
  typedef Ret (T::*Method_t)(T1,T2);
  typedef MethodHandler<Method_t, T> H_t;
  H_t methodhandler(handler, obj);
  FunctionAction2<SerializeTag,H_t,T1,T2> action(methodhandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class TLH, class TL1, class T1>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(TLH, T1),
			     const TL1 &x0)
{
  typedef Ret (*BaseH_t)(TLH, T1);
  typedef ExtraHandler1<BaseH_t, TL1> H_t;
  H_t extrahandler(handler, x0);
  FunctionAction1<SerializeTag,H_t,T1> action(extrahandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

// Handlers with three arguments, either from other context or local info.

template<class SerializeTag, class Ret, class T1, class T2, class T3>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(T1, T2, T3))
{
  typedef Ret (*H_t)(T1, T2, T3);
  FunctionAction3<SerializeTag,H_t,T1,T2,T3> action(handler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class T, class T1, class T2, class T3>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (T::*handler)(T1,T2,T3),
			     T *obj)
{
  typedef Ret (T::*Method_t)(T1,T2,T3);
  typedef MethodHandler<Method_t, T> H_t;
  H_t methodhandler(handler, obj);
  FunctionAction3<SerializeTag,H_t,T1,T2,T3> action(methodhandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class TLH, class TL1, class T1,
         class T2>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(TLH, T1, T2),
			     const TL1 &x0)
{
  typedef Ret (*BaseH_t)(TLH, T1, T2);
  typedef ExtraHandler1<BaseH_t, TL1> H_t;
  H_t extrahandler(handler, x0);
  FunctionAction2<SerializeTag,H_t,T1,T2> action(extrahandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

// Handlers with four arguments, either from other context or local info.

template<class SerializeTag, class Ret, class T1, class T2, class T3, class T4>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(T1, T2, T3, T4))
{
  typedef Ret (*H_t)(T1, T2, T3, T4);
  FunctionAction4<SerializeTag,H_t,T1,T2,T3,T4> action(handler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class T, class T1,
         class T2, class T3, class T4>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (T::*handler)(T1,T2,T3,T4),
			     T *obj)
{
  typedef Ret (T::*Method_t)(T1,T2,T3,T4);
  typedef MethodHandler<Method_t, T> H_t;
  H_t methodhandler(handler, obj);
  FunctionAction4<SerializeTag,H_t,T1,T2,T3,T4> action(methodhandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}

template<class SerializeTag, class Ret, class TLH, class TL1, class T1,
         class T2, class T3>
void
MatchingHandler::requestWith(const SerializeTag &, int fromContext,
			     int matchingTag,
			     Ret (*handler)(TLH, T1, T2, T3),
			     const TL1 &x0)
{
  typedef Ret (*BaseH_t)(TLH, T1, T2, T3);
  typedef ExtraHandler1<BaseH_t, TL1> H_t;
  H_t extrahandler(handler, x0);
  FunctionAction3<SerializeTag,H_t,T1,T2,T3> action(extrahandler);
  handle(group_m.rank(fromContext), matchingTag, action);
}


//----------------------------------------------------------------------
//
// MatchingHandler<SerializeTag>::handle
//
// Actually carry out the work of a request in MatchingHandler.
// Set up an action to take when a message comes in from a particular
// remote context with a particular tag through this MatchingHandler.
// The action is in the form of a general functor, and we store the
// action to take in a map of actions.  The matchingTag value indicates
// what 'event' this particular handler is interested in.
//
// If the appropriate message has already arrived, execute the action
// immediately.  Otherwise queue up the action object for later use.
//
//----------------------------------------------------------------------

template <class Handler>
void
MatchingHandler::handle(int otherContext, int matchingTag,
			Handler &handler)
{
  //
  // Look through the list of waiting messages to see if what we match
  // with has already arrived.
  //
  MessageMap_t &messages = messages_m[otherContext];
  MessageMap_t::iterator mp = messages.find(matchingTag);

  if ( mp == messages.end() )
    {
      //
      // It wasn't there.  Save this action for later.  The type
      // 'Handler' must be subclass of MatchingAction.
      //
      actions_m[otherContext][matchingTag] = new Handler(handler);
    }
  else
    {
      //
      // It is there.  Just execute it and take the message out of the map.
      // We give the handler the buffer from the message, and its length.
      //
      handler.run((*mp).second.buffer(), (*mp).second.length());
      messages.erase(mp);
    }
}


} // namespace Cheetah

#endif // CHEETAH_MATCHINGHANDLER_MATCHINGHANDLER_H
