#ifndef CHEETAH_MATCHINGHANDLER_MATCHINGMESSAGE_H
#define CHEETAH_MATCHINGHANDLER_MATCHINGMESSAGE_H

//-----------------------------------------------------------------------------
// Classes:
//   MatchingMessage
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
//
// MatchingMessage stores data for a single buffer that has been received
// from another node.  They are stored as MatchingMessage items until
// a handler is available to process them.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include Files:
//-----------------------------------------------------------------------------

#include "Utilities/CheetahAssert.h"
#include <string.h>


namespace Cheetah {

//----------------------------------------------------------------------
//
// class MatchingMessage
//
// MatchingMessage stores message buffers and lengths for items that
// have been received on a context but are not yet ready for use in
// a user-provided handler.  The buffers might have a header section that
// will not be needed by the handlers, so a header size may be provided
// in the constructor that is used in the buffer() and length() methods
// to skip over the header.
//
// Right now this makes a copy of the data given to it; an optimization
// of this would be to somehow just keep a reference to the buffer given
// to it and then somehow notify the agency that provided the buffer that
// it is no longer needed, when the MatchingMessage is deleted.
//
//----------------------------------------------------------------------

class MatchingMessage
{
public:
  inline MatchingMessage()
    : buf_m(0), length_m(-1), header_m(0)
  {
  }

  inline MatchingMessage(void *b, int l, int hl = 0)
    : buf_m(new char[l]), length_m(l), header_m(hl)
  {
    memcpy(buf_m, b, l);
  }

  inline MatchingMessage(const MatchingMessage& m)
    : buf_m(m.buf_m), length_m(m.length_m), header_m(m.header_m)
  {
    m.buf_m = 0;
    m.length_m = (-1);
    m.header_m = 0;
  }

  inline ~MatchingMessage()
  {
    delete [] buf_m;
  }

  inline void operator=(const MatchingMessage& m)
  {
    CheetahRTAssert(buf_m == 0);
    CheetahRTAssert(length_m == -1);
    CheetahRTAssert(header_m == 0);
    buf_m = m.buf_m;
    length_m = m.length_m;
    header_m = m.header_m;
    m.buf_m = 0;
    m.length_m = (-1);
    m.header_m = 0;
  }

  //
  // Return a pointer to the buffer, skipping over the header.
  //
  inline char *buffer() const
  {
    CheetahRTAssert(buf_m != 0);
    CheetahRTAssert(header_m >= 0);
    return (buf_m + header_m);
  }
  
  //
  // Return the total length of the buffer, minus the header.
  //
  inline int length() const
  {
    CheetahRTAssert(length_m >= 0);
    CheetahRTAssert(header_m >= 0);
    CheetahRTAssert(length_m >= header_m);
    return (length_m - header_m);
  }

private:
  mutable char *buf_m;
  mutable int length_m;
  mutable int header_m;
};


} // namespace Cheetah


#endif // CHEETAH_MATCHINGHANDLER_MATCHINGMESSAGE_H
