#ifndef CHEETAH_UTILITIES_ASSERT_H
#define CHEETAH_UTILITIES_ASSERT_H

//-----------------------------------------------------------------------------
// Classes:
// CheetahRTAssert macro, CheetahRTInsist macro, CheetahCTAssert struct
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// CheetahCTAssert(bool c) is a compile-time assertion macro.
// CheetahRTAssert(bool c) is a run-time assertion macro.
// CheetahRTInsist(bool c, char *m) is a run-time insistence, with an explanatory
// message.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Includes:
//-----------------------------------------------------------------------------

#include "CheetahConfiguration.h"

namespace Cheetah {

//-----------------------------------------------------------------------------
//
// Assertion: exception notification class for assertions.
//
// This class should really be derived from std::runtime_error, but
// unfortunately we don't have good implementation of the library standard
// yet, on compilers other than KCC.  So, this class will keep with the
// "what" method evidenced in the standard, but dispense with inheriting from
// classes for which we don't have implementations...
// 
//-----------------------------------------------------------------------------

class Assertion /* : std::runtime_error */
{
  char *msg_m;
  char *file_m;
  int line_m;
public:
  Assertion(const char *msg, const char *file, int line);
  Assertion(const Assertion &a);
  ~Assertion() { delete[] msg_m; delete [] file_m; }
  Assertion &operator=(const Assertion &a);
  const char *what() const { return msg_m; }
  const char *file() const { return file_m; }
  int line() const { return line_m; }
  
  template<class OS>
  void print(OS &os) const
  {
    os << "### Cheetah Assertion Failure ###\n";
    os << "### " << what() << "\n";
    os << "### File " << file() << "; Line " << line();
  }
};

//
// This is the function called in the assert/insist macros.
//

void toss_cookies(const char *msg, const char *file, int line ...);

} // namespace Cheetah

//-----------------------------------------------------------------------------
//
// CheetahCTAssert: CheetahCTAssert is a compile-time assert.
//
// That is, it tests the condition at compile time and if it is false
// you get a compile error that it can't find CheetahCTAssert<false>::test().
// 
// If NOCheetahCTAssert is defined, CheetahCTAssert will revert to the equivalent of
// CheetahRTAssert. To turn off the test completely, define NOCheetahRTAssert as well.
//
//-----------------------------------------------------------------------------

#if defined(NOCheetahCTAssert)

#if defined(NOCheetahRTAssert)

#define CheetahCTAssert(c)

#else

#define CheetahCTAssert(c) if (c) {} else ::Cheetah::toss_cookies(#c, __FILE__, __LINE__)

#endif

#else

#define CheetahCTAssert(c) CheetahCTAssertStruct<(c)>::test()

template<bool B> struct CheetahCTAssertStruct {};

template<> struct CheetahCTAssertStruct<true> { static void test() {} };

#endif


//-----------------------------------------------------------------------------
//
// CheetahRTAssert: Run-time assertion
//
// Now we define a run time assertion mechanism.  We will call it "CheetahRTAssert",
// to reflect the idea that this is for use in Cheetah per se, recognizing that
// there are numerous other assertion facilities in use in client codes.
//
// The CheetahRTAssert macro is intended to be used for validating preconditions
// which must be true in order for following code to be correct, etc.  For
// example, CheetahRTAssert( x > 0. ); y = sqrt(x);  If the assertion fails, the code
// should just bomb.  Philosophically, it should be used to feret out bugs in
// preceding code, making sure that prior results are within reasonable
// bounds before proceeding to use those results in further computation, etc.
// 
// If exceptions are completely turned off (by setting CHEETAH_NO_EXCEPTIONS)
// or NOCheetahRTAssert is defined, CheetahRTAssert will just be an empty macro.
//
//-----------------------------------------------------------------------------

#if defined NOCheetahRTAssert
#define CheetahRTAssert(c) 
#else
#define CheetahRTAssert(c) if (c) {} else ::Cheetah::toss_cookies(#c, __FILE__, __LINE__)
#endif


//-----------------------------------------------------------------------------
//
// CheetahRTInsist:
//
// The CheetahRTInsist macros are like the CheetahRTAssert macro, but they provide the
// opportunity to specify an instructive message.  The idea here is that you
// should use Insist for checking things which are more or less under user
// control.  If the user makes a poor choice, we "insist" that it be
// corrected, providing a corrective hint.
//
//-----------------------------------------------------------------------------

#define CheetahRTInsist(c,m) if (c) {} else ::Cheetah::toss_cookies(m, __FILE__, __LINE__)
#define CheetahRTInsist1(c,m,a1)                                                    \
  if (c) {} else ::Cheetah::toss_cookies(m,__FILE__,__LINE__,a1)
#define CheetahRTInsist2(c,m,a1,a2)                                                 \
  if (c) {} else ::Cheetah::toss_cookies(m,__FILE__,__LINE__,a1,a2)
#define CheetahRTInsist3(c,m,a1,a2,a3)                                              \
  if (c) {} else ::Cheetah::toss_cookies(m,__FILE__,__LINE__,a1,a2,a3)
#define CheetahRTInsist4(c,m,a1,a2,a3,a4)                                           \
  if (c) {} else ::Cheetah::toss_cookies(m,__FILE__,__LINE__,a1,a2,a3,a4)
#define CheetahRTInsist5(c,m,a1,a2,a3,a4,a5)                                        \
  if (c) {} else ::Cheetah::toss_cookies(m,__FILE__,__LINE__,a1,a2,a3,a4,a5)
#define CheetahRTInsist6(c,m,a1,a2,a3,a4,a5,a6)                                     \
  if (c) {} else ::Cheetah::toss_cookies(m,__FILE__,__LINE__,a1,a2,a3,a4,a5,a6)
#define CheetahRTInsist7(c,m,a1,a2,a3,a4,a5,a6,a7)                                  \
  if (c) {} else ::Cheetah::toss_cookies(m,__FILE__,__LINE__,a1,a2,a3,a4,a5,a6,a7)


//-----------------------------------------------------------------------------
// NOTE:  We provide a way to eliminate assertions, but not insistings.  The
// idea is that CheetahRTAssert is used to perform sanity checks during program
// development, which you might want to eliminate during production runs for
// performance sake.  CheetahRTInsist is used for things which really really must be
// true, such as "the file must've been opened", etc.  So, use CheetahRTAssert for
// things which you want taken out of production codes (like, the check might
// inhibit inlining or something like that), but use CheetahRTInsist for those things
// you want checked even in a production code.
//-----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////

#endif // CHEETAH_UTILITIES_PASSERT_H
